/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Setup a pingreply using an Click CutThrough channel.
 */
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_event.h>
#include <nodeos/an_info.h>
#include <nodeos/impl/an_gprof.h>

#define RUNFOR	60	/* run for this many seconds */
#define MAXBUFS	64

static void ee_init(void *arg);
an_flowinit_func_t InitEE = { ee_init };

static void ee_die(void *event, void *arg);
static void setup_channels(void);

static an_chan_t *ethinchan, ethinchan_data, *ethoutchan, ethoutchan_data;
static an_chan_t *cutchan, cutchan_data;
static an_pbuf_t cutbufs_data[MAXBUFS];

an_ethaddr_t myaddr = {
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, {0}, ETHERTYPE_IP
};

/*
 * A totally contrived Click router graph. Mostly to ensure that
 * stuff is working. What is does? Take all incoming packets and
 * classify them into ARP requests, IP packets, or anything else.
 * Reply to ARP requests, so we caan hand out our own address.
 * Send IP packets into the pinger. Drop everything else.
 */
char click_graph[] =
	"in :: FromInChan;"
	"out :: ToOutChan;"
	"cl :: Classifier(12/0806 20/0001, 12/0800, -); "
#if 1
	"ar :: ARPResponder(155.99.214.151 255.255.255.255 00:a0:c9:61:66:d1);"
#else
	"ar :: ARPResponder(206.163.153.21 255.255.255.255 00:00:c0:67:6f:d6);"
#endif
	"in -> cl;"
#ifdef DEBUG
	"cl [0] -> Print(arin) -> ar -> Print(arout) -> out;"
	"cl [1] -> "
		  "Print(ok) -> "
		  "ICMPPing() -> out;"
	"cl [2] -> Print(cl2) -> Discard;"
#else
	"cl [0] -> ar -> out;"
	"cl [1] -> ICMPPing() -> out;"
	"cl [2] -> Discard;"
#endif
	;

int
ee_main(int argc, const char **argv)
{
	assert(argc >= 1);

	printf("%p/%p: %s started\n",
	       an_flow_current(), an_thread_current(), argv[0]);

	/*
	 * Setup channels
	 */
	setup_channels();

	printf("(%02x:%02x:%02x:%02x:%02x:%02x) ready\n",
	       myaddr.daddr[0], myaddr.daddr[1], myaddr.daddr[2],
	       myaddr.daddr[3], myaddr.daddr[4], myaddr.daddr[5]);

	an_gprof_start();
	return 0;
}

static void
setup_channels(void)
{
	an_error_t err;
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_cpuspec_t cpu;
		an_pbuf_t *bufs[MAXBUFS];
	} attr;
	an_flow_t *flow;
	an_infoif_t info[8];
	int i, nif = 8;

	flow = an_flow_current();
	memset(&attr, 0, sizeof attr);

	/*
	 * Find my address.
	 * XXX assumes first interface is ethernet.
	 */
	err = an_info_getifs(info, &nif);
	assert(err == 0);
	assert(nif != 0);
	assert(info[0].type == ANIF_ETH);
	memcpy(myaddr.daddr, info[0].addr.eth.daddr, sizeof myaddr.daddr);

	/*
	 * Setup a timer event for RUNFOR seconds
	 */
	err = an_event_schedule(ee_die, 0, RUNFOR * 1000000);
	assert(err == 0);
	
	/*
	 * Setup in channel for ICMP/IP packets.
	 * No demux; all packets are handed off to the Click graph.
	 * No address; means a "raw" interface at the moment. Entire packet
	 * including the ethernet header is handed off.
	 */
	strncpy(attr.proto.data, info[0].pname, sizeof info[0].pname);
	attr.proto.len = strlen(attr.proto.data);
	memset(&attr.addr, 0, sizeof attr.addr);
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 8;
	for (i = 0; i < MAXBUFS; i++) {
		attr.bufs[i] = an_pbuf_create(&cutbufs_data[i], AN_PBUF_ETH,
					      0, 0);
		assert(attr.bufs[i] != 0);
	}
	ethinchan = an_inchan_create(&ethinchan_data, flow, 0,
				     &attr.proto, &attr.addr,
				     attr.bufs, MAXBUFS, &attr.cpu, 0, 0);
	assert(ethinchan != 0);

	printf("%p: inchannel for %02x:%02x:%02x:%02x:%02x:%02x, type=%x\n",
	       ethinchan,
	       myaddr.daddr[0], myaddr.daddr[1], myaddr.daddr[2],
	       myaddr.daddr[3], myaddr.daddr[4], myaddr.daddr[5],
	       ntohs(myaddr.type));

	/*
	 * Setup a raw ethernet output channel.
	 */
	memset(&attr, 0, sizeof attr);
	strncpy(attr.proto.data, info[0].pname, sizeof info[0].pname);
	attr.proto.len = strlen(attr.proto.data);
	memset(&attr.addr, 0, sizeof attr.addr);
	ethoutchan = an_outchan_create(&ethoutchan_data, flow,
				       &attr.proto, &attr.addr, 0);
	assert(ethoutchan != 0);
	printf("%p: raw ethernet outchannel\n", ethoutchan);

	/*
	 * Create the Click Cut Channel. 
	 */
	cutchan = an_clickchan_create(&cutchan_data, flow,
				      click_graph, ethinchan, ethoutchan);
	
	printf("%p: cut channel\n", cutchan);
	assert(cutchan);
}

static void
ee_die(void *event, void *arg)
{
	int err;
	an_chanstats_t chanstats;

	an_gprof_stop();

	err = an_chan_stats(cutchan, &chanstats);
	assert(err == 0);
	printf("All done!\n");

	printf("All done: %d pkts (%d pkts/sec), %d bytes (%d bytes/sec)\n",
	       chanstats.inpackets, chanstats.inpackets/RUNFOR,
	       chanstats.inbytes, chanstats.inbytes/RUNFOR);
	printf("  Packets: %d queued, %d incopied, %d outcopied, %d dropped\n",
	       chanstats.inbufqueued, chanstats.inbufcopies,
	       chanstats.outbufcopies, chanstats.droppackets);
	printf("  Drops: %d no bufs, %d/%d failed thread/buf\n",
	       chanstats.dropnobufs, chanstats.dropfailedthread,
	       chanstats.dropfailedbuf);

	if (cutchan)
		an_clickchan_destroy(cutchan);
	an_flow_destroy(an_flow_current());
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "ClickPingreplyEE" }; 

	status = ee_main(argc, argv);
}

