/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Forward ETHERTYPE_IP packets destined to our MAC address on if1,
 * out to if2 and visa versa.  Note that this is not true IP forwarding
 * as we do no checksuming or TTL adjustment.  This just tests a simple
 * cut channel.
 *
 *	host1 <-- net1 --> me <-- net2 --> host2
 *
 *	falcon		 chukar		  vulture
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_info.h>

#include <nodeos/impl/an_gprof.h>	/* XXX */

#define RUNFOR	(1*60)	/* run for this many seconds */
#define MAXBUFS	32

/*
 * For some reason the headers define ETHER_MAX_LEN as
 * (ETH_FRAME_LEN + ETH_CRC_LEN), but then don't define ETH_CRC_LEN.
 */
#ifdef __linux__
#ifndef ETH_CRC_LEN
#define ETH_CRC_LEN ETHER_CRC_LEN
#endif
#endif

static void ee_init(void *arg);
an_flowinit_func_t InitEE = { ee_init };

static void ee_arp(an_pbuf_t *pbuf, void *arg);
static void ee_die(an_pbuf_t *pbuf, void *arg);
static an_pbuf_t diebuf_data;
static void setup_channels(void);
static an_chan_t *n1in, n1in_data, *n1out, n1out_data, *n1cut, n1cut_data;
static an_chan_t *n2in, n2in_data, *n2out, n2out_data, *n2cut, n2cut_data;
static an_chan_t *die, die_data;
static an_chan_t *arpinchan1, arpinchan1_data;
static an_chan_t *arpinchan2, arpinchan2_data;
static an_chan_t *arpoutchan, arpoutchan_data;
static an_demuxkey_t arpdemux1_data, *arpdemux1;
static an_demuxkey_t arpdemux2_data, *arpdemux2;
static an_pbuf_t arpbuf1_data, arpbuf2_data;
static an_pbuf_t cutbufs_data[MAXBUFS];

static unsigned char myn1ip[4] = { 155, 99, 213, 190 };
static unsigned char myn2ip[4] = { 155, 99, 213, 198 };

static unsigned int myip1addr, myip2addr;

an_ethaddr_t bcastaddr = {
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff}, {0}, ETHERTYPE_ARP
};
an_ethaddr_t myn1addr = {
	/* chukar eth1 */
	{0x00, 0xa0, 0xc9, 0xdb, 0xf7, 0xb4}, {0}, ETHERTYPE_IP
};
an_ethaddr_t myn2addr = {
	/* chukar eth2 */
	{0x00, 0x90, 0x27, 0xa4, 0xa0, 0xbc}, {0}, ETHERTYPE_IP
};
an_ethaddr_t h1addr = {
	/* falcon eth1 */
	{0x00, 0xa0, 0xc9, 0xdb, 0xcc, 0x62}, {0}, ETHERTYPE_IP	
};
an_ethaddr_t h2addr = {
	/* vulture eth1 */
	{0x00, 0x90, 0x27, 0xa4, 0xa0, 0xb8}, {0}, ETHERTYPE_IP
};
an_ethaddr_t arpoutaddr = {
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, {0}, ETHERTYPE_ARP
};

int
ee_main(int argc, const char **argv)
{
	assert(argc >= 1);

	printf("%p/%p: %s started\n",
	       an_flow_current(), an_thread_current(), argv[0]);
	myip1addr = *(unsigned int *)myn1ip;
	myip2addr = *(unsigned int *)myn2ip;

	setup_channels();

	printf("Ready\n");

	an_gprof_start();
#ifdef OSKIT
	{
		extern int memlogging;
		memlogging = 1;
	}
#endif        
	return 0;
}

static void
setup_channels(void)
{
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_cpuspec_t cpu;
		an_pbuf_t *bufs[MAXBUFS];
	} attr;
	an_flow_t *flow;
	static an_infoif_t info[8];
	an_infoif_t *if1info = 0, *if2info = 0;
	an_error_t err;
	int i, nif = 8;
	struct ether_arp *arp = 0;
	unsigned short val_s;

	flow = an_flow_current();
	memset(&attr, 0, sizeof attr);

	/*
	 * Find addresses for desired interfaces.
	 */
	err = an_info_getifs(info, &nif);
	assert(err == 0);
	assert(nif != 0);
	for (i = 0; i < nif; i++) {
		if (info[i].type != ANIF_ETH)
			continue;
		if (memcmp(info[i].addr.eth.daddr, myn1addr.daddr, 6) == 0) {
			if1info = &info[i];
			if1info->id = 1; /* XXX */
		} else if (memcmp(info[i].addr.eth.daddr,
				  myn2addr.daddr, 6) == 0) {
			if2info = &info[i];
			if2info->id = 2; /* XXX */
		}
	}
	assert(if1info != 0);
	assert(if2info != 0);

	/*
	 * Setup a clock channel to time us out in RUNFOR seconds
	 */
	strcpy(attr.proto.data, "clock");
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.clock.freq = (RUNFOR * 100);
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 1;
	attr.bufs[0] = an_pbuf_create(&diebuf_data, AN_PBUF_CLOCK, 0, 0);
	assert(attr.bufs[0] != 0);
	die = an_inchan_create(&die_data, flow, 0, &attr.proto, &attr.addr,
			       attr.bufs, 1, &attr.cpu, ee_die, 0);
	assert(die != 0);
	printf("%p: clockchannel for %d seconds\n", die, RUNFOR);
	
	/*
	 * Setup in channel for broadcast (arp) packets
	 * First, create a demux to match ARP requests.
	 */
	arpdemux1 = an_demuxkey_create(&arpdemux1_data, an_flow_current());
	assert(arpdemux1 != 0);
	val_s = htons(ARPHRD_ETHER);
	err = an_demuxkey_add(arpdemux1, (int)&arp->arp_hrd, sizeof val_s,
			       &val_s, 0);
	assert(err == 0);
	val_s = htons(ETHERTYPE_IP);
	err = an_demuxkey_add(arpdemux1, (int)&arp->arp_pro, sizeof val_s,
			       &val_s, 0);
	assert(err == 0);
	val_s = htons(ARPOP_REQUEST);
	err = an_demuxkey_add(arpdemux1, (int)&arp->arp_op, sizeof val_s,
			       &val_s, 0);
	assert(err == 0);

	arpdemux2 = an_demuxkey_clone(&arpdemux2_data, arpdemux1);
	assert(arpdemux2 != 0);

	/*
	 * Now create an inchan for both interfaces
	 */
	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", if1info->id);
	attr.proto.len = strlen(attr.proto.data);
	bcastaddr.type = htons(bcastaddr.type);
	attr.addr.eth = bcastaddr;
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 2;
	attr.bufs[0] = an_pbuf_create(&arpbuf1_data, AN_PBUF_ETH, 0, 0);
	assert(attr.bufs[0] != 0);
	arpinchan1 = an_inchan_create(&arpinchan1_data, flow, arpdemux1,
				      &attr.proto, &attr.addr, attr.bufs, 1,
				      &attr.cpu, ee_arp, if1info);
	assert(arpinchan1 != 0);
	printf("%p: ARP inchannel for %s\n", arpinchan1, attr.proto.data);

	sprintf(attr.proto.data, "if%d", if2info->id);
	attr.proto.len = strlen(attr.proto.data);
	attr.bufs[0] = an_pbuf_create(&arpbuf2_data, AN_PBUF_ETH, 0, 0);
	assert(attr.bufs[0] != 0);
	arpinchan2 = an_inchan_create(&arpinchan2_data, flow, arpdemux2,
				      &attr.proto, &attr.addr, attr.bufs, 1,
				      &attr.cpu, ee_arp, if2info);
	assert(arpinchan2 != 0);
	printf("%p: ARP inchannel for %s\n", arpinchan2, attr.proto.data);

	/*
	 * Build up cutchan to forward from if1 -> if2:
	 *	- inchan for our MAC addr on if1 (eth1)
	 *	- outchan to host2 MAC addr on if2 (eth2)
	 */
	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", if1info->id);
	attr.proto.len = strlen(attr.proto.data);
	myn1addr.type = htons(myn1addr.type);
	attr.addr.eth = myn1addr;
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 8;
	for (i = 0; i < MAXBUFS/2; i++) {
		attr.bufs[i] = an_pbuf_create(&cutbufs_data[i], AN_PBUF_ETH,
					      0, 0);
		assert(attr.bufs[i] != 0);
	}
	n1in = an_inchan_create(&n1in_data, flow, 0, &attr.proto, &attr.addr,
				attr.bufs, MAXBUFS/2, &attr.cpu, 0, 0);
	assert(n1in != 0);
	printf("%p: inchan for %02x:%02x:%02x:%02x:%02x:%02x, type=%x\n",
	       n1in,
	       myn1addr.daddr[0], myn1addr.daddr[1], myn1addr.daddr[2],
	       myn1addr.daddr[3], myn1addr.daddr[4], myn1addr.daddr[5],
	       ntohs(myn1addr.type));
	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", if2info->id);
	attr.proto.len = strlen(attr.proto.data);
	h2addr.type = htons(h2addr.type);
	attr.addr.eth = h2addr;
	n2out = an_outchan_create(&n2out_data, flow, &attr.proto, &attr.addr,
				  0);
	assert(n2out != 0);
	printf("%p: outchan to %02x:%02x:%02x:%02x:%02x:%02x, type=%x\n",
	       n2out,
	       h2addr.daddr[0], h2addr.daddr[1], h2addr.daddr[2],
	       h2addr.daddr[3], h2addr.daddr[4], h2addr.daddr[5],
	       ntohs(h2addr.type));
	n1cut = an_cutchan_create(&n1cut_data, flow, 0, n1in, n2out);
	assert(n1cut != 0);

	/*
	 * Build up cutchan to forward from if2 -> if1:
	 *	- inchan for our MAC addr on if2 (eth2)
	 *	- outchan to host1 MAC addr on if1 (eth1)
	 */
	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", if2info->id);
	attr.proto.len = strlen(attr.proto.data);
	myn2addr.type = htons(myn2addr.type);
	attr.addr.eth = myn2addr;
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 8;
	for (i = 0; i < MAXBUFS/2; i++) {
		attr.bufs[i] = an_pbuf_create(&cutbufs_data[i+MAXBUFS/2],
					      AN_PBUF_ETH, 0, 0);
		assert(attr.bufs[i] != 0);
	}
	n2in = an_inchan_create(&n2in_data, flow, 0, &attr.proto, &attr.addr,
				attr.bufs, MAXBUFS/2, &attr.cpu, 0, 0);
	assert(n2in != 0);
	printf("%p: inchan for %02x:%02x:%02x:%02x:%02x:%02x, type=%x\n",
	       n2in,
	       myn2addr.daddr[0], myn2addr.daddr[1], myn2addr.daddr[2],
	       myn2addr.daddr[3], myn2addr.daddr[4], myn2addr.daddr[5],
	       ntohs(myn2addr.type));
	memset(&attr, 0, sizeof attr);
	sprintf(attr.proto.data, "if%d", if1info->id);
	attr.proto.len = strlen(attr.proto.data);
	h1addr.type = htons(h1addr.type);
	attr.addr.eth = h1addr;
	n1out = an_outchan_create(&n1out_data, flow, &attr.proto, &attr.addr,
				  0);
	assert(n1out != 0);
	printf("%p: outchan to %02x:%02x:%02x:%02x:%02x:%02x, type=%x\n",
	       n1out,
	       h1addr.daddr[0], h1addr.daddr[1], h1addr.daddr[2],
	       h1addr.daddr[3], h1addr.daddr[4], h1addr.daddr[5],
	       ntohs(h1addr.type));
	n2cut = an_cutchan_create(&n2cut_data, flow, 0, n2in, n1out);
	assert(n2cut != 0);
}

static void
swapn(unsigned char a[], unsigned char b[], int n)
{
	int i;
	unsigned char t[n];

	for (i = 0; i < n; i++) {
		t[i] = a[i];
		a[i] = b[i];
		b[i] = t[i];
	}
}

static char *
iptoa(unsigned long x)
{
	static char buf[4*4];		/* nnn.nnn.nnn.nnnz */
	unsigned char *p = (unsigned char *)&x;

	sprintf(buf, "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);
	return buf;
}

static void
ee_arp(an_pbuf_t *pbuf, void *arg)
{
	an_infoif_t *info = arg;
	int psize = an_pbuf_size(pbuf);
	void *data = an_pbuf_data(pbuf);
	struct ether_arp *arp;
	struct in_addr ip;
	an_error_t err;

	assert(info != 0);

	arp = (struct ether_arp *)data;
	assert(psize >= sizeof *arp);

	ip.s_addr = *(unsigned long *)arp->arp_tpa;
	assert(ntohs(arp->arp_hrd) == ARPHRD_ETHER);
	assert(ntohs(arp->arp_pro) == ETHERTYPE_IP);
	assert(ntohs(arp->arp_op) == ARPOP_REQUEST);

	/* See if it is for us */
	if (ip.s_addr != myip1addr && ip.s_addr != myip2addr)
		goto done;

	/* Send the reply. */
	arp->arp_op = htons(ARPOP_REPLY);
	swapn(arp->arp_spa, arp->arp_tpa, sizeof arp->arp_tpa);
	swapn(arp->arp_sha, arp->arp_tha, sizeof arp->arp_tha);
	if (ip.s_addr == myip1addr)
		memcpy(arp->arp_sha, myn1addr.daddr, sizeof myn1addr.daddr);
	else
		memcpy(arp->arp_sha, myn2addr.daddr, sizeof myn2addr.daddr);

#if 1
	printf("sending ARP reply to %s (%02x:%02x:%02x:%02x:%02x:%02x)\n",
	       iptoa(*(unsigned int *)arp->arp_tpa),
	       arp->arp_tha[0], arp->arp_tha[1], arp->arp_tha[2],
	       arp->arp_tha[3], arp->arp_tha[4], arp->arp_tha[5]);
#endif

	/*
	 * Got a legit packet to send.
	 */
	if (memcmp(arpoutaddr.daddr, arp->arp_tha, sizeof arpoutaddr.daddr) != 0) {
		struct {
			an_chanaddr_t addr;
			an_chanproto_t proto;
		} attr;

		if (arpoutchan)
			an_outchan_destroy(arpoutchan);
		else
			arpoutaddr.type = htons(arpoutaddr.type);
		memset(&attr, 0, sizeof attr);
		sprintf(attr.proto.data, "if%d", info->id);
		attr.proto.len = strlen(attr.proto.data);
		memcpy(arpoutaddr.daddr, arp->arp_tha, sizeof arpoutaddr.daddr);
		attr.addr.eth = arpoutaddr;
		arpoutchan = an_outchan_create(&arpoutchan_data,
					       an_flow_current(),
					       &attr.proto, &attr.addr, 0);
		assert(arpoutchan != 0);
	}
	err = an_outchan_send(arpoutchan, pbuf);
	assert(err == 0);

done:
	an_inchan_addpbuf(info->id == 1 ? arpinchan1 : arpinchan2, pbuf);
}


static void
ee_die(an_pbuf_t *pbuf, void *arg)
{
	an_chanstats_t n1stats, n2stats;
	an_error_t err;

	an_gprof_stop();
	err = an_chan_stats(n1cut, &n1stats);
	assert(err == 0);
	err = an_chan_stats(n2cut, &n2stats);
	assert(err == 0);
	printf("All done!\n");
	printf("Net1 -> Net2: %d pkts (%d pkts/sec), %d bytes (%d bytes/sec)\n",
	       n1stats.inpackets, n1stats.inpackets/RUNFOR,
	       n1stats.inbytes, n1stats.inbytes/RUNFOR);
	printf("Net2 -> Net1: %d pkts (%d pkts/sec), %d bytes (%d bytes/sec)\n",
	       n2stats.inpackets, n2stats.inpackets/RUNFOR,
	       n2stats.inbytes, n2stats.inbytes/RUNFOR);
	an_chan_destroy(n1cut);
	an_chan_destroy(n2cut);
	an_flow_destroy(an_flow_current());
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "IPforward EE" }; 

	status = ee_main(argc, argv);
}
