/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Test the packet scheduler stuff. Send many packets very fast to a set of
 * UDP ports on another machine, which gets them and tells you the relative
 * percentages. At present, there is no way to specify the bandwidth parm,
 * so every outgoing channel gets the same ticket allocation. In the test
 * below, each of 3 channels gets different numbers of packets sent over it,
 * but the receiver should get exactly 33% on each link.
 *
 * The receiver program is the hpfq_udp_tally program that Bart wrote to test
 * the HPFQ stuff. Its actually a standalone program, but lives in the oskit
 * tree. Maybe it should move here?
 */
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <net/if.h> // solaris
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>

#define __FAVOR_BSD	/* XXX get BSD field names in Linux udp.h */
#include <netinet/udp.h>

#include <nodeos/an_impl.h>
#include <nodeos/an_flow.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_mutex.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_info.h>

#include <nodeos/impl/an_gprof.h>	/* XXX */

#define RUNFOR		20	/* run for this many seconds */
#define MAXPORTS	3
#define SRC_PORT	10000
#define DST_PORT	5050
#define PKT_SIZE	256

#define SRC_IP		"206.163.153.26"		/* Poohbear, eth0 */
#define DST_IP		"206.163.153.25"		/* Golden, eth1 */

/* Golden, eth1 */
static unsigned char dst_ethaddr[6] = { 0x00,0xa0,0xc9,0xf1,0x09,0xcb };

/*
 * For some reason the headers define ETHER_MAX_LEN as
 * (ETH_FRAME_LEN + ETH_CRC_LEN), but then don't define ETH_CRC_LEN.
 */
#ifdef __linux__
#ifndef ETH_CRC_LEN
#define ETH_CRC_LEN ETHER_CRC_LEN
#endif
#endif

static void ee_init(void *arg);
an_flowinit_func_t InitEE = { ee_init };

static void ee_die(an_pbuf_t *pbuf, void *arg);
static an_pbuf_t *diebuf, diebuf_data;
static void setup_channels(void);

static an_pbuf_t *buffers[MAXPORTS], buffer_data[MAXPORTS];
static void setup_buffers(void);

static an_chan_t *die, die_data;
static an_chan_t *ethoutchan[MAXPORTS], ethoutchan_data[MAXPORTS];
static int dying;

an_ethaddr_t myaddr = {
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, {0}, ETHERTYPE_IP
};
an_ethaddr_t ipoutaddr = {
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00}, {0}, ETHERTYPE_IP
};

int
ee_main(int argc, const char **argv)
{
	int i;
	
	assert(argc >= 1);

	printf("%p/%p: %s started\n",
	       an_flow_current(), an_thread_current(), argv[0]);

	/*
	 * Setup channels
	 */
	setup_buffers();
	setup_channels();

	printf("%s (%02x:%02x:%02x:%02x:%02x:%02x) ready\n",
	       SRC_IP,
	       myaddr.daddr[0], myaddr.daddr[1], myaddr.daddr[2],
	       myaddr.daddr[3], myaddr.daddr[4], myaddr.daddr[5]);

	an_gprof_start();
#ifdef OSKIT
	{
		extern int memlogging;
		memlogging = 1;
	}
#endif

	for (i = 0; i < 20000; i++) {
		/*
		 * If all is well, the receiver should see exactly 33% on
		 * each port, even though the number of packets sent on
		 * each port is very unequal. Note that the channel code
		 * currently gives every channel an equal share. 
		 */
		an_outchan_send(ethoutchan[0], buffers[0]);
		an_outchan_send(ethoutchan[1], buffers[1]);
		an_outchan_send(ethoutchan[2], buffers[2]);
		
		an_outchan_send(ethoutchan[0], buffers[0]);
		an_outchan_send(ethoutchan[1], buffers[1]);

		an_outchan_send(ethoutchan[0], buffers[0]);
	}
	return 0;
}

static unsigned short
ipcksum(void *buf, int nbytes)
{
	unsigned long sum;

	sum = 0;
	while (nbytes > 1) {
		sum += *(unsigned short *)buf;
		buf += 2;
		nbytes -= 2;
	}
	/* Add in odd byte if any. */
	if (nbytes == 1)
		sum += *(char *)buf;

	sum = (sum >> 16) + (sum & 0xffff);     /* hi + lo */
	sum += (sum >> 16);                     /* add carry, if any */

	return (~sum);
}

static void
setup_channels(void)
{
	an_error_t err;
	struct {
		an_chanaddr_t addr;
		an_chanproto_t proto;
		an_cpuspec_t cpu;
		an_pbuf_t *bufs[1];
	} attr;
	an_flow_t *flow;
	an_infoif_t info[8];
	int nif = 8;
	int i;

	flow = an_flow_current();
	memset(&attr, 0, sizeof attr);

	/*
	 * Find my address.
	 * XXX assumes first interface is ethernet.
	 */
	err = an_info_getifs(info, &nif);
	assert(err == 0);
	assert(nif != 0);
	assert(info[0].type == ANIF_ETH);
	memcpy(myaddr.daddr, info[0].addr.eth.daddr, sizeof myaddr.daddr);

	/*
	 * Setup a clock channel to time us out in RUNFOR seconds
	 */
	strcpy(attr.proto.data, "clock");
	attr.proto.len = strlen(attr.proto.data);
	attr.addr.clock.freq = (RUNFOR * 100);
	attr.cpu.ttype = AN_TH_IMPLICIT;
	attr.cpu.threads.im.maxthreads = 1;
	attr.bufs[0] = diebuf;
	die = an_inchan_create(&die_data, flow, 0, &attr.proto, &attr.addr,
			       attr.bufs, 1, &attr.cpu, ee_die, 0);
	assert(die != 0);
	printf("%p: clockchannel for %d seconds\n", die, RUNFOR);

	/*
	 * Setup raw ethernet output channels.
	 */
	for (i = 0; i < MAXPORTS; i++) {
		memset(&attr, 0, sizeof attr);
		strncpy(attr.proto.data, info[0].pname, sizeof info[0].pname);
		attr.proto.len = strlen(attr.proto.data);
		memset(&attr.addr, 0, sizeof attr.addr);
		ethoutchan[i] = an_outchan_create(&ethoutchan_data[i], flow,
						  &attr.proto, &attr.addr, 0);
		assert(ethoutchan[i] != 0);
		printf("%p: raw ethernet outchannel\n", ethoutchan[i]);
	}
}

static void
ee_die(an_pbuf_t *pbuf, void *arg)
{
	int i;

	dying = 1;
	an_gprof_stop();
	printf("All done!\n");

	for (i = 0; i < MAXPORTS; i++)
		an_outchan_destroy(ethoutchan[i]);

	an_flow_destroy(an_flow_current());
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "PackSchedEE" }; 

	status = ee_main(argc, argv);
}

/*
 * Packet buffer management
 */

void
setup_buffers(void)
{
	an_pbuf_t *bufptr;
	int i;
	static void setup_ether_packet(int which, char *p);

	diebuf = an_pbuf_create(&diebuf_data, AN_PBUF_CLOCK, 0, 0);
	assert(diebuf != 0);

	for (i = 0; i < MAXPORTS; i++) {
		bufptr = an_pbuf_create(&buffer_data[i], AN_PBUF_ETH, 0, 0);
		assert(bufptr != 0);
		buffers[i] = bufptr;

		setup_ether_packet(i, an_pbuf_data(bufptr));
		an_pbuf_setsize(bufptr, sizeof(struct ether_header) +
				sizeof(struct ip) + sizeof(struct udphdr) +
				PKT_SIZE, 0);
	}
}

static void
setup_ether_packet(int which, char *p)
{
	struct ether_header	eth;
	struct ip		ip;
	struct udphdr		udp;
	int			i;
	struct in_addr		src_ip, dst_ip;

	inet_aton(DST_IP, &dst_ip);
	inet_aton(SRC_IP, &src_ip);

	/*
	 * Set up ether header
	 */
	memcpy(&eth.ether_shost, myaddr.daddr, sizeof(myaddr.daddr));
	memcpy(&eth.ether_dhost, dst_ethaddr, sizeof(dst_ethaddr));
	eth.ether_type = htons(ETHERTYPE_IP);

	/*
	 * Setup UDP Header
	 */
	udp.uh_sport = htons(SRC_PORT + which);
	udp.uh_dport = htons(DST_PORT);
	udp.uh_ulen  = htons(sizeof(struct udphdr) + PKT_SIZE);
	udp.uh_sum   = 0;

	/*
	 * Setup IP header
	 */
	ip.ip_v		 = IPVERSION;
	ip.ip_hl	 = (sizeof ip) >> 2;
	ip.ip_tos	 = 0;
	ip.ip_len	 = htons(sizeof(struct ip) +
				 sizeof(struct udphdr) + PKT_SIZE);
	ip.ip_id	 = 69;		/* apparently can't leave this
					  zero or we get dropped by FreeBSD  */
	ip.ip_off	 = 0;
	ip.ip_ttl	 = 64;
	ip.ip_p		 = IPPROTO_UDP;
	ip.ip_src.s_addr = src_ip.s_addr;	/* already in net order */
	ip.ip_dst.s_addr = dst_ip.s_addr;	/* already in net order */
	ip.ip_sum        = 0;
	ip.ip_sum	 = ipcksum(&ip, sizeof(ip));

	/*
	 * Now copy into packet to avoid alignment breakage on ARM32
	 */
	memcpy(p, &eth, sizeof(struct ether_header));
	p += sizeof(struct ether_header);
	
	memcpy(p, &ip, sizeof(struct ip));
	p += sizeof(struct ip);

	memcpy(p, &udp, sizeof(struct udphdr));
	p += sizeof(struct udphdr);

	/*
	 * And fill in some data for the payload.
	 */
	for (i = 0; i < PKT_SIZE; i++)
		*p++ = i;
}

