/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Test various thread functions.
 */

#include <assert.h>
#include <stdio.h>

#include <nodeos/an_thread.h>
#include <nodeos/an_mempool.h>
#include <nodeos/an_flow.h>

static void ee_init(void *arg);
static void dotests(void);

an_flowinit_func_t InitEE = { ee_init };

int
ee_main(int argc, const char **argv)
{
#ifdef GPROF
	oskit_start_gprof();
#endif
	dotests();
#ifdef GPROF
	oskit_stop_gprof();
#endif
	return 0;
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	const char *argv[] = { "NullEE" }; 
	an_flow_t *flow = an_flow_current();

	status = ee_main(argc, argv);
	an_flow_destroy(flow);
}

static void testcurrent(void);
static void testcreatedestroy(void);
static void teststackinfo(void);
static void testdata(void);
static void testyield(void);
static void testsleep(void);

static void dotests(void)
{
	printf("test an_thread_current()\n");
	testcurrent();
	printf("test an_thread_create/destroy()\n");
	testcreatedestroy();
	printf("test an_thread_stackinfo()\n");
	teststackinfo();
	printf("test an_thread_getdata/setdata()\n");
	testdata();
	printf("test an_thread_yield()\n");
	testyield();
	printf("test an_thread_sleep()\n");
	testsleep();

	printf("All tests complete.\n");
}

static void testcurrent(void)
{
	an_thread_t* th;

	th = an_thread_current();
	assert(th);
}

static void testcreatedestroy(void)
{
	an_thread_t newth;
	int stacksize = 20 * 1024; // OSKit pthread min stack size
	void* stack;
	an_thread_t *th;
	an_error_t err;

	stack = an_mempool_alloc(an_mempool_current(), stacksize);
	th = an_thread_create(&newth, stack, stacksize);
	assert(th);
	assert(th == &newth);

	err = an_thread_destroy(th);
	assert(err == 0);
	
	an_mempool_free(an_mempool_current(), stack, stacksize);
}

static void teststackinfo(void)
{
	void *stackbase;
	int stacksize;
	void *stackptr;
	an_error_t err;
	void *fp = __builtin_frame_address(0);

	err = an_thread_stackinfo(an_thread_current(),
				  &stackbase,
				  &stacksize,
				  &stackptr);
	if (err == ENODEV)
	{
		/* Running in unix mode, there is no support for stackinfo(). */
		return;
	}

	assert(err == 0); /* XXX */

	/*     |***********|--ACTIVE-STACK---|          */
	/* stackbase    stackptr    stackbase+stacksize */

	assert(fp > stackbase);
	assert(fp < (stackbase + stacksize));
	assert(fp >= stackptr);
}

static void testdata(void)
{
	int data;
	int data2;
	void *d;

	d = an_thread_getdata();
	assert(d == NULL);

	an_thread_setdata(&data);
	d = an_thread_getdata();
	assert(d = &data);

	an_thread_setdata(&data2);
	d = an_thread_getdata();
	assert(d = &data2);
}

static void testyield(void)
{
	an_thread_yield();
}

static void testsleep(void)
{
	int sleepdelay = 3;

	an_timespec_t start, end, wait;

	an_gettime(&start);

	wait.tv_sec = sleepdelay;
	wait.tv_nsec = 0;
	an_thread_sleep(&wait);

	an_gettime(&end);

	assert(end.tv_sec > start.tv_sec);
	assert((end.tv_sec - start.tv_sec) >= sleepdelay);
}

