/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Channel API
 */
#ifndef _NODEOS_CHANNEL_H_
#define _NODEOS_CHANNEL_H_

#ifdef NODEOS_SPEC_COMPAT
#define an_Channel an_chan_t
#define an_inchannelCreate an_inchan_create
#define an_outchannelCreate an_outchan_create
#define an_cutchannelCreate an_cutchan_create
#define an_channelDestroy an_chan_destroy
#define an_channelSend an_chan_send
#define an_channelStats an_chan_stats
#endif

#include <nodeos/an_error.h>
#include <nodeos/an_impl.h>
#include <nodeos/an_resource.h>
#include <nodeos/an_demuxkey.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/impl/an_chan.h>

/*
 * Create an in-channel as described by the given parameters.  Every packet
 * matching the given protocol, addressing and demultiplex criteria is
 * handed off to the specified delivery function.  Returns a pointer to the
 * new channel object or zero on an error.
 *
 * ``flow'' is the flow with which the channel should be associated,
 * ``key'' contains the filtering criteria to be applied to every packet,
 * ``proto'' describes the protocol processing to apply to every packet,
 * ``addr'' contains the addressing info appropriate for the protocols,
 * ``mem'' specifies the buffering resources to use for packets,
 * ``cpu'' specifies the CPU resources to use,
 * ``func'' is the packet delivery function,
 * ``farg'' is an additional argument provided to the delivery function.
 */
an_chan_t *an_inchan_create(void *mem, an_flow_t *flow, an_demuxkey_t *key,
			    an_chanproto_t *proto, an_chanaddr_t *addr,
			    an_pbuf_t *pbufs[], int npbufs, an_cpuspec_t *cpu,
			    an_chanrecvfunc_t func, void *farg);

/*
 * Create an out-channel as described by the given parameters.  Every packet
 * sent on the channel undergoes the specified protocol processing and is
 * routed based on the addressing info.  Returns a pointer to the new channel
 * object or zero on an error.
 *
 * ``flow'' is the flow with which the channel should be associated,
 * ``proto'' describes the protocol processing to apply to every packet,
 * ``addr'' contains the addressing info appropriate for the protocols,
 * ``net'' specifies the network resources to use,
 */
an_chan_t *an_outchan_create(void *mem, an_flow_t *flow, an_chanproto_t *proto,
			     an_chanaddr_t *addr, an_netspec_t *net);

/*
 * Create a cut-channel by composing the provided in- and out- channels.
 * Additional protocol processing is specified in the given protocol spec.
 * The in-chan and out-chan can no longer be used once they are composed
 * into a cut-channel.
 *
 * ``flow'' is the flow with which the channel should be associated,
 * ``proto'' describes the protocol processing to apply to every packet,
 * ``inchan'' is the in-channel created for receiving packets,
 * ``outchan'' is the out-channel created for sending the packets
 */
an_chan_t *an_cutchan_create(void *mem, an_flow_t *flow, an_chanproto_t *proto,
			     an_chan_t *inchan, an_chan_t *outchan);

/*
 * Cleanly shutdown a channel.
 * Further operations on the channel should return an error.  For
 * in-channels, no further packets will be received, though processing
 * of currently active packets continues.
 *
 * ``chan'' is the channel to destroy.
 */
an_error_t an_chan_destroy(an_chan_t *chan);

#define an_inchan_destroy an_chan_destroy
#define an_outchan_destroy an_chan_destroy
#define an_cutchan_destroy an_chan_destroy

/*
 * Send the indicated packet to an out-channel.
 * Once this function returns, the packet may be safely destroyed or
 * returned to the channel buffer pool.  No indication is provided as
 * to whether the packet was successfully transmitted.  Returns an error
 * if the channel has been destroyed.
 *
 * ``chan'' is the channel to send on.
 * ``packet'' is the packet to send.
 */
an_error_t an_outchan_send(an_chan_t *chan, an_pbuf_t *pbuf);
an_error_t an_outchan_sendv(an_chan_t *chan, an_pbuf_t *pbufs[], int npbufs);

/*
 * Add a packet buffer to a channel's pool.
 * The given pbuf is associated with the specified channel and becomes
 * available for receiving a packet.  Returns zero on success an error
 * otherwise.  Possible errors include an invalid channel or a buffer
 * which is "incompatible" (e.g., too small) with the channel.
 *
 * ``chan'' is the channel to associate the buffer with.
 * ``pbuf'' is the buffer to add.
 */
an_error_t an_inchan_addpbuf(an_chan_t *chan, an_pbuf_t *pbuf);

/*
 * Remove a packet buffer from a channel's pool.
 * Remove and return a pbuf of the type indicated by flags from the
 * channel's available pool.  Fills in the provided pointer with the
 * returned pbuf or zero if there are none.  Returns zero on success
 * or an error code otherwise.
 *
 * ``chan'' is the channel to associate the buffer with.
 * ``pbufp'' is a pointer in which the buffer object is returned.
 */
an_error_t an_inchan_rempbuf(an_chan_t *chan, int flags, an_pbuf_t **pbufp);

/*
 * Return statistics for a channel.
 * Return packets statistics for the channel in the provided buffer.
 * Returns zero on success or an error code otherwise.
 *
 * ``chan'' is the channel whose stats are to be returned.
 * ``stats'' is a pointer to the buffer to return the stats in.
 */
an_error_t an_chan_stats(an_chan_t *chan, an_chanstats_t *stats);

#define an_inchan_stats an_chan_stats
#define an_outchan_stats an_chan_stats
#define an_cutchan_stats an_chan_stats

#endif /* _NODEOS_CHANNEL_H_ */
