/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * ani_ilock_*
 *
 * Internal locking, just use pthreads sans errors.
 *
 * All the ani_ilock and ani_icond functions are inlined.
 */
#ifndef _NODEOS_IMPL_ANI_ILOCK_H_
#define _NODEOS_IMPL_ANI_ILOCK_H_

#include <pthread.h>
#include <errno.h>
#include <assert.h>

#include <nodeos/impl/decls.h>

/* types */

typedef pthread_mutex_t ani_ilock_t;
typedef pthread_cond_t ani_icond_t;


/* function prototypes */

static inline void 
ani_ilock_init(ani_ilock_t *l) NODEOS_FUNC_UNUSED;

static inline void 
ani_ilock_deinit(ani_ilock_t *l) NODEOS_FUNC_UNUSED;

static inline void 
ani_ilock_lock(ani_ilock_t *l) NODEOS_FUNC_UNUSED;

static inline int
ani_ilock_trylock(ani_ilock_t *l) NODEOS_FUNC_UNUSED;

static inline void 
ani_ilock_unlock(ani_ilock_t *l) NODEOS_FUNC_UNUSED;

static inline int
ani_ilock_islocked(ani_ilock_t *l) NODEOS_FUNC_UNUSED;

static inline void 
ani_icond_init(ani_icond_t *c) NODEOS_FUNC_UNUSED;

static inline void 
ani_icond_deinit(ani_icond_t *c) NODEOS_FUNC_UNUSED;

static inline void 
ani_icond_signal(ani_icond_t *c) NODEOS_FUNC_UNUSED;

static inline void 
ani_icond_broadcast(ani_icond_t *c) NODEOS_FUNC_UNUSED;

static inline void 
ani_icond_wait(ani_icond_t *c, ani_ilock_t *l) NODEOS_FUNC_UNUSED;

static inline void
ani_icond_wait_withhandler(ani_icond_t *c, ani_ilock_t *l,
			   void (*handler)(void *), void *harg) NODEOS_FUNC_UNUSED;

static inline void 
ani_icond_timedwait(ani_icond_t *c, ani_ilock_t *l,
		    int timeoutsec, long timeoutnsec) NODEOS_FUNC_UNUSED;


/* inline functions */

static inline void
ani_ilock_init(ani_ilock_t *l) 
{
	int err;
	assert(l);
	err = pthread_mutex_init(l, 0);
	assert(err == 0);
}

static inline void
ani_ilock_deinit(ani_ilock_t *l) 
{
	int err;
	assert(l);
	err = pthread_mutex_destroy(l);
	assert(err == 0);
}

static inline void
ani_ilock_lock(ani_ilock_t *l) 
{
	int err;
	assert(l);
	err = pthread_mutex_lock((l));
	assert(err == 0);
}

static inline int
ani_ilock_trylock(ani_ilock_t *l) 
{
	assert(l);
	return pthread_mutex_trylock(l);
}


static inline void
ani_ilock_unlock(ani_ilock_t *l) 
{
	int err;
	assert(l);
	err = pthread_mutex_unlock(l);
	assert(err == 0);
}

static inline int
ani_ilock_islocked(ani_ilock_t *l)
{
	assert(l);
	return (pthread_mutex_trylock((l)) == 0 ? (pthread_mutex_unlock((l)), 0) : 1);
}

static inline void
ani_icond_init(ani_icond_t *c) 
{
	int err;
	assert(c);
	err = pthread_cond_init(c, 0);
	assert(err == 0);
}

static inline void
ani_icond_deinit(ani_icond_t *c) 
{
	int err;
	assert(c);
	err = pthread_cond_destroy(c);
	assert(err == 0);
}

static inline void
ani_icond_wait(ani_icond_t *c, ani_ilock_t *l)
{
	int err;

 	pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock, l);
	err = pthread_cond_wait(c, l);
	pthread_cleanup_pop(0);
	assert(err == 0);
}

static inline void
ani_icond_wait_withhandler(ani_icond_t *c, ani_ilock_t *l,
			   void (*handler)(void *), void *harg)
{
	int err;

	assert(handler != 0);
 	pthread_cleanup_push(handler, harg);
	err = pthread_cond_wait(c, l);
	pthread_cleanup_pop(0);
	assert(err == 0);
}

static inline void 
ani_icond_timedwait(ani_icond_t *c, ani_ilock_t *l,
		    int timeoutsec, long timeoutnsec)
{
	int err;
#ifdef OSKIT
	oskit_timespec_t ts;
#else
	struct timespec ts;
#endif

	/* Convert timespec */
	ts.tv_sec = timeoutsec;
	ts.tv_nsec = timeoutnsec;

 	pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock, l);
	err = pthread_cond_timedwait(c, l, &ts);
	pthread_cleanup_pop(0);
	/* XXX does not return timedout notification */
	assert(err == 0 || err == (int)ETIMEDOUT);
}

static inline void
ani_icond_signal(ani_icond_t *c)
{
	int err;
	assert(c);
	err = pthread_cond_signal((c));
	assert(err == 0);
}

static inline void
ani_icond_broadcast(ani_icond_t *c)
{
	int err;
	assert(c);
	err = pthread_cond_broadcast((c));
	assert(err == 0);
}

#endif /* _NODEOS_IMPL_ILOCK_H_ */
