/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _NODEOS_IMPL_ANI_PBUFFER_H_
#define _NODEOS_IMPL_ANI_PBUFFER_H_

#include <assert.h>
#include <sys/types.h>
#include <nodeos/impl/decls.h>
#ifdef OSKIT_SKBUFIOS
#include <oskit/io/skbufio.h>
#endif

/*
 * A packet buffer is a contiguous range of memory described by a pointer
 * (bufdata) and size (bufsize).  Within the buffer is a "valid" region
 * likewise described by a pointer (data) and size (size).  The valid region
 * bounds the valid data in the buffer.  It may be explicitly expanded or
 * contracted at either end up to the limits of the buffer itself.
 */

struct ani_pbuf;
struct ani_chan;
struct ani_flow;

typedef void (*ani_pbuf_free_t)(struct ani_pbuf *pbuf, void *arg);

typedef struct ani_pbuf {
#ifdef OSKIT_SKBUFIOS
	oskit_skbufio_t *skbufio;/* skbufio object */
	struct sk_buff *skb;	/* sk_buff struct */
#else
	size_t off;		/* start offset of valid region */
	size_t size;		/* size of valid region */
	void *bufdata;		/* pointer to buffer memory */
	size_t bufsize;		/* size of buffer */
#endif
	int refcount;		/* reference count */
	ani_pbuf_free_t ffunc;	/* free function, called when refcount==0 */
	void *farg;		/* free function argument */

	struct ani_flow *flow;	/* associated flow */
	struct ani_chan *chan;	/* associated channel */

	struct ani_pbuf *next;	/* handy for linkage */
} ani_pbuf_t;

/* field accessors */
#ifdef OSKIT_SKBUFIOS
#define ani_pbuf_offset(pb)	((pb)->skb->data - (pb)->skb->head)
#define ani_pbuf_data(pb)	((void *)(pb)->skb->data)
#define ani_pbuf_size(pb)	((pb)->skb->len)
#define ani_pbuf_bufdata(pb)	((pb)->skb->head)
#define ani_pbuf_bufsize(pb)	((pb)->skb->truesize)
#define ani_pbuf_isvalid(pb)	((pb)->skbufio != 0) 
#else
#define ani_pbuf_offset(pb)	((pb)->off)
#define ani_pbuf_data(pb)	((void *)((int)(pb)->bufdata + (pb)->off))
#define ani_pbuf_size(pb)	((pb)->size)
#define ani_pbuf_bufdata(pb)	((pb)->bufdata)
#define ani_pbuf_bufsize(pb)	((pb)->bufsize)
#define ani_pbuf_isvalid(pb)	((pb)->bufsize != 0)
#endif
 
/* reference counting */
#define ani_pbuf_reference(pb)	((pb)->refcount++)
#define ani_pbuf_release(pb)	\
	if ((--(pb)->refcount) == 0) (pb)->ffunc((pb), (pb)->farg)

/* inline functions */

NODEOS_FUNC_INLINE void
ani_pbuf_enqueue(ani_pbuf_t **queue, ani_pbuf_t *pbuf)
{
	assert(queue != 0);
	assert(pbuf != 0);
	pbuf->next = *queue;
	*queue = pbuf;
}

NODEOS_FUNC_INLINE ani_pbuf_t *
ani_pbuf_dequeue(ani_pbuf_t **queue)
{
	ani_pbuf_t *pbuf;

	assert(queue != 0);
	pbuf = *queue;
	if (pbuf != 0) {
		*queue = pbuf->next;
		pbuf->next = 0;
	}
	return pbuf;
}

NODEOS_FUNC_INLINE void
ani_pbuf_append(ani_pbuf_t **queue, ani_pbuf_t *list)
{
	assert(queue != 0);
	while (*queue != 0)
		queue = &(*queue)->next;
	*queue = list;
}

ani_error_t ani_pbuf_init(ani_pbuf_t *pbuf, struct ani_flow *flow,
			  struct ani_chan *chan, size_t bsize, size_t off,
			  size_t size);
void ani_pbuf_init_external(ani_pbuf_t *pbuf, struct ani_flow *flow,
			    struct ani_chan *chan,
#ifdef OSKIT_SKBUFIOS
			    struct sk_buff *skb
#else
			    void *bdata, size_t bsize, size_t off, size_t size,
			    ani_pbuf_free_t ffunc, void *farg
#endif
			    );
void ani_pbuf_deinit(ani_pbuf_t *pbuf);
ani_error_t ani_pbuf_clone(ani_pbuf_t *pbuf, ani_pbuf_t *newpbuf);
ani_error_t ani_pbuf_setoffset(ani_pbuf_t *pbuf, size_t newoff,
			       void **out_data);
ani_error_t ani_pbuf_setsize(ani_pbuf_t *pbuf, size_t newsize,
			     void **out_data);
void *ani_pbuf_setmax(ani_pbuf_t *pbuf);
void ani_pbuf_swap(ani_pbuf_t *pbuf1, ani_pbuf_t *pbuf2);

#ifdef OSKIT
struct oskit_netio;
struct oskit_bufio;

ani_error_t ani_pbuf_frombufio(struct ani_flow *flow, struct ani_chan *chan,
			       struct oskit_bufio *bufio, size_t off,
			       size_t size, ani_pbuf_t **out_pbuf);
ani_error_t ani_pbuf_tobufio(ani_pbuf_t *pbuf, void *hdr, size_t hsize,
			     struct oskit_netio *nio,
			     struct oskit_bufio **out_bufio, int *didcopy);
#else
ani_error_t ani_pbuf_tomem(ani_pbuf_t *pbuf, void *hdr, size_t hsize,
			   void **out_buf, size_t *out_size, int *didcopy);
void ani_pbuf_relmem(void *buf, size_t bufsize);
#endif

#endif /* _NODEOS_IMPL_ANI_PBUFFER_H_ */
