/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _NODEOS_IMPL_ANI_THREAD_H_
#define _NODEOS_IMPL_ANI_THREAD_H_

#include <pthread.h>
#include <sched.h>
#include <setjmp.h>
#include <nodeos/an_impl.h>
#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_cond.h>

/*
 * Stuff That Doesn't Belong
 * XXX hack around pthread deficiencies:
 *	OSKit doesn't support attribute scope stuff currently,
 *		the default (only!) scope is system.
 *	FreeBSD 3.0 doesn't support cancelation,
 *		the EE had really better cooperate!
 * XXX hack around missing functions in the OSKit/ARM C library:
 *	No sig{set,long}jmp functions, use _{set,long}jmp instead.
 */
#ifndef PTHREAD_CANCEL_ASYNCHRONOUS
#define PTHREAD_CANCEL_ASYNCHRONOUS	0
#define PTHREAD_CANCEL_DEFERRED		0
#define pthread_setcanceltype(t, sp)	(*(sp) = 0)
#define pthread_cancel(t)		(0)
#define pthread_testcancel()		((void)0)
#undef PTHREAD_SCOPE_SYSTEM /* XXX freebsd defines this but doesn't impl */
#endif

#ifndef PTHREAD_SCOPE_SYSTEM
#define pthread_attr_setscope(a,s) 	(0)
#endif

#ifdef OSKIT_ARM32_SHARK
#define sigsetjmp(b, m)		_setjmp(b)
#define siglongjmp(b, v)	_longjmp(b, v)
#endif

struct ani_threadpool;
struct ani_threadwork;
struct ani_chan;
struct an_cpuspec;

typedef struct ani_thread {
	struct ani_thread *next;	/* threadpool link */
	struct ani_thread *freenext;	/* threadpool freelist link */
	struct ani_threadpool *pool;	/* our threadpool */
	struct ani_chan *chan;	 /* channel we are processing for */
	pthread_t id;		 /* pthread ID */
	pthread_attr_t attr;	 /* pthread attributes (stackptr, stacksize) */
	struct sched_param pprio;/* pthread scheduling priority */
	int nprio;		 /* NodeOS scheduling priority */
	ani_icond_t waiting;	 /* place for this thread to wait */
	ani_icond_t waitedon;	 /* place for others to wait for this thread */
	struct ani_threadwork *work;	/* work targetted to this thread */
	an_threadintr_func_t handler;	/* interrupt/trap handler */
	ani_ilock_t lock;	 /* lock to protect interrupt state */
	ani_cond_t *cond;	 /* non-zero if in ani_cond_wait */
	int flags;		 /* various flags */
	sigjmp_buf jmpbuf;	 /* warp the thread back to its origin */
} ani_thread_t;

#define THREAD_INKERNEL	0x01	/* thread is in the NodeOS */
#define THREAD_WAITEDON	0x02	/* someone is waiting for thread */
#define THREAD_DONE	0x04	/* waiter is finished with the thread */
#define THREAD_INTSON	0x08	/* interrupts should be allowed */
#define THREAD_ISFREE	0x10	/* thread is on the pool free list */
#define THREAD_ONESHOT	0x20	/* thread runs a specific function and dies */
#define THREAD_BEDEAD	0x80	/* thread is not attached to a threadpool */

#define ani_thread_lock(t)	ani_ilock_lock(&(t)->lock)
#define ani_thread_unlock(t)	ani_ilock_unlock(&(t)->lock)

struct an_cpuspec;

#define _assert_ani_thread_valid(t) \
	assert((t)->pool != NULL)

#ifdef MEMDEBUG
#define assert_ani_thread_valid(t) \
	assert((t) != 0 && (t) != (void *)0xdddddddd), \
	assert(*(unsigned int *)(t) != 0xdddddddd), \
	_assert_ani_thread_valid(t)
#else
#define assert_ani_thread_valid(t) \
	assert((t) != NULL), \
	_assert_ani_thread_valid(t)
#endif

/* inline functions */

NODEOS_FUNC_INLINE ani_thread_t *
ani_thread_current(void)
{
	extern pthread_key_t ani_nodeos_thread;
	ani_thread_t *thread;

	/*
	 * In the current implementation has no control over scheduling
	 * of threads, so we have to query the pthread state.
	 */
	thread = (ani_thread_t *)pthread_getspecific(ani_nodeos_thread);
	assert(thread == 0 || thread->id == pthread_self());
	return thread;
}

/* prototypes */

void ani_thread_setup(void);
ani_error_t ani_thread_init(ani_thread_t *thread,
			    void *stackptr, size_t stacksize);
ani_error_t ani_thread_deinit(ani_thread_t *thread);
ani_thread_t *ani_thread_current(void);
ani_error_t ani_thread_spawn(struct ani_threadpool *tpool,
			     ani_thread_t **out_thread);
ani_error_t ani_thread_start(ani_thread_t *thread,
			     struct ani_threadpool *tpool);
void ani_thread_kill(ani_thread_t *thread, int dowait);
void ani_thread_kill_locked(ani_thread_t *thread, int dowait);
void ani_thread_exit(void);
void ani_thread_yield(void);
void ani_thread_sleep(an_timespec_t *delay);
void *ani_thread_getdata(void);
void ani_thread_setdata(void *);
void ani_thread_interrupt(ani_thread_t *thread);
void ani_thread_block_interrupt(void);
void ani_thread_unblock_interrupt(void);
void ani_thread_sethandler(an_threadintr_func_t handler);
void ani_thread_condtimedwait(ani_cond_t *cond, an_timespec_t* abstime);
int ani_thread_getprio(ani_thread_t *thread);
ani_error_t ani_thread_setprio(ani_thread_t *thread, int prio);
ani_error_t ani_thread_stackinfo(ani_thread_t *thread,
				 void **base, size_t *size, void **ptr);

#endif /* _NODEOS_IMPL_THREAD_H_ */
