/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _ANI_THREADPOOL_H_
#define _ANI_THREADPOOL_H_

#define ANI_THREADPOOL_CTHREADS	4	/* number of cached threads */
#undef ANI_THREADPOOL_STATS

#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_resource.h>
#include <nodeos/impl/ani_thread.h>

/*
 * The threadpool: a group of threads sharing the same scheduling params
 */

struct ani_flow;
struct ani_chan;

typedef void (*ani_thread_func_t)(void *, void *);

struct ani_threadwork {
	struct ani_threadwork *next;	/* next item on list */
	int inkern;		/* function should run in kernel context */
	struct ani_chan *chan;	/* channel to associate thread with */
	ani_thread_func_t func;	/* function the thread should start in */
	void *farg[2];		/* and arguments for that function */
};

struct ani_threadpool {
	struct ani_flow *flow;		/* associated flow */
	int flags;			/* various flags */
	int curthreads;			/* current # of threads in pool */
	ani_thread_t *threads;		/* list of current threads */
	int reserve;			/* desired number of free threads */
	int freethreads;		/* current # of free threads */
	ani_thread_t *freelist;		/* list of free threads */
	int curosthreads;		/* number of one-shot threads */
	ani_thread_t *osthreads;	/* list of one-shot threads */
	ani_icond_t waiting;		/* place to wait for this pool */
	int worklen;			/* length of work queue */
	struct ani_threadwork *workhead;/* head of work queue */
	struct ani_threadwork *worktail;/* tail of work queue */
	an_cpuspec_t params;		/* resource parameters */
	int refs;			/* reference count */
	ani_ilock_t lock;		/* lock for entire pool */
#ifdef DEBUG /* XXX ifdef STATS */
	struct {
		int starts;
		int creates;
		int queued;
		int failed;
		int maxactive;
		int maxcached;
		int maxqueued;
	} stats;
#endif
};
typedef struct ani_threadpool ani_threadpool_t;

#define TPOOL_DYING	1
#define TPOOL_WAITING	2

#define ani_threadpool_lock(t)		ani_ilock_lock(&(t)->lock)
#define ani_threadpool_unlock(t)	ani_ilock_unlock(&(t)->lock)
#define ani_threadpool_islocked(t)	ani_ilock_islocked(&(t)->lock)

#define ani_threadpool_reference(t)	((t)->refs++)
#define ani_threadpool_release(t)	\
	if (--(t)->refs == 0) ani_threadpool_cleanup(t)

ani_error_t ani_threadpool_init(ani_threadpool_t *tpool, struct ani_flow *flow,
				an_cpuspec_t *cspec);
void ani_threadpool_deinit(ani_threadpool_t *tpool, int waitfor);
void ani_threadpool_disable(ani_threadpool_t *tpool, int waitfor);
void ani_threadpool_cleanup(ani_threadpool_t *tpool);

#ifdef MEMDEBUG
#define assert_ani_threadpool_valid(t) \
	assert((t) != 0 && (t) != (void *)0xdddddddd), \
	assert(*(unsigned int *)(t) != 0xdddddddd)
#else
#define assert_ani_threadpool_valid(t)	assert((t) != 0)
#endif

/*
 * Wait for the indicated thread to terminate or until the given time has
 * elapsed.  Returns zero on success, an error code otherwise.
 */
ani_error_t ani_threadpool_thread_wait(ani_threadpool_t *tpool,
				       ani_thread_t *thread,
				       an_timespec_t *timo);

/*
 * Schedule the function to be run in a thread from the indicated pool.
 * If inkern is non-zero, the thread will be run in the NodeOS context.
 * If no threads are currently available or queueonly is non-zero, the
 * request will be queued.  Returns zero on success, an error code otherwise.
 */
ani_error_t ani_threadpool_thread_schedule(ani_threadpool_t *tpool,
					   struct ani_chan *chan,
					   int inkern, int queueonly,
					   ani_thread_func_t func,
					   void *farg1, void *farg2);

/*
 * Start a thread from the given pool running the indicated function.
 * If inkern is non-zero, the thread will be run in the NodeOS context.
 * If there are no threads available and the pool is at the thread limit,
 * returns an error.  If the pool is under limit but there are no cached
 * threads available, the call is delayed briefly while a new thread is started.
 * That is, upon return from this call, an actual thread has been commited
 * to run the function.  If out_threadp is non-zero, the call returns a
 * pointer to the new thread.  Returns zero if a thread was successfully
 * started, an error otherwise.
 */
ani_error_t ani_threadpool_thread_start(ani_threadpool_t *pool, int inkern,
					ani_thread_func_t func, void *farg1,
					void *farg2, ani_thread_t **out_thread);

/*
 * Start a thread from the given pool and wait til it terminates or the
 * timeout expires.  If the timeout expires, the thread is killed and a
 * distinguished error code is returned.  If inkern is non-zero, the
 * thread will be run in the NodeOS context.  Returns zero on success,
 * an error code otherwise.
 */
ani_error_t ani_threadpool_thread_run(ani_threadpool_t *tpool, int inkern,
				      ani_thread_func_t func, void *farg1,
				      void *farg2, an_timespec_t *timo);

/*
 * Start the given thread as a ``one-shot'' thread in the context of the
 * given pool running the indicated function.  Returns zero if a thread was
 * successfully started, an error otherwise.
 */
ani_error_t ani_threadpool_osthread_start(ani_threadpool_t *tpool,
					  ani_thread_t *thread, int inkern,
					  ani_thread_func_t func,
					  void *farg1, void *farg2);

void ani_threadpool_thread_exit(ani_threadpool_t *tpool);

ani_error_t ani_threadpool_thread_setreserve(ani_threadpool_t *tpool,
					     int nthreads, int wait);

#endif /* _ANI_THREADPOOL_H_ */
