/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Channel API
 */
#include <nodeos/an_error.h>
#include <nodeos/an_chan.h>
#include <nodeos/an_pbuf.h>
#include <nodeos/an_demuxkey.h>

#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_mempool.h>
#include "os.h"

/*
 * @allocation (optional): sizeof(an_flow_t) bytes for channel
 * @allocation (temporary): sizeof(ani_pbuf_t *) * npbufs bytes for pbuf ptrs
 */
an_chan_t *
an_inchan_create(void *mem, an_flow_t *flow, an_demuxkey_t *key,
		 an_chanproto_t *proto, an_chanaddr_t *addr,
		 an_pbuf_t *pbufs[], int npbufs, an_cpuspec_t *cpu,
		 an_chanrecvfunc_t func, void *farg)
{
	ani_chan_t *ichan;
	ani_pbuf_t **ipbufs = 0;
	ani_flow_t *iflow;
	ani_demuxkey_t *ikey = NULL;
	ani_mempool_t *cpool;
	ani_error_t err;
	int didalloc = 0;

	NODEOS_ENTER;

	/*
	 * Allocate memory for the object if needed.
	 */
	cpool = ani_flow_mempool(ani_flow_current());
	if (mem == 0) {
		mem = ani_mempool_alloc(cpool, sizeof(an_chan_t));
		didalloc = 1;
	}

	/*
	 * Validate arguments.
	 * N.B.: no copyin of proto/addr/cpu specs as copyin/out are not
	 * needed in our implementation.
	 */
	ichan = NODEOS_INTERN_CHAN(mem);
	iflow = NODEOS_INTERN_FLOW(flow);
	if (ichan == 0 || iflow == 0 || proto == 0 || addr == 0 || cpu == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "some argument was null");
		goto done;
	}
	if (key != NULL)
		ikey = NODEOS_INTERN_DEMUXKEY(key);
	if (pbufs != 0 && npbufs != 0) {
		int i;

		ipbufs = ani_mempool_alloc(cpool, sizeof *ipbufs * npbufs);
		if (ipbufs == 0) {
			err = ani_error_set(AN_ERR_NOMEMORY,
					    "could not allocate pbufs");
			goto done;
		}
		for (i = 0; i < npbufs; i++) {
			ipbufs[i] = NODEOS_INTERN_PBUF(pbufs[i]);
			if (ipbufs[i] == 0) {
				err = ani_error_set(AN_ERR_INVALIDARG,
						    "could not intern provided pbuf[%d]", i);
				goto done;
			}
		}
	}

	/*
	 * Create and enable the channel.
	 */
	err = ani_chan_init(ichan, CHAN_IN, iflow, ani_flow_threadpool(iflow),
			    proto, addr, ikey, cpu, ipbufs, npbufs, 0,
			    func, farg);
	if (!ani_error_occured(err) && func != 0) {
		err = ani_chan_enable(ichan);
		if (ani_error_occured(err))
			ani_chan_deinit(ichan);
	}

 done:
	if (ani_error_occured(err) && didalloc)
		ani_mempool_free(cpool, mem, sizeof(an_chan_t));
	if (ipbufs)
		ani_mempool_free(cpool, ipbufs, sizeof *ipbufs * npbufs);

	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_CHAN(ichan);
}

an_chan_t *
an_outchan_create(void *mem, an_flow_t *flow,
		  an_chanproto_t *proto, an_chanaddr_t *addr, an_netspec_t *net)
{
	ani_chan_t *ichan;
	ani_flow_t *iflow;
	ani_mempool_t *cpool;
	ani_error_t err;
	int didalloc = 0;

	NODEOS_ENTER;

	/*
	 * Allocate memory for the object if needed.
	 */
	cpool = ani_flow_mempool(ani_flow_current());
	if (mem == 0) {
		mem = ani_mempool_alloc(cpool, sizeof(an_chan_t));
		didalloc = 1;
	}

	/*
	 * Validate arguments.
	 */
	ichan = NODEOS_INTERN_CHAN(mem);
	iflow = NODEOS_INTERN_FLOW(flow);
	if (ichan == 0 || iflow == 0 || proto == 0 || addr == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "one of the parameters is null");
		goto done;
	}

	/*
	 * Create the channel.
	 */
	err = ani_chan_init(ichan, CHAN_OUT, iflow, ani_flow_threadpool(iflow),
			    proto, addr, 0, 0, 0, 0, net, 0, 0);

 done:
	if (ani_error_occured(err) && didalloc)
		ani_mempool_free(cpool, mem, sizeof(an_chan_t));

	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_CHAN(ichan);
}

an_chan_t *
an_cutchan_create(void *mem, an_flow_t *flow, an_chanproto_t *proto,
		  an_chan_t *inchan, an_chan_t *outchan)
{
	ani_flow_t *iflow;
	ani_chan_t *ichan, *iinchan, *ioutchan;
	ani_mempool_t *cpool;
	ani_error_t err;
	int didalloc = 0;

	NODEOS_ENTER;

	/*
	 * Allocate memory for the object if needed.
	 */
	cpool = ani_flow_mempool(ani_flow_current());
	if (mem == 0) {
		mem = ani_mempool_alloc(cpool, sizeof(an_chan_t));
		didalloc = 1;
	}

	/*
	 * Validate arguments.
	 */
	ichan = NODEOS_INTERN_CHAN(mem);
	iflow = NODEOS_INTERN_FLOW(flow);
	iinchan = NODEOS_INTERN_CHAN(inchan);
	ioutchan = NODEOS_INTERN_CHAN(outchan);
	if (ichan == 0 || iflow == 0 || iinchan == 0 || ioutchan == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "one of the parameters is null");
		goto done;
	}

	/*
	 * Create the channel.
	 */
	err = ani_cutchan_init(ichan, iflow, ani_flow_threadpool(iflow), 0,
			       proto, iinchan, ioutchan,
			       ani_cutchan_receive, ichan);

	/*
	 * Enable the inchan
	 */
	if (!ani_error_occured(err)) {
		err = ani_chan_enable(iinchan);
		if (ani_error_occured(err))
			ani_chan_deinit(ichan);
	}

 done:
	if (ani_error_occured(err) && didalloc)
		ani_mempool_free(cpool, mem, sizeof(an_chan_t));

	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_CHAN(ichan);

}

an_error_t
an_chan_destroy(an_chan_t *chan)
{
	ani_error_t err;

	NODEOS_ENTER;
	err = ani_chan_deinit(NODEOS_INTERN_CHAN(chan));
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_outchan_send(an_chan_t *chan, an_pbuf_t *pbuf)
{
	ani_error_t err;

	NODEOS_ENTER;
	err = ani_chan_send(NODEOS_INTERN_CHAN(chan), NODEOS_INTERN_PBUF(pbuf));
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_outchan_sendv(an_chan_t *chan, an_pbuf_t *pbufs[], int npbufs)
{
	panic("ani_chan_sendv called");
	return AN_ERR_INVALIDARG;
}

an_error_t
an_inchan_addpbuf(an_chan_t *chan, an_pbuf_t *pbuf)
{
	ani_chan_t *ichan;
	ani_pbuf_t *ipbuf[1];
	ani_error_t err;

	NODEOS_ENTER;

	ichan = NODEOS_INTERN_CHAN(chan);
	ipbuf[0] = NODEOS_INTERN_PBUF(pbuf);
	if (ichan == 0 || ipbuf[0] == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "null argument passed");
	} else {
		ani_chan_pbuf_add(ichan, ipbuf, 1);
		err = ani_error_clear();
	}

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_inchan_rempbuf(an_chan_t *chan, int flags, an_pbuf_t **pbufp)
{
	panic("ani_chan_rempbuf called");
	return AN_ERR_INVALIDARG;
}

an_error_t
an_chan_stats(an_chan_t *chan, an_chanstats_t *stats)
{
	ani_chan_t *ichan;
	an_chanstats_t *istats;
	ani_error_t err;

	NODEOS_ENTER;

	ichan = NODEOS_INTERN_CHAN(chan);
	if (ichan == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "chan parameter is null");
	else {
		istats = ani_chan_stats(ichan);
		NODEOS_COPYOUT(stats, istats, sizeof *istats, &err);
	}

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}
