/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Flow API
 */
#include <nodeos/an_flow.h>
#include <nodeos/an_cred.h>
#include <nodeos/an_thread.h>

#include <nodeos/impl/ani_cred.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_nodeos.h>

an_flow_t *
an_flow_create(void *mem, an_cred_t *cred,
	       an_cpuspec_t *cspec, an_memspec_t *mspec,
	       an_flowinit_func_t init, void *initarg, an_flowterm_func_t term)
{
	ani_cred_t *icred;
	ani_flow_t *iflow;
	ani_mempool_t *cpool;
	int didalloc = 0;
	ani_error_t err;

	NODEOS_ENTER;

	/*
	 * Allocate memory for the object if needed.
	 */
	cpool = ani_flow_mempool(ani_flow_current());
	if (mem == 0) {
		mem = ani_mempool_alloc(cpool, sizeof(an_flow_t));
		didalloc = 1;
	}

	iflow = NODEOS_INTERN_FLOW(mem);
	icred = NODEOS_INTERN_CRED(cred);
	if (iflow == 0 || icred == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid flow or cred argument");
	else
		err = ani_flow_init(iflow, icred, cspec, mspec,
				    init, initarg, term);

	if (ani_error_occured(err) && didalloc)
		ani_mempool_free(cpool, mem, sizeof(an_flow_t));

	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_FLOW(iflow);
}

an_error_t
an_flow_destroy(an_flow_t *flow)
{
	ani_flow_t *iflow;
	ani_error_t err;

	NODEOS_ENTER;

	iflow = NODEOS_INTERN_FLOW(flow);
	
	if (iflow == NULL || iflow == ani_flow_root_flow())
		err = ani_error_set(AN_ERR_INVALIDARG, "invalid flow argument");
	else 
		err = ani_flow_deinit(iflow);

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_flow_startthread(an_flow_t *flow, void (*func)(void *arg), void *arg,
		    an_thread_t **out_thread)
{
	ani_flow_t *iflow;
	ani_thread_t *ith;
	ani_error_t err;

	NODEOS_ENTER;

	iflow = NODEOS_INTERN_FLOW(flow);
	if (iflow == 0 || func == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid flow or func argument");
	else {
		err = ani_threadpool_thread_start(&iflow->tpool, 0,
						  (ani_thread_func_t)func,
						  arg, 0, &ith);
		if (!ani_error_occured(err) && out_thread)
			*out_thread = NODEOS_EXTERN_THREAD(ith);
	}

	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

