/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Information API
 */
#include <nodeos/an_info.h>
#include <nodeos/impl/ani_if.h>

#include "options.h"
#include "os.h"

/*
 * Retrieve information about network interfaces.
 *
 * ``info'' is an array of an_ifinfo structs to be filled in
 * ``count'' the size of the array on call, the number filled in on return
 *
 * Returns 0 on success, an error otherwise.
 */
an_error_t
an_info_getifs(an_infoif_t info[], int *count)
{
	ani_error_t err;
	int icount, num;
	an_infoif_t *iinfo;

	NODEOS_ENTER;

	/*
	 * Validate arguments.
	 */
	if (count == 0) {
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid count argument");
		goto done; 
	}

	/*
	 * Read desired size.
	 */
	NODEOS_COPYIN(&icount, count, sizeof icount, &err);
	if (ani_error_occured(err))
		goto done;

	/*
	 * Get a pointer to the info
	 */
	num = icount;
	iinfo = ani_if_getinfo(&num);

	/*
	 * Copy out data
	 */
	NODEOS_COPYOUT(count, &num, sizeof *count, &err);
	if (!ani_error_occured(err) && iinfo != 0)
		NODEOS_COPYOUT(info, iinfo, sizeof *iinfo * num, &err);

done:
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

static char options_str[256];

const char*
an_info_getoptlist(void)
{
	LW_NODEOS_ENTER;
	if (options_str[0] == 0) {
#ifdef HAVE_NET
		strcat(options_str, "HAVE_NET ");
#endif
#ifdef HAVE_DISK
		strcat(options_str, "HAVE_DISK ");
#endif
#ifdef CHAN_CLOCK
		strcat(options_str, "CHAN_CLOCK ");
#endif
#ifdef CHAN_ETH
		strcat(options_str, "CHAN_ETH ");
#endif
#ifdef CHAN_SOCKET
		strcat(options_str, "CHAN_SOCKET ");
#endif
#ifdef CHAN_IF
		strcat(options_str, "CHAN_IF ");
#endif
	}
	LW_NODEOS_LEAVE;

	return options_str;
}


/*
 * Return the hostname of this node.  Returns a read-only constant.
 */
const char *
an_info_gethostname(void)
{
	const char *name;

	LW_NODEOS_ENTER;
	name = ani_os_gethostname();
	LW_NODEOS_LEAVE;

	/* NB: don't need to copyout the name. */

	return name;
}


/*
 * Set the hostname of this node.  The nodeos will
 * make a copy of the argument.
 */
an_error_t
an_info_sethostname(const char *name)
{
	ani_error_t err;
	int ierr;

	/* NB: don't need to copyin the name. */
	
	NODEOS_ENTER;
	ierr = ani_os_sethostname(name);
	if (ierr != 0)
		err = ani_error_set(ierr, "os_sethostname failed");
	else
		err = ani_error_clear();
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}
