/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Thread API
 */
#include <nodeos/an_error.h>
#include <nodeos/an_thread.h>

#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_threadpool.h>

an_thread_t *
an_thread_create(void *mem, void *stackptr, size_t stacksize)
{
	ani_thread_t *ith;
	ani_error_t err;

	NODEOS_ENTER;

	ith = NODEOS_INTERN_THREAD(mem);
	if (ith == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid thread argument");
	else
		err = ani_thread_init(ith, stackptr, stacksize);

	NODEOS_LEAVE;
	return ani_error_occured(err) ? 0 : NODEOS_EXTERN_THREAD(ith);
}

int
an_thread_destroy(an_thread_t *thread)
{
	ani_error_t err;

	NODEOS_ENTER;
	err = ani_thread_deinit(NODEOS_INTERN_THREAD(thread));
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_thread_start(an_thread_t *thread, an_flow_t *flow,
		void (*func)(void *arg), void *arg)
{
	ani_thread_t *ithread;
	ani_flow_t *iflow;
	ani_error_t err;

	NODEOS_ENTER;

	ithread = NODEOS_INTERN_THREAD(thread);
	iflow = NODEOS_INTERN_FLOW(flow);
	if (ithread == 0 || iflow == 0 || func == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid thread/flow/func argument");
	else
		err = ani_threadpool_osthread_start(&iflow->tpool, ithread, 0,
						    (ani_thread_func_t)func,
						    arg, 0);
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

void
an_thread_exit(void)
{
	ani_thread_t *cur;
	
	NODEOS_ENTER;
	cur = ani_thread_current();
	assert_ani_thread_valid(cur);
	assert(cur->pool);
	ani_threadpool_thread_exit(cur->pool);
	NODEOS_LEAVE;
}

void
an_thread_yield(void)
{
	NODEOS_ENTER;
	ani_thread_yield();
	NODEOS_LEAVE;
}

void
an_thread_sleep(an_timespec_t *delay)
{
	NODEOS_ENTER;
	ani_thread_sleep(delay);
	NODEOS_LEAVE;
}

void
an_thread_setdata(void *data)
{
	NODEOS_ENTER;
	ani_thread_setdata(data);
	NODEOS_LEAVE;
}

void *
an_thread_getdata(void)
{
	void *data;

	NODEOS_ENTER;
	data = ani_thread_getdata();
	NODEOS_LEAVE;

	return data;
}

void
an_thread_intr(an_thread_t *thread)
{
	ani_thread_t *ithread;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread != 0)
		ani_thread_interrupt(ithread);
	NODEOS_LEAVE;
}

void
an_thread_blockintr(void)
{
	NODEOS_ENTER;
	ani_thread_block_interrupt();
	NODEOS_LEAVE;
}

void
an_thread_unblockintr(void)
{
	NODEOS_ENTER;
	ani_thread_unblock_interrupt();
	NODEOS_LEAVE;
}

an_error_t
an_thread_setprio(an_thread_t *thread, int prio)
{
	ani_thread_t *ithread;
	ani_error_t err;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid thread argument");
	else
		err = ani_thread_setprio(ithread, prio);
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_thread_getprio(an_thread_t *thread, int *priop)
{
	ani_thread_t *ithread;
	ani_error_t err;
	int iprio;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid thread argument");
	else {
		iprio = ani_thread_getprio(ithread);
		NODEOS_COPYOUT(priop, &iprio, sizeof *priop, &err);
	}
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Implementation specific API calls
 */

/*
 * Establish a thread_interrupt handler.
 */
void
an_thread_sethandler(an_threadintr_func_t handler)
{
	NODEOS_ENTER;
	ani_thread_sethandler(handler);
	NODEOS_LEAVE;
}

/*
 * Get various bits of stack information on a specific
 * thread.  The thread may be the current thread.
 * If the target thread is running (on an SMP machine)
 * the results are undefined.
 *
 * XXX this call may evolve into a more generic thread_info call.
 */
an_error_t 
an_thread_stackinfo(an_thread_t *thread, void **base, size_t *size, void **ptr)
{
	ani_thread_t *ithread;
	void *ibase, *iptr;
	size_t isize;
	ani_error_t err;
	
	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread == 0)
		err = ani_error_set(AN_ERR_INVALIDARG,
				    "invalid thread argument");
	else {
		err = ani_thread_stackinfo(ithread, &ibase, &isize, &iptr);
		if (!ani_error_occured(err))
			NODEOS_COPYOUT(base, &ibase, sizeof ibase, &err);
		if (!ani_error_occured(err))
			NODEOS_COPYOUT(size, &isize, sizeof isize, &err);
		if (!ani_error_occured(err))
			NODEOS_COPYOUT(ptr, &iptr, sizeof iptr, &err);
	}
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}
