/*
 * Copyright (c) 2000 The University of Utah. 
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
/*
 * ani_debug.c
 */

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_error.h>

#include "ani_debug.h"

#if !defined(NDEBUG) && defined(DEBUG)
/* --- Debugging is enabled --- */ 

/* The debugging mask all DBG() statements check. */
ani_debugmask_t ani_debug_mask = 0;

/*
 * Global variables related to the debugging buffer
 *
 * Note: these are not 'static' so that they're easy to find
 * in a debugger.
 */
char *ani_debug_buffer = NULL;
int ani_debug_bufferbegin = 0;
int ani_debug_buffersize = 0;

/*
 * For each debugging option in ani_debug.h, provide a
 * string name, a mask (may be more than one option |'d),
 * and a description string.
 */
static struct {
	const char *name;
	ani_debugmask_t mask;
	const char *desc;
} ani_debug_opttable[] = {
#define D(name, str) { #name, DBG_ ## name, str }
	D(NONE,		"Nothing"),
	D(SOCKFD,	"Show socket-channel related file descriptor operations."),
	D(DEMUX_MATCH_DUMP,"Show demux key usage."),
	D(THREADS,	"Show thread-related operations."),
	D(THREADPOOLS,	"Show threadpool-related operations."),
	
	/* Options handled special-like by ani_debug_setmaskstr() */
	{ "buffer", 0,	"Write all ani_dprintf() output into an internal buffer." },
	{ "dump", 0,	"Dump said internal buffer when moab exits." },
		
	/* These must be last, as they are markers for some GDB macros. */
	D(ALL,		"Everything under the sun...."),
	D(ANY,		"Ibid.")
};
	
#define ANI_DEBUG_DEFAULT_BUFFERSIZE (16 * 1024)
static void ani_debug_bufferinit(int size);
static void ani_debug_bufferprint(void);

#define NELEMS(a)	(sizeof (a) / sizeof(a[0]))

static int ani_debug_initialized = 0;

/*
 * initialize ani_debug system
 */
void
ani_debug_init(void)
{
	if (ani_debug_initialized)
		return;
	ani_debug_initialized = 1;
}

/*
 * Set the debug mask to the given mask
 */
void
ani_debug_setmask(ani_debugmask_t mask)
{
	ani_debug_init();  /* in case its done before moab initialization. */
	assert(ani_debug_initialized);

	ani_debug_mask = mask;
}


/*
 * Set the debug mask as defined by the entries in the given string.
 * Each entry should be a name in the ani_debug_opttable table.  Entries may
 * be separated by "|" or ",".  The special entry 'list' causes all
 * of the debugging entries to be listed.
 * 
 * Returns zero if there was an error or the 'list' argument was passed.
 * The caller should cleanup and shutdown.
 */
ani_error_t
ani_debug_setmaskstr(char *mask_str)
{
	int i;
	const char *separators = "|,";
	char *opt;
	
	ani_debug_init();  /* in case its done before moab initialization. */
	assert(ani_debug_initialized);
	
	opt = strtok(mask_str, separators);

	if (!opt) {
		ani_debug_mask = 0;
		return ani_error_set(EINVAL,
			"No valid options passed to ani_debug_setmaskstr");
	}

	/* Special target 'list' lists all the defined options, and returns 0 */
	if (!strcasecmp(opt, "list")) {
		ani_dprintf("Available debug opts: \n");
		ani_dprintf("  %-15s\t%16s  %s\n",
			    "Option", "Mask", "Description");
		for (i = 0; i < NELEMS(ani_debug_opttable); i++)
		{
			const char* name;
			const char* desc;
			int hi, lo;

			name = ani_debug_opttable[i].name;
			desc = ani_debug_opttable[i].desc;
			hi = (int)(ani_debug_opttable[i].mask>>32);
			lo = (int)(ani_debug_opttable[i].mask);

			if (hi)
				ani_dprintf("  %-15s\t%8X%08X  %s\n",
					    name, hi, lo, desc);
			else
				ani_dprintf("  %-15s\t        %8X  %s\n", 
					    name, lo, desc);
		}
		return ani_error_set(EINVAL,
				"'list' option passed to ani_debug_setmaskstr");
	}
	

	while (opt) {
		/* Handle special-case opts */
		if (!strcasecmp(opt, "buffer"))
			ani_debug_bufferinit(ANI_DEBUG_DEFAULT_BUFFERSIZE);
		else if (!strcasecmp(opt, "dump"))
			atexit(ani_debug_bufferprint);
		/* Handle run-of-the-mill opts */
		else
		{
			for (i = 0; i < NELEMS(ani_debug_opttable); i++)
				if (!strcasecmp(opt, ani_debug_opttable[i].name))
				{
					ani_debug_mask |= ani_debug_opttable[i].mask;
					break;
				}
		
			/* Be polite. */
			if (i == NELEMS(ani_debug_opttable)) 
			{
				return ani_error_set(EINVAL,
						     "Unknown option '%s' passed to "
						     "-ani_debug_setmaskstr()\n",
						     opt);
			}
		}
		
		/* Get next opt */
		opt = strtok(NULL, separators);
	}

	return ani_error_clear();
}

/*
 * Create a buffer into which debugging output will be written.
 * See the 'buffer' option.
 */
static void
ani_debug_bufferinit(int size)
{
	ani_debug_buffersize = size;
	ani_debug_buffer = ani_mempool_alloc(ani_flow_mempool(ani_flow_current()),
					     size);
	assert(ani_debug_buffer);
}


/*
 * The following function is atexit'd if the 'dump' option was
 * given.
 */
void
ani_debug_bufferprint(void)
{
	int i = 0;
	int end = ani_debug_bufferbegin;

	i = ani_debug_bufferbegin;
	assert(i != 0);

	while(i < ani_debug_buffersize)
		putc(ani_debug_buffer[i++], stdout);
	i = 0;
	while(i < end)
		putc(ani_debug_buffer[i++], stdout);
}

void
ani_dprintf(const char *fmt, ...)
{
	int n;
	va_list args;

	va_start(args, fmt);

	if (ani_debug_buffersize != 0)
	{
		int max = ani_debug_buffersize - ani_debug_bufferbegin - 1;
		n = vsnprintf(ani_debug_buffer + ani_debug_bufferbegin, max, fmt, args);

		/* The return value is bytes *needed* not bytes *used* */
		if (n > max)
			n = max;
		ani_debug_bufferbegin += n;
		assert(ani_debug_bufferbegin < ani_debug_buffersize);/* XXX */

		if (ani_debug_bufferbegin >= (ani_debug_buffersize - 60))
			ani_debug_bufferbegin = 0;
	}
	else
	{
		/* Boring. Print to stderr */
		n = vfprintf(stderr, fmt, args);
		fflush(stderr);
	}

	va_end(args);

	return;
}
#endif /* Debugging is enabled */

