/*
 * Copyright (c) 2000 The University of Utah.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
/*
 * ani_debug.h
 *
 * A dynamic debugging framework for Moab.  Through the magic
 * of hideous macros, we can control the internal debugging output of
 * Moab in a couple of ways.  First, it can be completely
 * disabled in which case all of the code "just goes away,"
 * or, if it's enabled, specific debug statements can be dynamically
 * enabled at run time by setting ani_debug_mask, which is done
 * via the -anidebug command line switch.
 */
#ifndef _ani_debug_h
#define _ani_debug_h

/*
 * Available debugging options.
 * For an explanation, refer to the table in debug.c
 *
 * Adding an option is simple: pick a free number and add a define,
 * then #include "debug.h" in every file where you want to use the
 * option, and then use either DBG or DBGEXPR.
 *
 * If you add options, try to invent sensible categories.
 */
/* Debug Masks: (1 bit per option) */
# define DBG_BIT(x) (((ani_debugmask_t)1)<<x)
# define DBG_NONE		(0)
# define DBG_SOCKFD		DBG_BIT(0)
# define DBG_DEMUX_MATCH_DUMP	DBG_BIT(1)
# define DBG_THREADS		DBG_BIT(2)
# define DBG_THREADPOOLS	DBG_BIT(3)
# define DBG_THREADPOOL_STATS	DBG_BIT(4)

# define DBG_ALL		((ani_debugmask_t)(-1))
# define DBG_ANY                DBG_ALL


#if defined(NDEBUG) || !defined(DEBUG)
/* --- Debugging is NOT enabled --- */

/*
 * DBGIF:
 * 	if debugging is disabled, resolves to nothing.
 *	if debugging is enabled, resolve to statement.
 *	shortcut for #if DEBUG ... #endif
 */
# define DBGONLY(statement)  { ((void)0); }

/*
 * DBG:
 * 	if debugging is disabled, resolves to nothing.
 *	if debugging is enabled, statement is executed if mask matches.
 */
# define DBG(mask, statement)  { ((void)0); }

/* Define static inline, no-op versions of debugging functions. */
#define ani_debug_init()
#define ani_dprintf(fmt, args...)
#define ani_debug_setmask(mask)
#define ani_debug_setmaskstr(mask_str) (0)

#else 
/* --- Debugging is enabled --- */

/* Defines what debugging output is seen. Needs to be 64-bit. */
typedef long long ani_debugmask_t;

extern ani_debugmask_t ani_debug_mask;

# define DBGONLY(statement)  statement

/* Debug macros that are selected with the above flags. */

# define DBG(mask, statements) {	\
	if ((mask)&(ani_debug_mask)) {	\
		statements;		\
	}				\
	}

/* Initialize the ani_debug layer. */
void ani_debug_init(void);

/* A debugging-only printf.  Use this inside of DBG() statements. */
void ani_dprintf(const char *fmt, ...);

/* Set the debugging mask to use. (give the mask) */
void ani_debug_setmask(ani_debugmask_t mask);

/* 
 * Set the debugging mask to use. Give a string. For
 * parsing a command line option.  The string is a comma-separated
 * list of options which combined form the debugging mask.
 *
 * Return false if nothing was set
 */
ani_error_t ani_debug_setmaskstr(char *mask_str);

#endif /* defined(NDEBUG) || !defined(DEBUG) */

#endif /* _ani_debug_h */
