/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <nodeos/an_flow.h>

#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_cred.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_event.h>
#include <nodeos/impl/ani_moab_setup.h>

#include "os.h"
#include "demux.h"
#include "ani_debug.h"

/*
 * Setup Moab.
 *
 * See main.c for example code that sets up Moab.
 * These functions should only be called once and should
 * be called immediately by the EE's main().
 */

#include "version.h" 
static int version = VERSION;

static ani_icond_t shutdownCond;
static ani_ilock_t shutdownLock;
static int childdead;

#define MOABSTATE_NUTHIN         0
#define MOABSTATE_BARESETUPSTART 1
#define MOABSTATE_BARESETUPDONE  2
#define MOABSTATE_INITNODEOS     3
#define MOABSTATE_ALLSETUPDONE   4
#define MOABSTATE_STARTSHUTDOWN  5
#define MOABSTATE_ROOTFLOWGONE   6
#define MOABSTATE_COMPLETETOAST  7

int ani_moab_state = MOABSTATE_NUTHIN;

/*
 * Initialize the underlying device layers.
 */
void
ani_moab_setup(void)
{
	ani_moab_state = MOABSTATE_BARESETUPSTART;

	ani_debug_init();

	os_init_basic();
#ifdef GPROF
	os_gprof_init();
#endif

	os_init_mem();
	os_init_devices();
	os_init_net();
	os_init_fs();

	ani_moab_state = MOABSTATE_BARESETUPDONE;
}


/*
 * Return the compiled version of Moab, as an integer.
 */
int
ani_moab_version(void)
{
	return version;
}


/*
 * Start a new thread in the root flow on its way.
 * 
 * Initialize various NodeOS components.  Register the (uninitialized)
 * global objects passed in.
 *
 * XXX initialization of credentials will probably have to change.
 *
 * DOES NOT RETURN UNTIL MOAB IS SHUTDOWN.
 */
void
ani_moab_start(ani_flow_t *root_flow,
	       ani_mempool_t *root_mempool,
	       int nthreads, int stacksize,
	       void **root_threads, void **root_threadstacks,
	       ani_cred_t *root_cred,
	       ani_cred_t *initee_cred,
	       void (*initee_func)(void *),
	       void *initee_arg)
{
	ani_error_t err;

	assert(root_flow);
	assert(root_mempool);
	assert(root_cred);
	assert(initee_func);

	ani_moab_state = MOABSTATE_INITNODEOS;

	/*
	 * Initialize NodeOS components, setup global objects.
	 */
	ani_mempool_setup(root_mempool);
	ani_cred_setup(root_cred, initee_cred);
	ani_thread_setup();
	demux_init();		/* XXX needs name change pass */
	ani_chan_setup();
	ani_flow_setup(root_flow);
	assert(root_flow == ani_flow_root_flow());
	ani_event_setup();

	/* Initialize the root flow shutdown condition variable */
	ani_ilock_init(&shutdownLock);
	ani_icond_init(&shutdownCond);

	ani_moab_state = MOABSTATE_ALLSETUPDONE;

	childdead = 0;
	ani_ilock_lock(&shutdownLock);

	/*
	 * Create the root flow.
	 * This will fire off a thread at initee_func which does
	 * real root flow initialization in the context of the root
	 * flow.
	 *
	 * All memory allocations are charged to a flow from
	 * this point on.
	 *
	 * NOTE that there is no shutdown callback for the root
	 * flow.  It is never destroyed.  ani_moab_shutdown()
	 * should be called instead.
	 */
	err = ani_flow_rootflowinit(root_flow, root_cred, root_mempool,
				    nthreads, stacksize,
				    root_threads, root_threadstacks,
				    initee_func, initee_arg, NULL);
	assert(!ani_error_occured(err)); /* XXX */

	/*
	 * Wait until ani_moab_shutdown is called.
	 */
	while (childdead == 0)
		ani_icond_wait(&shutdownCond, &shutdownLock);

	ani_moab_state = MOABSTATE_STARTSHUTDOWN;

	/* Destroy the root flow and all its children. */
	ani_flow_rootflowdeinit(root_flow);

	ani_moab_state = MOABSTATE_ROOTFLOWGONE;

	/*
	 * XXX wait a while longer to allow things to shutdown cleanly.
	 */
	{
		an_timespec_t ts;

		ts.tv_sec = 1;
		ts.tv_nsec = 0;
		ani_thread_sleep(&ts);
	}

	/* Return to the Flow-free environment from whence we came. */
	ani_moab_state = MOABSTATE_COMPLETETOAST;
}

void
ani_moab_shutdown(void)
{
	ani_ilock_lock(&shutdownLock);
	childdead = 1;
	ani_icond_signal(&shutdownCond);
	ani_ilock_unlock(&shutdownLock);
}
