/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A "clock" channel driver.
 *
 * In-channel only.  Packets are delivered at a rate given by the address info.
 * (e.g., if address is "5", a packet is delivered every 5th clock tick).
 * The delivered packet contains the 4-byte internal time value (``mytime'').
 */
#include "options.h"

#ifdef CHAN_CLOCK

#include <sys/time.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_pbuf.h>
#include <nodeos/impl/ani_event.h>

#define CBUFSIZ	sizeof(unsigned int)

struct chandata {
	ani_event_t event;
	ani_chan_t *chan;
	int freq;
};

static ani_event_t ticker;
static unsigned int mytime;
static void chan_clock_intr(void *event, void *arg);
static unsigned long uspertick;

void
chan_clock_init(void)
{
	/*
	 * XXX chan_clock_init is called too early in the wrong context
	 * so we don't do anything.
	 */
}

void
chan_clock_start(void)
{
	ani_error_t err;
	static int done;

	if (done++)
		return;

	ani_event_setup();
	ani_event_info(&uspertick);
	err = ani_event_init(&ticker, ani_flow_threadpool(ani_flow_current()),
			     ANI_EV_INKERN|ANI_EV_PERIODIC|ANI_EV_DIRECT,
			     chan_clock_intr, 0, uspertick);
	assert(!ani_error_occured(err));
}

ani_error_t
chan_clock_create(ani_chan_t *chan)
{
	int freq;
	struct chandata *cdata;

	chan_clock_start();	/* XXX */

	freq = chan->addr.clock.freq;
	if (freq < 1)
		return ani_error_set(EINVAL, "invalid clockchan freqency");

	cdata = ani_mempool_alloc(ani_mempool_current(), sizeof *cdata);
	if (cdata == 0)
		return ani_error_set(ENOMEM, "no memory for clockchan data");

	cdata->chan = chan;
	cdata->freq = freq;
	chan->data = cdata;

	return ani_error_clear();
}

void
chan_clock_destroy(ani_chan_t *chan)
{
	struct chandata *cdata;

	assert(chan->data != 0);
	cdata = chan->data;
	assert(ani_error_code(ani_event_deinit(&cdata->event)) == ENOENT);

	ani_mempool_free(ani_mempool_current(), cdata, sizeof *cdata);
	chan->data = 0;
}

ani_error_t
chan_clock_enable(ani_chan_t *chan)
{
	struct chandata *cdata;
	ani_error_t err;

	assert(chan->data != 0);
	cdata = chan->data;

	err = ani_event_init(&cdata->event, chan->tpool,
			     ANI_EV_INKERN|ANI_EV_PERIODIC|ANI_EV_DIRECT,
			     chan_clock_intr, chan, cdata->freq * uspertick);
	if (ani_error_occured(err))
		ani_mempool_free(ani_mempool_current(), cdata, sizeof *cdata);

	return err;
}

void
chan_clock_disable(ani_chan_t *chan)
{
	struct chandata *cdata;
	ani_error_t err;

	assert(chan->data != 0);
	cdata = chan->data;

	err = ani_event_deinit(&cdata->event);
	assert(!ani_error_occured(err));
}

static void
chan_clock_intr(void *event, void *arg)
{
	ani_chan_t *chan = arg;
	ani_pbuf_t *pbuf;
	ani_error_t err;

	/*
	 * No channel means the ticker event.
	 */
	if (chan == 0) {
		mytime++;
		return;
	}

	ani_chan_lock(chan);
	err = ani_chan_pbuf_alloc(chan, CBUFSIZ, &pbuf);
	if (!ani_error_occured(err)) {
		*(unsigned int *)ani_pbuf_data(pbuf) = mytime;
		err = ani_chan_thread_start(chan, pbuf);
		if (ani_error_occured(err))
			ani_chan_pbuf_free(chan, pbuf);
	}
	ani_chan_unlock(chan);
}

#endif
