/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A raw ethernet channel driver.
 */
#include "options.h"

#ifdef CHAN_ETH

#include <stdio.h>
#include <string.h>
#include <assert.h>

#ifdef OSKIT
#include <stdlib.h>
#include <oskit/startup.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/ethernet.h>
#include <oskit/dev/net.h>
#include <oskit/net/ether.h>
#include <oskit/io/bufio.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#ifndef ETHER_HDR_SIZE
#define ETHER_HDR_SIZE 14
#endif

#include "os/unix_eth.h"
#endif

#include <nodeos/an_info.h>

#include <nodeos/impl/an_impl.h>
#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_pbuf.h>

#include "os.h"
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_demuxkey.h>
#include <nodeos/impl/ani_threadpool.h>
#include <nodeos/impl/ani_time.h>
#include "demux.h"
#include "chan_eth.h"
#include "pq/pq.h"

/*
 * Per-device structure.
 * Most oskit drivers restrict to only a single open of a device,
 * we use a single send/recv netio shared by many channels.
 *
 * The etherdev lock is lower in the locking hierarchy than the channel
 * lock.
 */
struct etherdev {
	ani_ilock_t	 lock;
	char		 name[NODEOS_MAXIFNAME]; /* "eth0", etc. */
	unsigned char	 addr[6];	/* our address */
	demux_t		 *demux;	/* list of demuxers for this device */
	char		 recv_inuse;	/* Flag */
	char		 send_inuse;	/* Flag */

#ifdef OSKIT	    
	/* OSKit specific */
	oskit_etherdev_t *eth;		/* device handle */
	oskit_netio_t	 *send;		/* handle for outgoing packets */
	oskit_netio_t	 *recv;		/* handle for incoming packets */
#else
	char		if_name[IFNAMSIZ]; /* system ethernet device name */
	int		recv;		/* socket file descriptor */
#endif
	/* Packet Scheduler */
	pq_scheduler_t	*pq_scheduler;
};

/*
 * Per-channel structure.
 * You get one of these per channel.
 */
struct etherdata {
	struct etherdev	*edev;		/* Per device Per channel */
	pq_client_t	*pq;		/* Each channel talks to a PQ */
};

static struct etherdev *etherdevs;
static int netherdevs;
static char chanprefix[] = "eth";
static ani_ilock_t ethlock;

static ani_error_t chan_eth_demux_prepend(ani_chan_t *chan);
static struct etherdev *chan_edev_find(char *name);
static ani_error_t chan_edev_init(struct etherdev *edev);
static void chan_edev_deinit(struct etherdev *edev);
#ifdef OSKIT
static oskit_error_t chan_edev_intr(void *arg, oskit_bufio_t *b,
				    oskit_size_t psize);
extern oskit_netio_t *oskit_threaded_netio_create(
  oskit_error_t (*func)(void *data, oskit_bufio_t *b, oskit_size_t pkt_size),
  void *data);
#else
static void chan_edev_intr(void *arg1, void *arg2);
#endif

void
chan_eth_init(void)
{
	int ndev, i;
#ifdef OSKIT
	oskit_etherdev_t **edev;
#else
	struct if_info *edev;
#endif	      

	ani_ilock_init(&ethlock);
	ani_ilock_lock(&ethlock);

	/*
	 * Look up all the ethernet devices
	 */
#ifdef OSKIT
	ndev = osenv_device_lookup(&oskit_etherdev_iid, (void***)&edev);
#else 
	ndev = unix_eth_get_if_info(&edev);
#endif
	if (ndev <= 0) {
		ani_ilock_unlock(&ethlock);
		return;
	}

	/*
	 * Create a device to name mapping
	 */
	etherdevs = ani_mempool_alloc(ani_mempool_current(),
				      ndev * sizeof(struct etherdev));
	if (etherdevs == 0)
	{
		ani_ilock_unlock(&ethlock);
		return;
	}
	
	memset(etherdevs, 0, ndev * sizeof(struct etherdev));
	for (i = 0; i < ndev; i++) {
		snprintf(etherdevs[i].name, sizeof etherdevs[i].name,
			 "%s%d", chanprefix, i);
#ifdef OSKIT
		etherdevs[i].eth = edev[i];
		oskit_etherdev_getaddr(edev[i], etherdevs[i].addr);
#else 
		strncpy(etherdevs[i].if_name, edev[i].name, IFNAMSIZ);
		memcpy(etherdevs[i].addr, edev[i].addr, 6);
#endif

		ani_ilock_init(&etherdevs[i].lock);
	}
	netherdevs = ndev;
	
#ifdef OSKIT
	free(edev);  /* XXX should be osenv_devicearray_free() */
#else
	unix_eth_free_if_info(ndev, edev);
#endif

	ani_ilock_unlock(&ethlock);
}

ani_error_t
chan_eth_info(an_infoif_t info[], int *count)
{
	int i, num;

	assert(count != 0);

	num = *count;
	if (num == 0) {
		*count = netherdevs;
		return ani_error_clear();
	}
	if (num > netherdevs)
		num = netherdevs;

	/*
	 * Fill in information about the interface.
	 * Note that the info id is assigned elsewhere and that
	 * the info ethaddr saddr and type are zero (indicating "any").
	 */
	memset(info, 0, sizeof info[0] * num);
	ani_ilock_lock(&ethlock);
	for (i = 0; i < num; i++) {
		info[i].type = ANIF_ETH;
		strncpy(info[i].pname, etherdevs[i].name,
			sizeof etherdevs[i].name);
		memcpy(info[i].addr.eth.daddr, etherdevs[i].addr,
		       sizeof etherdevs[i].addr);
	}
	ani_ilock_unlock(&ethlock);

	*count = num;
	return ani_error_clear();
}

/*
 * Hook up to the ethernet device.  Note that we do not allocate any actual
 * channel buffers for in-chans.  We use the buffer memory for incoming
 * packets.
 *
 * @allocation: ethernet specific channel data structure
 */
ani_error_t
chan_eth_create(ani_chan_t *chan)
{
	struct etherdev *edev;
	struct etherdata *edata;
	ani_error_t err;

	assert(ani_chan_type(chan) != CHAN_CUT);

	/*
	 * Lookup the device corresponding to the name given
	 */
	edev = chan_edev_find(ani_chan_name(chan));
	if (edev == 0)
		return ani_error_set(EINVAL, "not a valid chaneth device");

	/*
	 * Special in-channel processing.
	 */
	if (ani_chan_type(chan) == CHAN_IN) {
		/*
		 * Augment the demux filter with the address and demux info
		 * as necessary.
		 */
		err = chan_eth_demux_prepend(chan);
		if (err)
			return err;
	}

	/*
	 * Special out-channel processing.
	 */
	if (ani_chan_type(chan) == CHAN_OUT) {
		/*
		 * XXX debatable: use the source address as returned
		 * by the interface, not the one provided by the user.
		 * Maybe only do this if the user-supplied saddr is zero.
		 */
		memcpy(chan->addr.eth.saddr, edev->addr, sizeof edev->addr);
	}

	/*
	 * Allocate per-channel thingie.
	 */
	edata = ani_mempool_alloc(ani_mempool_current(),
				  sizeof(struct etherdata));
	if (edata == 0)
		return ani_error_set(ENOMEM, "no memory for chaneth data");
	chan->data = edata;

	/*
	 * Now hook up to the ethernet device
	 */
	ani_ilock_lock(&edev->lock);

	/*
	 * If device hasn't been initialized, do so now
	 */
	if (edev->recv == 0) {
		err = chan_edev_init(edev);
		if (err) {
			ani_ilock_unlock(&edev->lock);
			return err;
		}
	}
	assert(edev->recv != 0);
	edata->edev = edev;

#ifdef	OSKIT
	/*
	 * Create a packet scheduler client of the packet scheduler
	 * for the associated etherdev.
	 *
	 * XXX: Every channel gets 100 tickets until we have a way to
	 * specify such things. Also, let the PQ pick the default Q size.
	 */
	if (ani_chan_type(chan) == CHAN_OUT) {
		int ierr;

		ierr = pq_create_client(edata->edev->pq_scheduler, 0, 100,
					&edata->pq);
		if (ierr != 0) {
			/*
			 * Hmm, whats the right way to back out of this mess?
			 */
			ani_ilock_unlock(&edev->lock);
			return ani_error_set(ierr, "pq_create_client failed");
		}
	} else
		edata->pq = NULL;
#endif
	/*
	 * For deallocation. 
	 */
	if (ani_chan_type(chan) == CHAN_IN)
		edev->recv_inuse++;
	else
		edev->send_inuse++;

	ani_ilock_unlock(&edev->lock);
	return ani_error_clear();
}

/*
 * Called with the channel locked
 */
void
chan_eth_destroy(ani_chan_t *chan)
{
	struct etherdata *edata;
	struct etherdev *edev;

	assert(chan->tdata.in.demux == 0);

	edata = chan->data;
	
#ifdef	OSKIT
	/* Kill the packet scheduler. Any queued packets will be killed! */
	if (ani_chan_type(chan) == CHAN_OUT)
		pq_destroy_client(edata->pq);
#endif
	edev = edata->edev;
	ani_ilock_lock(&edev->lock);

	/*
	 * Destroy underlying goo when there are no more channels using it.
	 */
	if (ani_chan_type(chan) == CHAN_IN)
		edev->recv_inuse--;
	else
		edev->send_inuse--;

	assert(edev->send_inuse >= 0);
	assert(edev->recv_inuse >= 0);

	if (edev->recv_inuse == 0 && edev->send_inuse == 0)
		chan_edev_deinit(edev);

	/* XXX kill service threads, free up packets/bufios */
	
	ani_ilock_unlock(&edev->lock);

	/* free up the per-channel structure. */
	ani_mempool_free(ani_mempool_current(),	 edata, sizeof(*edata));
}

ani_error_t
chan_eth_enable(ani_chan_t *chan)
{
	struct etherdev *edev;
	ani_error_t err;

	assert(ani_chan_type(chan) == CHAN_IN);

	edev = ((struct etherdata *)chan->data)->edev;
	ani_ilock_lock(&edev->lock);

	err = ani_chan_demux_add(chan, &edev->demux);
	if (ani_error_occured(err) && chan->tdata.in.demux != 0) {
#if 0
		ani_demuxkey_destroy(chan->tdata.in.demux);
#endif
		chan->tdata.in.demux = 0;
	}

	ani_ilock_unlock(&edev->lock);
	return err;
}

void
chan_eth_disable(ani_chan_t *chan)
{
	struct etherdev *edev;

	assert(ani_chan_type(chan) == CHAN_IN);

	edev = ((struct etherdata *)chan->data)->edev;
	ani_ilock_lock(&edev->lock);
	ani_chan_demux_remove(chan, &edev->demux);
	ani_ilock_unlock(&edev->lock);

	if (chan->tdata.in.demux != 0) {
#if 0
		ani_demuxkey_destroy(chan->tdata.in.demux);
#endif
		chan->tdata.in.demux = 0;
	}
}

/*
 * Send a packet out on a channel.
 * Called with the channel locked.
 */
void
chan_eth_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
#ifdef OSKIT
	oskit_bufio_t *bio;
#else
	void *buf;
	size_t bufsize;
#endif
	struct etherdata *edata;
	struct etherdev *edev;
	int didcopy;
	size_t hsize;
	void *hdr;

	assert(ani_chan_type(chan) == CHAN_OUT);

	edata = chan->data;
	edev = edata->edev;

	if ((chan->flags & CHAN_ISRAW) == 0) {
		hdr = &chan->addr.eth;
		hsize = ETHER_HDR_SIZE;
	} else {
		hdr = 0;
		hsize = 0;
	}

#ifdef OSKIT	    
	/*
	 * Create a bufio for the packet and header.
	 */
	if (ani_pbuf_tobufio(pbuf, hdr, hsize, edev->send, &bio, &didcopy) != 0)
		return;

	/*
	 * Push the packet out. We should probably give our bufio reference
	 * to the pq code instead, since it does an addref and this does
	 * a release.
	 */
	pq_send_packet(edata->pq, bio, hsize+ani_pbuf_size(pbuf));
	oskit_bufio_release(bio);
	if (didcopy)
		CHANSTATS(chan->stats.outbufcopies++);
#else
	if (ani_pbuf_tomem(pbuf, hdr, hsize, &buf, &bufsize, &didcopy) != 0)
		return;

	os_write_packet(edev->recv, buf, bufsize);
	if (didcopy) {
		ani_pbuf_relmem(buf, bufsize);
		CHANSTATS(chan->stats.outbufcopies++);
	}
#endif /* OSKIT */	  
}

/*
 * Prepend channel address info into the demux information.
 * Called with the channel locked.
 *
 * XXX this function changes the user-provided demux,
 * it may also screw it up royally on an error.
 */
static ani_error_t
chan_eth_demux_prepend(ani_chan_t *chan)
{
	an_ethaddr_t *eaddr;
	ani_demuxkey_t *list = 0;
	ani_error_t err;
	int flag;
	static unsigned char zaddr[6] = { 0 };

	assert(ani_chan_type(chan) == CHAN_IN);
	assert(chan->tdata.in.demux != 0);

	/*
	 * If RAW, nothing to do.
	 */
	if (chan->flags & CHAN_ISRAW)
		return ani_error_clear();

	eaddr = &chan->addr.eth;

	/*
	 * Determine how many demux elements we are adding.
	 * XXX is it safe to use 0 addresses as wildcards?
	 */
	flag = 0;
	if (memcmp(eaddr->daddr, &zaddr, sizeof zaddr) != 0)
		flag |= 1;
	if (memcmp(eaddr->saddr, &zaddr, sizeof zaddr) != 0)
		flag |= 2;
	if (eaddr->type != 0)
		flag |= 4;

	/*
	 * Adjust offsets to account for enet header
	 */
	list = chan->tdata.in.demux;
	ani_demuxkey_adjust(list, ETHER_HDR_SIZE);

	/*
	 * Tack on matches for the address and type information
	 */
	err = ani_error_clear();
	if ((flag & 4) != 0 && err == 0)
		err = ani_demuxkey_prepend(list, 12, 2, &eaddr->type);
	if ((flag & 2) != 0 && err == 0)
		err = ani_demuxkey_prepend(list, 6, 6, eaddr->saddr);
	if ((flag & 1) != 0 && err == 0)
		err = ani_demuxkey_prepend(list, 0, 6, eaddr->daddr);

	return err;
}

static struct etherdev *
chan_edev_find(char *name)
{
	int i, len;

	len = strlen(name);

	ani_ilock_lock(&ethlock);
	for (i = 0; i < netherdevs; i++)
		if (strncmp(etherdevs[i].name, name, len) == 0) {
			ani_ilock_unlock(&ethlock);
			return &etherdevs[i];
		}

	ani_ilock_unlock(&ethlock);
	return 0;
}

#ifdef OSKIT
/*
 * Receive netio push function, called from the interrupt handler but in
 * a full thread context.  Bufio comes in with a ref we must release to
 * free the buffer.
 *
 * This routine serializes all packets coming in on the netio (currently one
 * netio per interface) so lets get in and out as quick as possible.
 */
static oskit_error_t
chan_edev_intr(void *arg, oskit_bufio_t *b, oskit_size_t psize)
{
	struct etherdev *edev = arg;
	ani_chan_t *chan;
	ani_pbuf_t *pbuf;
	size_t hsize;
	void *data;
	ani_error_t err;

	/*
	 * Map the packet for demux examination
	 */
	assert(psize >= ETHER_HDR_SIZE);
	if (oskit_bufio_map(b, &data, 0, psize) != 0) {
		oskit_bufio_release(b);
		return 0;
	}

	/*
	 * Find a channel for this packet.
	 * If no channel matches, the packet is dropped.
	 */
	ani_ilock_lock(&edev->lock);
	chan = demux_apply(edev->demux, data, psize);

	/*
	 * XXX as soon as we drop the device lock, a pending ani_chan_destroy
	 * blocked in disable could complete.  Grab a reference to the
	 * channel to keep our pointer valid til we can lock the channel
	 * (see below).
	 */
	if (chan) {
		assert_ani_chan_valid(chan);
		ani_chan_reference(chan);
	}

	ani_ilock_unlock(&edev->lock);
	oskit_bufio_unmap(b, data, 0, psize);
	if (chan == 0) {
		oskit_bufio_release(b);
		return 0;
	}

	/*
	 * XXX deal with a race condition involving channel destruction.
	 * Channel destroy takes the channel lock and then attempts to
	 * disable the channel.	 Disabling the channel requires the edev
	 * lock, so channel destruction might block while we are doing the
	 * demux above.	 In that case, we will finish the demux, drop the
	 * edev lock and then block on the channel lock here.  In the
	 * meantime, the destroy call will acquire the edev lock and finish
	 * the channel destruction.  We then wake up after acquiring the
	 * channel lock with a dead channel.
	 */
	ani_chan_lock(chan);
	if (chan->flags & CHAN_DYING) {
		ani_chan_unlock(chan);
		ani_chan_release(chan);
		oskit_bufio_release(b);
		return 0;
	}
	assert(chan->refs > 1);
	ani_chan_release(chan);

	/*
	 * Wrap the buffio with a NodeOS packet buffer
	 * and fire off a handler thread.
	 */
	hsize = (chan->flags & CHAN_ISRAW) ? 0 : ETHER_HDR_SIZE;
	psize -= hsize;
	err = ani_pbuf_frombufio(chan->flow, chan, b, hsize, psize, &pbuf);
	oskit_bufio_release(b);
	if (!ani_error_occured(err)) {
		err = ani_chan_thread_start(chan, pbuf);
		if (ani_error_occured(err))
			ani_chan_pbuf_free(chan, pbuf);
	}

	ani_chan_unlock(chan);

	return 0;
}

#else

void
chan_edev_intr(void *arg, void *arg2)
{
	struct etherdev *edev = arg;
	int count;
	ani_error_t err;
	char *buf;
	int fd, bufsize;
	ani_pbuf_t *rbuf, *pbuf;
	ani_chan_t *chan;

	/*
	 * Locking forces synchronization with our creator
	 */
	ani_ilock_lock(&edev->lock);
	fd = edev->recv;
	ani_ilock_unlock(&edev->lock);

	/*
	 * Create an initial read buffer.  We swap this buffer with
	 * empty ones from the devices' channels when we hand off packets.
	 *
	 * XXX should the caller allocate this?
	 */
	rbuf = ani_mempool_alloc(ani_mempool_current(), sizeof *rbuf);
	assert(rbuf != 0);
	err = ani_pbuf_init(rbuf, ani_flow_current(), 0,
			    NODEOS_CHAN_BUFSIZE, 0, NODEOS_CHAN_BUFSIZE);
	assert(!ani_error_occured(err));
	buf = ani_pbuf_data(rbuf);
	bufsize = ani_pbuf_size(rbuf);

	while (1) {
		chan = 0;

		count = os_read_packet(fd, buf, bufsize);
		if (count < 0) {
			err = ani_error_set(errno, "os_read_packet failed");
			break;
		}

		/* XXX should we delay here? */
		if (count == 0)
			continue;
		
		/*
		 * Find a channel for this packet.
		 * If no channel matches, the packet is dropped.
		 */
		ani_ilock_lock(&edev->lock);
		chan = demux_apply(edev->demux, buf, count);
		ani_ilock_unlock(&edev->lock);
		if (chan == 0)
			continue;

		/*
		 * Allocate a packet buffer from the channel.
		 * If none are available, just drop the packet.
		 */
		ani_chan_lock(chan);
		err = ani_chan_pbuf_alloc(chan, NODEOS_CHAN_BUFSIZE, &pbuf);
		if (ani_error_occured(err)) {
			ani_chan_unlock(chan);
			continue;
		}

		/*
		 * Swap the buffer obtained from the channel with our buffer.
		 * The channel gets the full buffer in exchange for one of its
		 * empty ones that we will use for the next read.
		 */
		ani_pbuf_setsize(rbuf, count, 0);
		ani_pbuf_swap(rbuf, pbuf);

		/*
		 * If this is not a raw ethernet channel, strip off the
		 * ethernet header.
		 */
		if ((chan->flags & CHAN_ISRAW) == 0)
			ani_pbuf_setoffset(pbuf, ETHER_HDR_SIZE, 0);

		/*
		 * Finally, fire up a handler thread.
		 */
		err = ani_chan_thread_start(chan, pbuf);
		if (ani_error_occured(err))
			ani_chan_pbuf_free(chan, pbuf);

		ani_chan_unlock(chan);

		/*
		 * Reset the buffer pointer/size since we have a new
		 * physical buffer.
		 */
		buf = ani_pbuf_data(rbuf);
		bufsize = ani_pbuf_size(rbuf);
	}

	/*
	 * We have broken out of the recvfrom.	Most likely because the
	 * socket was closed.  Free the read buffer and exit.
	 */
	ani_pbuf_deinit(rbuf);
	panic("NodeOS: enet interface error (%d)", err);
}
#endif /* OSKIT */

static ani_error_t
chan_edev_init(struct etherdev *edev)
{
#ifdef OSKIT	    
	oskit_error_t err;

	/*
	 * Create the netio object we will use for receiving packets.
	 */
	edev->recv = oskit_threaded_netio_create(chan_edev_intr, edev);
	if (edev->recv == 0)
		return ani_error_set(ENOMEM,
				     "oskit_threaded_netio_create failed");

	/*
	 * Open the device registering our receive netio and getting
	 * back the send netio.
	 */
	assert(edev->send == 0);
	err = oskit_etherdev_open(edev->eth, 0, edev->recv, &edev->send);
	if (err) {
		oskit_netio_release(edev->recv);
		return ani_error_set(err, "oskit_etherdev_open failed");
	}

	/*
	 * Create a packet scheduler for this device. Eventually, this might
	 * work under non-oskit situations.
	 */
	err = pq_create_scheduler(edev->send, 0, &edev->pq_scheduler);
	if (err) {
		oskit_netio_release(edev->recv);
		oskit_netio_release(edev->send);
		return ani_error_set(err, "pq_create_scheduler failed");
	}
#else
	ani_error_t err;

	edev->recv = os_open_eth(edev->if_name);
	if (edev->recv < 0) 
		return ani_error_set(EINVAL, "cannot open ethernet device");
	
	/*
	 * Create the reader thread
	 */
	err = ani_threadpool_thread_start(&ani_flow_root_flow()->tpool, 1,
					  chan_edev_intr, edev, 0, 0);
	assert(err == 0);
#endif	      
	return ani_error_clear();
}

/*
 * Gotta release the ethernet devices or my PC hangs up in netboot reboot.
 */
static void
chan_edev_deinit(struct etherdev *edev)
{
#ifdef OSKIT
	oskit_netio_release(edev->send);
	oskit_netio_release(edev->recv);
	oskit_etherdev_release(edev->eth);
	pq_destroy_scheduler(edev->pq_scheduler);

	/* XXX This will cause another trip through chan_edev_init() */
	edev->send = edev->recv = 0;
#else
	os_close_eth(edev->recv);
	edev->recv = -1;
#endif	      
}

#endif /* CHAN_ETH */
