/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <assert.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_demuxkey.h>
#include <nodeos/impl/ani_mempool.h>
#include "demux.h"
#include "ani_debug.h"


struct ani_demux_triple { /* XXX rename 'ani_demuxkey_segment' */
	int off;
	int len;
	unsigned char key[0];
};

static struct ani_demux_triple *demux_triple_create(int off, int len, void *key);
static void demux_triple_destroy(struct ani_demux_triple *dt);
#ifdef DEBUG
static void demux_triple_dump(struct ani_demux_triple *dt);
#endif

ani_error_t
ani_demuxkey_init(ani_demuxkey_t *list)
{
	assert(list != 0);

	memset(list, 0, sizeof *list);
	list->listsize = ANI_DEMUXKEY_MAX;

	return ani_error_clear();
}

ani_error_t
ani_demuxkey_deinit(ani_demuxkey_t *list)
{
	int i;

	assert(list != 0);

	for (i = 0; i < list->ntriples; i++) {
		assert(list->list[i] != 0);
		demux_triple_destroy(list->list[i]);
	}
	list->ntriples = 0;

	return ani_error_clear();
}

ani_error_t
ani_demuxkey_clone(ani_demuxkey_t *newlist, ani_demuxkey_t *list)
{
	int i;

	assert(newlist != 0);
	assert(list != 0);

	newlist->maxoff = list->maxoff;
	newlist->listsize = list->listsize;
	for (i = 0; i < list->ntriples; i++) {
		assert(list->list[i] != 0);
		newlist->list[i] = demux_triple_create(list->list[i]->off,
						       list->list[i]->len,
						       list->list[i]->key);
		if (newlist->list[i] == 0) {
			ani_demuxkey_deinit(newlist);
			return ani_error_set(ENOMEM,
					     "no memory for demuxkey element");
		}
		newlist->ntriples++;
	}

	return ani_error_clear();
}

/*
 * Create and add a match element to the front of the given list.
 */
ani_error_t
ani_demuxkey_prepend(ani_demuxkey_t *list, int off, int len, void *key)
{
	struct ani_demux_triple *dt;
	int i;

	assert(list != 0);

	i = list->ntriples;
	if (i == list->listsize)
		return ani_error_set(ENOMEM, "no space in demuxkey");

	dt = demux_triple_create(off, len, key);
	if (dt == 0)
		return ani_error_set(ENOMEM, "no memory for demuxkey element");

	while (i > 0) {
		list->list[i] = list->list[i-1];
		i--;
	}
	list->list[i] = dt;
	list->ntriples++;

	if (off + len > list->maxoff)
		list->maxoff = off + len;

	return ani_error_clear();
}


/*
 * Create and add a match element to the end of the given list.
 */
ani_error_t
ani_demuxkey_append(ani_demuxkey_t *list, int off, int len, void *key)
{
	struct ani_demux_triple *dt;
	int i;

	assert(list != 0);

	i = list->ntriples;
	if (i == list->listsize)
		return ani_error_set(ENOMEM, "no space in demuxkey");

	dt = demux_triple_create(off, len, key);
	if (dt == 0)
		return ani_error_set(ENOMEM, "no memory for demuxkey element");

	list->list[i] = dt;
	list->ntriples++;

	if (off + len > list->maxoff)
		list->maxoff = off + len;

	return ani_error_clear();
}

/*
 * Adjust the offset of all triples by the indicated amount
 */
void
ani_demuxkey_adjust(ani_demuxkey_t *list, int adjust)
{
	int i;

	assert(list != 0);

	if (list->ntriples != 0) {
		for (i = 0; i < list->ntriples; i++) {
			assert(list->list[i] != 0);
			list->list[i]->off += adjust;
		}
		list->maxoff += adjust;
	}
}

/*
 * Determine if the list has room for the given number of new elements.
 * Returns 0 if so, an error otherwise.
 */
ani_error_t
ani_demuxkey_expand(ani_demuxkey_t *list, int nelts)
{
	assert(list != 0);

	if (list->ntriples + nelts > list->listsize)
		return ani_error_set(ENOMEM, "no space in demuxkey");

	return ani_error_clear();
}

/* 
 * Compare a given key to a given packet.  Returns -1 if they don't
 * match.  Otherwise, returns the offset in the pdata of the first byte
 * in the packet *after* the key.
 *
 * 'ddata' is the demux key to compare against the 'pdata'/'plen' (the
 * packet).
 */
int
ani_demuxkey_match(void *ddata, void *pdata, int plen)
{
	ani_demuxkey_t *list = ddata;
	struct ani_demux_triple *dt;
	int i;

	assert(list != 0);

	/*
	 * Null demuxkey matches the entire packet.
	 */
	if (list->ntriples == 0)
		return plen;

	/*
	 * Packet not long enough for our filters, cannot match.
	 */
	if (list->maxoff > plen)
		return -1;

	/*
	 * Otherwise we must iterate through the list matching the data.
	 */
	DBG(DBG_DEMUX_MATCH_DUMP,{
		    int j;
		    unsigned char *cp = pdata;
		    
		    ani_dprintf("DATA: ");
		    for (j = 0; j < list->maxoff; j++)
			    ani_dprintf("%02x", cp[j]);
		    ani_dprintf("\n");
	});

	for (i = 0; i < list->ntriples; i++) {
		assert(list->list[i] != 0);
		dt = list->list[i];
		DBG(DBG_DEMUX_MATCH_DUMP, {
			int j;
			int len;
			unsigned char *cp = pdata + dt->off;

			demux_triple_dump(dt);
			ani_dprintf("DATA: ");
			len = plen - dt->off;
			if (len > dt->len)
				len = dt->len;
			for (j = 0; j < len; j++)
				ani_dprintf("%02x", cp[j]);
			ani_dprintf("\n");
		});

		if (memcmp(pdata+dt->off, dt->key, dt->len) != 0) {
			DBG(DBG_DEMUX_MATCH_DUMP,{
				ani_dprintf("FAILED\n");
			});
			return -1;
		}
	}

	DBG(DBG_DEMUX_MATCH_DUMP,{
		ani_dprintf("MATCHED\n");
	});

	return list->maxoff; /* XXX static value for static keys. */
}

static struct ani_demux_triple *
demux_triple_create(int off, int len, void *key)
{
	struct ani_demux_triple *dt;

	/* XXX memory allocation */
	dt = ani_mempool_alloc(ani_mempool_current(), sizeof *dt + len);
	if (dt == 0)
		return 0;

	dt->off = off;
	dt->len = len;
	if (key)
		memcpy(dt->key, key, len);
	else
		memset(dt->key, 0, len);

	return dt;
}

static void
demux_triple_destroy(struct ani_demux_triple *dt)
{
	/* XXX memory (de)allocation */
	ani_mempool_free(ani_mempool_current(), dt, sizeof *dt + dt->len);
}

#ifdef DEBUG
static void
demux_triple_dump(struct ani_demux_triple *dt)
{
	int j;

	ani_dprintf("TRIPLE %p: %d/%d/0x", dt, dt->off, dt->len);
	for (j = 0; j < dt->len; j++)
		ani_dprintf("%02x", dt->key[j]);
	ani_dprintf("\n");
}

void
ani_demuxkey_dump(ani_demuxkey_t *list)
{
	int i;

	ani_dprintf("TLIST %p, ntriples=%d, maxoff=%d:\n",
	       list, list->ntriples, list->maxoff);
	for (i = 0; i < list->ntriples; i++) {
		assert(list->list[i] != 0);
		demux_triple_dump(list->list[i]);
	}
}
#endif
