/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <nodeos/impl/an_impl.h>
#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_event.h>
#include <nodeos/impl/ani_threadpool.h>

#include "os.h"
#include "ani_debug.h"

static void *eventtimer;
static ani_event_t *eventlist;
static ani_ilock_t eventlock;
static unsigned long uspertick;
static unsigned long mytime;

static void ani_event_intr(void *oskitthingee, void *arg);

void
ani_event_setup(void)
{
	static int called;

	if (called++)
		return;

	ani_ilock_init(&eventlock);

	eventtimer = os_timer_init(ani_event_intr, 0, &uspertick);
	assert(eventtimer != 0);
	assert(uspertick != 0);

	os_timer_start(eventtimer, uspertick);
}

void
ani_event_info(unsigned long *timerres)
{
	assert(timerres);
	*timerres = uspertick;
}

static inline void
ani_event_add_locked(ani_event_t *event)
{
	ani_event_t *ev, **pevp;

	assert(ani_ilock_islocked(&eventlock));

	for (pevp = &eventlist; (ev = *pevp) != 0; pevp = &ev->next)
		if (event->ticks < ev->ticks)
			break;
	event->next = ev;
	*pevp = event;
}

static inline ani_error_t
ani_event_remove_locked(ani_event_t *event)
{
	ani_event_t *ev, **pevp;

	assert(ani_ilock_islocked(&eventlock));

	for (pevp = &eventlist; (ev = *pevp) != 0; pevp = &ev->next)
		if (event == ev) {
			*pevp = ev->next;
			return ani_error_clear();
		}

	return ani_error_set(ENOENT, "no such event");
}

ani_error_t
ani_event_init(ani_event_t *event, ani_threadpool_t *tpool, int flags,
	       an_eventfunc_t func, void *farg, unsigned long delta)
{
	unsigned long ticks;

	if (event == 0 || tpool == 0 || func == 0)
		return ani_error_set(EINVAL,
				     "NULL argument to ani_event_init");

	ticks = (delta + uspertick-1) / uspertick;
	if (ticks == 0 && (flags & ANI_EV_PERIODIC) != 0)
		return ani_error_set(EINVAL,
				     "invalid period for periodic event");

	memset(event, 0, sizeof *event);
	event->tpool = tpool;
	event->flags = flags;
	event->func = func;
	event->farg = farg;
	event->ticks = mytime + ticks;
	event->freq = ticks;
	ani_ilock_lock(&eventlock);
	ani_event_add_locked(event);
	ani_ilock_unlock(&eventlock);

	return ani_error_clear();
}

ani_error_t
ani_event_deinit(ani_event_t *event)
{
	ani_error_t err;

	ani_ilock_lock(&eventlock);
	err = ani_event_remove_locked(event);
	ani_ilock_unlock(&eventlock);

	return err;
}

/*
 * Called every uspertick microseconds.  Fires all the events that are due.
 *
 * XXX if we cannot schedule a thread to handle the event, we ignore the
 * event.  This should only happen if the flow (threadpool) is shutting down
 * (err == EINVAL).
 */
static
void ani_event_intr(void *oskitthingee, void *arg)
{
	ani_event_t *ev;
	ani_error_t err;

	mytime++;
	assert(mytime != 0);

	ani_ilock_lock(&eventlock);
	while ((ev = eventlist) != 0 && ev->ticks <= mytime) {
		eventlist = ev->next;
		ani_ilock_unlock(&eventlock);

		assert_ani_threadpool_valid(ev->tpool);

		if (ev->flags & ANI_EV_DIRECT) {
			assert((ev->flags & ANI_EV_INKERN) != 0);
			ev->func(0, ev->farg);
			err = ani_error_clear();
		} else
			err = ani_threadpool_thread_schedule(ev->tpool, 0,
						(ev->flags & ANI_EV_INKERN), 0,
						(ani_thread_func_t)ev->func, 0,
						ev->farg);
		ani_ilock_lock(&eventlock);

		if (ani_error_code(err) != EINVAL &&
		    (ev->flags & ANI_EV_PERIODIC)) {
			ev->ticks += ev->freq;
			ani_event_add_locked(ev);
		}
	}
	ani_ilock_unlock(&eventlock);
}

/*
 * Remove all events associated with a particular threadpool.  Should be
 * called with the threadpool unlocked, otherwise we go against the locking
 * order used by event_intr above (eventlock then threadpool lock).
 */
void
ani_event_removeall(ani_threadpool_t *tpool)
{
	ani_event_t *ev, *next;

	assert(ani_threadpool_islocked(tpool) == 0);

	ani_ilock_lock(&eventlock);
	for (ev = eventlist; ev != 0; ev = next) {
		next = ev->next;
		if (ev->tpool == tpool)
			ani_event_remove_locked(ev);
	}
	ani_ilock_unlock(&eventlock);
}

void
ani_event_dump(char *str)
{
#ifdef DEBUG
	ani_event_t *ev;

	ani_ilock_lock(&eventlock);
	ani_dprintf("%s:\n", str);
	for (ev = eventlist; ev != 0; ev = ev->next) {
		ani_dprintf("%p: ticks=%lu, freq=%lu, tpool=%p, flags=%#x, "
			    "func=%p, farg=%p\n",
			    ev, ev->ticks, ev->freq, ev->tpool, ev->flags,
			    ev->func, ev->farg);
	}
	ani_ilock_unlock(&eventlock);
#endif
}
