/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h>
#include <assert.h>

#include <nodeos/an_resource.h>
#include <nodeos/an_mempool.h>
#include <nodeos/an_thread.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_threadpool.h>
#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_cred.h>
#include <nodeos/impl/ani_resource.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>

#include "ani_debug.h"

static int ani_flow_disable_internal(struct hnode *h, void *arg);
static int ani_flow_destroy_internal(struct hnode *h, void *arg);

/* flags to destroy_internal */
#define DI_DISABLE	1
#define DI_WAITFOR	2

extern int bootstrapped;
ani_flow_t *root_flow = NULL;


void
ani_flow_setup(ani_flow_t *rflow)
{
	/*
	 * rflow has not been initialized at this point, its still
	 * just an arbitrary hunk of memory.
	 */
	root_flow = rflow;
}

/*
 * Initialize the root flow, firing off the first true NodeOS thread.
 */
ani_error_t
ani_flow_rootflowinit(ani_flow_t *rflow,
		      ani_cred_t *root_cred, ani_mempool_t *root_mempool,
		      int nthreads, int stacksize,
		      void **root_threads, void **root_threadstacks,
		      an_flowinit_func_t initfunc, void *initarg,
		      an_flowterm_func_t termfunc)
{
	an_cpuspec_t cspec;
	an_memspec_t mspec;
	ani_error_t err;

	/* Provide default stacksize if one isn't given. */
	if (stacksize <= 0) {
		assert(nthreads <= 0); /* XXX */
		assert(root_threads == NULL); /* XXX */
		assert(root_threadstacks == NULL); /* XXX */
		stacksize = 64 * 1024;
	}
	
	/* Provide default number of implicit threads if not given. */
	if (nthreads <= 0) {
		assert(root_threads == NULL); /* XXX */
		assert(root_threadstacks == NULL); /* XXX */
		nthreads = 8; /* XXX ? */
	}

	if (root_threads != NULL) {
		int i;

		assert(root_threadstacks); /* XXX */

		/* Convert unitialized thread object memory into ani_thread_t */
		for (i = 0; i < nthreads; i++) {
			ani_thread_t *ith;

			assert(root_threads[i]); /* XXX */
			assert(root_threadstacks[i]); /* XXX */
			ith = NODEOS_INTERN_THREAD(root_threads[i]);
			assert(ith);
			assert(ith == root_threads[i]); /* XXX */
			ani_thread_init(ith, root_threadstacks[i], stacksize);
		}
		
		cspec.ttype = AN_TH_EXPLICIT;
		cspec.threads.ex.nthreads = nthreads;
		/* XXX should not hold onto user's "root_threads".  Use a linked list. */
		cspec.threads.ex.threads = (struct an_thread**)root_threads;
	} else {
		cspec.ttype = AN_TH_IMPLICIT;
		cspec.threads.im.maxthreads = nthreads;
		cspec.threads.im.stacksize = stacksize;
	}

	cspec.rate = RESOURCE_ALL;
	cspec.slice.tv_sec = RESOURCE_ALL;
	cspec.sched = 0; /* XXX */

	mspec.npages = 100; /* XXX */
	mspec.pool = NODEOS_EXTERN_MEMPOOL(root_mempool);
	err = ani_flow_init(rflow, root_cred, &cspec, &mspec, initfunc,
			    initarg, termfunc);
	return err;
}

ani_flow_t *
ani_flow_current(void)
{
	ani_thread_t *thread = ani_thread_current();

	return thread ? thread->pool->flow : 0;
}

ani_flow_t *
ani_flow_root_flow(void)
{
	return root_flow;
}


ani_error_t
ani_flow_init(ani_flow_t *flow, ani_cred_t *cred,
	      an_cpuspec_t *cspec, an_memspec_t *mspec,
	      an_flowinit_func_t initfunc, void *initarg,
	      an_flowterm_func_t termfunc)
{
	ani_flow_t *pflow;
	ani_mempool_t *mpool;
	resource_t res;
	ani_error_t err;

	/* Sanity */
	if (flow == 0 || cred == 0 || cspec == 0 || mspec == 0)
		return ani_error_set(EINVAL,
				     "invalid argument to ani_flow_init");

	/* Can this flow create sub flows? */
	if (ani_cred_check(cred, CCHECK_FLOW_CREATE) == CCHECK_FAIL)
		return ani_error_set(EPERM,
				     "insufficient permission to create flow");

	/* Ensure requested resources are available and reserve them */
	resource_init(&res, cspec, mspec);
	err = ani_cred_resource_alloc(cred, &res);
	if (err)
		return err;

	pflow = ani_flow_current();

	/*
	 * Extract the target mempool so we can start charging it.
	 *
	 * XXX this doesn't really belong here.  We may eventually
	 * just pass in a mempool here, with the current memspec
	 * params being associated with the pool when it is created.
	 */
	mpool = NODEOS_INTERN_MEMPOOL(mspec->pool);
	if (mpool == 0) {
		ani_cred_resource_free(cred, &res);
		return ani_error_set(EINVAL,
				     "invalid mempool");
	}

	/* Initialize the flow */
	hnode_init(&flow->me);
	flow->cred = cred;
	flow->res = res;
	flow->chanlist = 0;
	flow->flags = 0;
	flow->termfunc = termfunc;
	flow->refs = 1;
	ani_ilock_init(&flow->lock);

	/* Link to the mempool now for threadpool memory allocations */
	flow->mpool = mpool;
	ani_mempool_addflow(flow->mpool, flow);

	/* Create the threadpool */
	err = ani_threadpool_init(&flow->tpool, flow, cspec);
	if (err) {
		ani_mempool_removeflow(mpool, flow);
		ani_ilock_deinit(&flow->lock);
		ani_cred_resource_free(cred, &res);
		return err;
	}

	/* Link into flow hierarchies (special handling for first flow) */
	if (flow == root_flow) {
		ani_flow_lock(flow);
		hnode_attach(&flow->me, 0, 0);
	} else {
		ani_flow_lock(root_flow);
		ani_flow_lock(flow);
		hnode_attach(&flow->me, &ani_flow_current()->me, 0);
		ani_flow_unlock(root_flow);
	}

	ani_flow_unlock(flow);

	/* All setup, fire up a thread in the flow to call the init function */
	if (initfunc) {
		err = ani_flow_startthread(flow, 0,
					   (void (*)(void *, void *))initfunc,
					   initarg, 0);
		if (err) {
			/*
			 * Cannot deinit the root flow if something
                         *  goes wrong at startup.
			 */
			if (flow != root_flow)
				ani_flow_deinit(flow);

			return err;
		}
	}

	return ani_error_clear();
}

/*
 * Deinit the root flow, cleaning up all child flows.
 * Similar to ani_flow_deinit, except there is no parent notification.
 * Called outside the context of any NodeOS flow.
 */
void
ani_flow_rootflowdeinit(ani_flow_t *flow)
{
	int err;

	assert(flow);
	assert(flow == root_flow);

	ani_flow_lock(root_flow);
	ani_flow_reference(root_flow);
	err = hnode_traverse(&flow->me, ani_flow_destroy_internal,
			     (void *)(DI_DISABLE|DI_WAITFOR), HNODE_PARENT, 1);
	assert(err == 0); /* XXX */
	ani_flow_unlock(root_flow);

	/*
	 * Release the reference we took above now that we have dropped the
	 * lock.  The NodeOS world is over at this point.
	 */
	ani_flow_release(root_flow);
}

ani_error_t
ani_flow_deinit(ani_flow_t *flow)
{
	ani_flow_t *oflow;
	int me = 0, err;

	assert_ani_flow_valid(flow);
	assert(flow != root_flow);

	ani_flow_lock(flow);

	/*
	 * Flow must be us or one of our immediate children
	 */
	if (flow == ani_flow_current())
		me = 1;
	else {
		for (oflow = ani_flow_child(ani_flow_current()); oflow != 0;
		     oflow = ani_flow_sibling(oflow))
			if (oflow == flow)
				break;
		if (oflow == 0) {
			ani_flow_unlock(flow);
			return ani_error_set(EINVAL, "invalid flow argument");
		}
	}

	/*
	 * Prevent recursive upcalls and, effectively, attempts to destroy
	 * an already dying flow.  Note that the flow lock is not held while
	 * in the upcall so we need this other mechanism to ensure this.
	 */
	if (flow->flags & FLOW_INUPCALL) {
		ani_flow_unlock(flow);
		return ani_error_clear();
	}
	flow->flags |= FLOW_INUPCALL;
	ani_flow_unlock(flow);

	/*
	 * First disable all the threadpools so that no new flow activity
	 * will occur while we are shutting down.
	 */
	ani_flow_lock(root_flow);
	err = hnode_traverse(&flow->me, ani_flow_disable_internal, 0,
			     HNODE_PARENT, 1);
	assert(err == 0);
	ani_flow_unlock(root_flow);

	/*
	 * Now make the termination upcall.
	 */
	if (flow->termfunc != 0) {
		oflow = ani_flow_parent(flow);
		assert(oflow != 0);
		(void) ani_threadpool_thread_run(&oflow->tpool, 0,
					(ani_thread_func_t)flow->termfunc,
					flow, 0, 0);
	}

	if ((flow->flags & FLOW_DYING) == 0) {
		ani_flow_lock(root_flow);
		err = hnode_traverse(&flow->me, ani_flow_destroy_internal,
				     (void *)(me ? 0 : DI_WAITFOR),
				     HNODE_PARENT, 1);
		assert(err == 0);
		ani_flow_unlock(root_flow);
	}

	/*
	 * If we just destroyed ourself, there is no point in returning.
	 * Our thread's reference is keeping the flow alive at this point,
	 * exiting will take care of that!
	 */
	if (me)
		ani_thread_exit();

	return ani_error_clear();
}

/*
 * Disable all future flow activity by removing all threads from the
 * flow's threadpool.
 */
static int
ani_flow_disable_internal(struct hnode *h, void *arg)
{
	ani_flow_t *flow = (ani_flow_t *)h;

	assert_ani_flow_valid(flow);

	ani_flow_lock(flow);
	ani_threadpool_disable(&flow->tpool, 1);
	ani_flow_unlock(flow);	

	return 0;
}

/*
 * The guts of flow destruction.  Called by the hnode_traverse in
 * ani_flow_destroy.
 */
static int
ani_flow_destroy_internal(struct hnode *h, void *arg)
{
	ani_flow_t *flow = (ani_flow_t *)h;
	int flags = (int)arg;
	ani_flow_t *pflow;
	ani_chan_t *chan, *next;

	assert_ani_flow_valid(flow);

	if (flow != root_flow)
		ani_flow_lock(flow);
	pflow = ani_flow_parent(flow);
	flow->flags |= FLOW_DYING;

	/* Remove from flow hierarchy */
	hnode_detach(h, HNODE_PARENT);

	/* Destroy all the channels */
	for (chan = flow->chanlist; chan != 0; chan = next) {
		next = chan->next;
		ani_chan_deinit(chan);
	}

	/* Kill any remaining threads */
	if (flags & DI_DISABLE)
		ani_threadpool_disable(&flow->tpool, 1);
	ani_threadpool_deinit(&flow->tpool, (flags & DI_WAITFOR));

	/* Release the initial reference to the flow */
	if (flow != root_flow)
		ani_flow_unlock(flow);
	ani_flow_release(flow);

	return 0;
}

/*
 * Called when there are no more references to the flow.
 * At this point we can safely free resources.
 */
void
ani_flow_cleanup(ani_flow_t *flow)
{
	assert_ani_flow_valid(flow);

	ani_flow_lock(flow);

	/* Remove from mempool */
	assert(hnode_listhead(&flow->me) != 0);
	ani_mempool_removeflow(flow->mpool, flow);

	/* Return resources to the system */
	ani_cred_resource_free(flow->cred, &flow->res);

	/* Uninitialize and free all the data structures */
	ani_flow_unlock(flow);
	ani_ilock_deinit(&flow->lock);
}

void
ani_flow_addchannel(ani_flow_t *flow, ani_chan_t *chan)
{
	assert_ani_flow_valid(flow);

	ani_flow_lock(flow);
	ani_chan_lock(chan);

	/* XXX make sure requested channel resources are consistent with flow */

	chan->flow = flow;
	chan->next = flow->chanlist;
	flow->chanlist = chan;

	ani_chan_unlock(chan);
	ani_flow_unlock(flow);
}

void
ani_flow_removechannel(ani_flow_t *flow, ani_chan_t *chan)
{
	ani_chan_t *ch, **pch;

	assert_ani_flow_valid(flow);
	assert(chan->flow == flow);

	/*
	 * Flows never return from the dead so we don't lock this access
	 */
	if (flow->flags & FLOW_DYING)
		return;

	ani_flow_lock(flow);
	ani_chan_lock(chan);

	for (pch = &flow->chanlist; (ch = *pch) != 0; pch = &ch->next)
		if (ch == chan) {
			*pch = ch->next;
			break;
		}
	chan->flow = 0;

	ani_chan_unlock(chan);
	ani_flow_unlock(flow);
}

ani_error_t
ani_flow_startthread(ani_flow_t *flow, int inkern,
		     void (*func)(void *, void *), void *farg1, void *farg2)
{
	return ani_threadpool_thread_start(&flow->tpool, inkern,
					   (ani_thread_func_t)func,
					   farg1, farg2, 0);
}

#ifdef DEBUG
#include <stdio.h>

static int
dump_flow(struct hnode *node, void *arg)
{
	ani_flow_t *flow = (ani_flow_t *)node;

	printf("FLOW %p: parent=%p, child=%p, sibling=%p, "
	       "cred=%p, mpool=%p, tpool=%p\n",
	       flow, hnode_parent(node), hnode_child(node), hnode_sibling(node),
	       flow->cred, flow->mpool, &flow->tpool);

	return 0;
}

void
ani_flow_dump(ani_flow_t *flow)
{
	if (flow == 0)
		flow = ani_flow_current();

	hnode_traverse(&flow->me, dump_flow, 0, HNODE_PARENT, 0);
}
#endif
