/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_hnode.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_flow.h>

#include "ani_debug.h"

extern int bootstrapped;

static struct hnode allpools;
static ani_mempool_t *root_mempool = NULL;

void
ani_mempool_setup(ani_mempool_t *pool)
{
	ani_error_t err;

	assert(pool);
	assert(root_mempool == 0);

	err = ani_mempool_init(pool, 0, 0);
	assert(!ani_error_occured(err));
	root_mempool = pool;
}

/*
 * Get the mempool associated with the current thread's flow.
 */
ani_mempool_t *
ani_mempool_current(void)
{
	ani_flow_t *flow;
	ani_mempool_t *p;

	flow = ani_flow_current();
	p = flow ? flow->mpool : root_mempool;

	assert(p);

	return p;
}

/*
 * Return the root mempool
 */
ani_mempool_t *
ani_mempool_rootpool(void)
{
	return root_mempool;
}

/*
 * Create a new mempool with no associated flows.  The parent
 * parameter defines this pool's position in the pool hierarchy.
 * The callback routine is invoked whenever the NodeOS detects that
 * the indicated pool is over its page limit.
 */
ani_error_t
ani_mempool_init(ani_mempool_t *pool, ani_mempool_t *parent,
		 an_mempoolfull_func_t callback)
{
	assert(pool != 0);

	memset(pool, 0, sizeof *pool);

	hnode_init(&pool->me);
	hnode_attach(&pool->me, 0, &allpools);
	if (parent)
		hnode_attach(&pool->me, (struct hnode *)parent, 0);
	pool->physcurrent = pool->physlimit = 0;
	pool->callback = callback;

	return ani_error_clear();
}

/*
 * Destroy a mempool.  There should be no flows associatd with the pool.
 */
ani_error_t
ani_mempool_deinit(ani_mempool_t *pool)
{
	assert(pool);

	if (hnode_child(&pool->me) != 0)
		return ani_error_set(EBUSY,
				     "mempool has attached flows");

	hnode_detach(&pool->me, HNODE_ALL);

	return ani_error_clear();
}

/*
 * Donates the memory resources associated with a flow to the pool.
 */
void
ani_mempool_addflow(ani_mempool_t *pool, ani_flow_t *flow)
{
	assert(pool);
	assert(flow);

	hnode_attach((struct hnode *)flow, 0, &pool->flowlist);

	pool->physlimit += flow->res.ppages;

#if 0
	/*
	 * XXX first attach to the root pool,
	 * debit for memory allocated before this point.
	 */
	if (pool == root_mempool && pool->physcurrent == 0) {
		extern unsigned long earlymempages;

		assert(flow == ani_flow_root_flow());
		assert(earlymempages <= pool->physlimit);

		pool->physcurrent += earlymempages;
	}
#endif
}

/*
 * Removes the memory resources associated with a flow from the pool.
 */
void
ani_mempool_removeflow(ani_mempool_t *pool, ani_flow_t *flow)
{
	hnode_detach((struct hnode *)flow, HNODE_GLOBAL);

	pool->physlimit -= flow->res.ppages;

#if 0
	/*
	 * If we are over the pool limit and it matters, invoke the callback
	 * to free up some memory.  If no callback, kill all the flows!
	 *
	 * XXX we need to establish the proper context for the callback to
	 *     run in (whatever that might be)
	 * XXX we need to do something reasonable if the callback doesn't
	 *     return or doesn't free memory.
	 */
	if (callback == 0)
		;
	while (pool->physcurrent > pool->physlimit)
		pool->callback(pool, pool->physcurrent, pool->physlimit);
#endif
}

/*
 * Simple malloc/free allocation interface
 *
 * XXX if pool == 0, root_mempool does not yet exist or has no resources yet,
 * remember allocations for later.
 *
 * XXX maintain seperate heaps for each pool, charge accordingly
 */

void *
ani_mempool_alloc(ani_mempool_t *pool, size_t size)
{
	assert(pool != 0);

	/* XXX fer now */
	return malloc(size);
}

void
ani_mempool_free(ani_mempool_t *pool, void *ptr, size_t size)
{
	assert(pool != 0);

	/* XXX fer now */
	free(ptr);
}

#ifdef DEBUG
#include <stdio.h>

static int
dump_mempool(struct hnode *node, void *arg)
{
	ani_mempool_t *pool = (ani_mempool_t *)node;

	printf("POOL %p: parent=%p, child=%p, sibling=%p, "
	       "physcur=%d, physlimit=%d\n",
	       pool, hnode_parent(node), hnode_child(node),
	       hnode_sibling(node), pool->physcurrent, pool->physlimit);
	return 0;
}

void
ani_mempool_dump(ani_mempool_t *pool)
{
	if (pool == 0)
		hnode_traverse(&ani_mempool_current()->me, dump_mempool, 0,
			       HNODE_GLOBAL, 0);
	else
		hnode_traverse(&pool->me, dump_mempool, 0, HNODE_PARENT, 0);
}
#endif
