/*
 * Copyright (c) 1997-2000 University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <oskit/dev/osenv.h>
#include <oskit/dev/osenv_mem.h>

/*
 * For the OSKit this is all handled in the device initialization.
 * This functions still has the purpose of ensuring that this file gets
 * brought in from the library, essential if you are overriding OSKit
 * implementations.
 */
void
os_init_net(void)
{
}

#ifndef OSKIT_NEW_NETIOS

/*
 * Bufio implementation for output.
 * Based on com/buf_io.c.
 */
#include <oskit/io/bufio.h>

struct mybufio {
	oskit_bufio_t	ioi;		/* COM I/O Interface */
	unsigned 	count;		/* reference count */
	oskit_size_t	size;		/* size of buffer */
	/* the buffer itself follows... */
};
typedef struct mybufio mybufio_t;

/* 
 * Query a buffer I/O object for it's interfaces.
 *
 */
static OSKIT_COMDECL
bufio_query(oskit_bufio_t *io, const oskit_iid_t *iid, void **out_ihandle) 
{
	struct mybufio *b;

	b = (mybufio_t *)io;
	assert(b != NULL);
	assert(b->count != 0);

	if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
	    memcmp(iid, &oskit_bufio_iid, sizeof(*iid)) == 0) {
		*out_ihandle = &b->ioi;
		++b->count;
		
		return 0;
	}

	*out_ihandle = NULL;
	return OSKIT_E_NOINTERFACE;
}

/*
 * Clone a reference to a device's block I/O interface.
 */
static OSKIT_COMDECL_U
bufio_addref(oskit_bufio_t *io)
{
	struct mybufio *b;

	b = (struct mybufio *)io;
	assert(b != NULL);
	assert(b->count != 0);

	return ++b->count;
}

static OSKIT_COMDECL_U
bufio_release(oskit_bufio_t *io)
{
	struct mybufio *b;
	char iszero;

	b = (struct mybufio *)io;

	assert(b != NULL);
	assert(b->count != 0);

	/*
	 * Atomic decrement and read of the counter.
	 *
	 * Needed since this will be called by the network transmit
	 * done interrupt handler.
	 *
	 * Note that we don't return an accurate count to the caller,
	 * but we know that the caller doesn't really care beyond
	 * zero or non-zero.
	 */
#ifdef i386
	__asm__ __volatile__("decl %0; sete %1"
			     : "=m" (b->count), "=qm" (iszero)
			     : "m" (b->count));
#else
	assert(osenv_intr_enabled());
	osenv_intr_disable();
	iszero = (--b->count == 0);
	osenv_intr_enable();
#endif
	if (iszero) {
		osenv_mem_free(b, OSENV_PHYS_WIRED, sizeof *b + b->size);
		return 0;
	}
	return 1;
}

/*
 * Return the block size of this blkio object - always 1.
 */
static OSKIT_COMDECL_U
bufio_getblocksize(oskit_bufio_t *io)
{
	return 1;
}

/*
 * Copy data from a user buffer into kernel's address space.
 */
static OSKIT_COMDECL
bufio_read(oskit_bufio_t *io, void *dest, oskit_off_t offset,
	   oskit_size_t count, oskit_size_t *out_actual)
{
	struct mybufio *b;

	b = (struct mybufio *)io;

	assert(b != NULL);
	assert(b->count != 0);

	if (offset >= b->size)
		return OSKIT_EINVAL;
	if (offset + count > b->size)
		count = b->size - offset;

	memcpy(dest, (void *)(b+1) + offset, count);

	*out_actual = count;
	return 0;
}

/*
 * Copy data from kernel address space to a user buffer.
 */
static OSKIT_COMDECL
bufio_write(oskit_bufio_t *io, const void *src, oskit_off_t offset,
	    oskit_size_t count, oskit_size_t *out_actual)
{
	struct mybufio *b;

	b = (struct mybufio *)io;

	assert(b != NULL);
	assert(b->count != 0);

	if (offset >= b->size)
		return OSKIT_EINVAL;
	if (offset + count > b->size)
		count = b->size - offset;

	memcpy((void *)(b+1) + offset, src, count);

	*out_actual = count;
	return 0;
}

/*
 * Return the size of this buffer.
 */
static OSKIT_COMDECL
bufio_getsize(oskit_bufio_t *io, oskit_off_t *out_size)
{
	struct mybufio *b;

	b = (struct mybufio *)io;
	assert(b != NULL);
	assert(b->count != 0);

	*out_size = b->size;
	return 0;
}

static OSKIT_COMDECL
bufio_setsize(oskit_bufio_t *io, oskit_off_t size)
{
	struct mybufio *b;

	b = (struct mybufio *)io;

	assert(b != NULL);
	assert(b->count != 0);

	return OSKIT_E_NOTIMPL;
}


static OSKIT_COMDECL
bufio_map(oskit_bufio_t *io, void **out_addr,
	  oskit_off_t offset, oskit_size_t count)
{
	struct mybufio *b;

	b = (struct mybufio *)io;

	assert(b != NULL);
	assert(b->count != 0);

	if (offset + count > b->size)
		return OSKIT_EINVAL;

	*out_addr = (void*)(b+1) + offset; 

	return 0;
}

static OSKIT_COMDECL
bufio_unmap(oskit_bufio_t *io, void *addr,
	    oskit_off_t offset, oskit_size_t count)
{
	struct mybufio *b;

	b = (struct mybufio *)io;

	assert(b != NULL);
	assert(b->count != 0);

	if (offset + count > b->size)
		return OSKIT_EINVAL;

	return 0;
}


static OSKIT_COMDECL 
bufio_wire(oskit_bufio_t *io, oskit_addr_t *out_physaddr,
	   oskit_off_t offset, oskit_size_t count)
{
	return OSKIT_E_NOTIMPL;
}

static OSKIT_COMDECL 
bufio_unwire(oskit_bufio_t *io, oskit_addr_t phys_addr,
	     oskit_off_t offset, oskit_size_t count)
{
	return OSKIT_E_NOTIMPL;
}

static OSKIT_COMDECL
bufio_copy(oskit_bufio_t *io, oskit_off_t offset, oskit_size_t count,
	   oskit_bufio_t **out_io)
{
	return OSKIT_E_NOTIMPL;
}


static struct oskit_bufio_ops bio_ops = {
	bufio_query, bufio_addref, bufio_release,
	bufio_getblocksize,
	bufio_read, bufio_write,
	bufio_getsize, bufio_setsize,
	bufio_map, bufio_unmap,
	bufio_wire, bufio_unwire,
	bufio_copy
};


oskit_bufio_t *
oskit_bufio_create(oskit_size_t size) 
{
	mybufio_t *b;

	/* XXX: must be wired memory for DMA drivers */

	b = osenv_mem_alloc(sizeof *b + size, OSENV_PHYS_WIRED, 0);
	if (b == 0)
		return 0;

	b->ioi.ops = &bio_ops;
	b->count = 1;
	b->size = size;
	
	return &b->ioi;
}
#endif

/*
 * XXX hack Linux interface to intercept skbuff allocations.
 */
#include <nodeos/impl/an_impl.h>

#define MAXFREEBUFS	64

#ifdef OSKIT_SKBUFIOS
#include <oskit/io/skbufio.h>

/*
 * XXX value is magic from OSKIT skbuff implmentation.
 * Ensures that skb->truesize winds up as NODEOS_CHAN_BUFSIZE.
 */
#define HDRSIZE	((sizeof(struct sk_buff)+7)&~7)
#else
struct sk_buff {
	struct sk_buff *next;
};
#endif

static int freecount;
static struct sk_buff *freelist;
static struct {
	unsigned int allocs;
	unsigned int alloc_empty;
	unsigned int frees;
	unsigned int free_full;
	unsigned int max_free;
} skb_stats;

void *
oskit_linux_skbmem_alloc(unsigned int hdrsize, unsigned int size, int flags,
			 unsigned int *out_size)
{
	struct sk_buff *skb;
	size_t bufsize;

	assert(flags == OSENV_NONBLOCKING);

#ifdef OSKIT_SKBUFIOS
	/*
	 * Skbufios are used for all packet buffers, not just incoming
	 * and outgoing ethernet packets.  In the case where a user creates
	 * a packet buffer, the data size (size) seen here will be CHAN_BUFSIZE
	 * so we need to allocate a slightly larger buffer (since the default
	 * skbufio implementation requests space for the skbuff struct itself
	 * as well (hdrsize)).
	 */
	assert(hdrsize == HDRSIZE);
	assert(size <= NODEOS_CHAN_BUFSIZE);
	bufsize = NODEOS_CHAN_BUFSIZE + HDRSIZE;
#else
	assert(hdrsize + size <= NODEOS_CHAN_BUFSIZE);
	bufsize = NODEOS_CHAN_BUFSIZE;
#endif

	skb_stats.allocs++;

	skb = freelist;
	if (skb == NULL) {
		assert(freecount == 0);
		skb = osenv_mem_alloc(bufsize, OSENV_PHYS_WIRED|flags, 0);
		skb_stats.alloc_empty++;
	} else {
		assert(freecount > 0);
		freelist = skb->next;
		skb->next = 0;
		freecount--;
	}

	if (skb)
		*out_size = bufsize;
	return skb;
}

void
oskit_linux_skbmem_free(void *ptr, unsigned int size)
{
	struct sk_buff *skb = ptr;

#ifdef OSKIT_SKBUFIOS
	assert(size == NODEOS_CHAN_BUFSIZE + HDRSIZE);
#else
	assert(size == NODEOS_CHAN_BUFSIZE);
#endif

	skb_stats.frees++;

	if (freecount < MAXFREEBUFS) {
		skb->next = freelist;
		freelist = skb;
		freecount++;
		if (freecount > skb_stats.max_free)
			skb_stats.max_free = freecount;
	} else {
		osenv_mem_free(ptr, OSENV_PHYS_WIRED, size);
		skb_stats.free_full++;
	}
}

#if 1
#include <stdio.h>

void
oskit_linux_skbmem_dump(void)
{
	printf("sk_buff freelist: %d allocations, %d frees\n",
	       skb_stats.allocs, skb_stats.frees);
	printf("  %d empty on alloc, %d full on free, %d max entries\n",
	       skb_stats.alloc_empty, skb_stats.free_full, skb_stats.max_free);
}
#endif
