/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <sys/types.h>
#ifdef OSKIT
#include <oskit/io/netio.h>
#include <oskit/io/bufio.h>
#endif

#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_pbuf.h>

/*
 * Packet buffer manipulation
 */

#ifndef OSKIT_SKBUFIOS
/*
 * XXX sigh...even though we now control Linux skbuff allocation,
 * it still isn't enough.  The bufio interface doesn't have a buffer
 * size independent of the data size.  When we receive a bufio via the
 * push method, we can only discover its current size, not its max size.
 * At first glance that seems ok, since we know the underlying buffer
 * has to be NODEOS_CHAN_BUFSIZE since we allocated it.  But there is
 * no way to "reset" a bufio to its maximum size (before we put it back
 * on a channel's pbuf list) so we cannot use the bufio interfaces to
 * access the complete buffer even though we know it should be ok.
 * Until we come up with a new oskit netio buffer interface, we are still
 * stuck copying all but the max sized packet.
 */
#undef NODEOS_CHAN_COPYSIZE
#define NODEOS_CHAN_COPYSIZE	1513
#endif

#ifdef OSKIT_SKBUFIOS
static void
pbuffer_skbufiofree(ani_pbuf_t *pbuf, void *arg)
{
	oskit_skbufio_metadata_unmap(pbuf->skbufio, pbuf->skb);
	oskit_skbufio_release(pbuf->skbufio);
	pbuf->skbufio = 0;
}
#else
/*
 * Allocate system-wide packet buffer memory.
 * Right now we just malloc every time.  Someday we might manage
 * a cache of memory.
 *
 * @allocation: "size" bytes charged to the given flow.
 */
static void *
pbuffer_memalloc(ani_flow_t *flow, size_t size)
{
	assert_ani_flow_valid(flow);
	assert(size > 0);

	/* XXX round up size or verify it is a valid size */

	return ani_mempool_alloc(ani_flow_mempool(flow), size);
}

/*
 * Free system-wide packet buffer memory.
 * Called with a pbuf when its reference count drops to zero.
 * Frees just the buffer, not the header.
 *
 * @deallocation: pbuf->bufsize bytes returned to pbuffer's flow.
 */
static void
pbuffer_memfree(ani_pbuf_t *pbuf, void *arg)
{
	assert(pbuf != 0);
	assert_ani_flow_valid(pbuf->flow);
	assert(pbuf->bufdata != 0);
	assert(pbuf->bufsize > 0);

	ani_mempool_free(ani_flow_mempool(pbuf->flow),
			 pbuf->bufdata, pbuf->bufsize);
	pbuf->bufdata = 0;
}
#endif

/*
 * Initialize the given pbuf with the indicated info.
 */
void
ani_pbuf_init_external(ani_pbuf_t *pbuf, ani_flow_t *flow, ani_chan_t *chan,
#ifdef OSKIT_SKBUFIOS
		       struct sk_buff *skb)
#else
		       void *bdata, size_t bsize, size_t off, size_t size,
		       ani_pbuf_free_t ffunc, void *farg)
#endif
{
	assert(pbuf != 0);
	assert_ani_flow_valid(flow);

#ifdef OSKIT_SKBUFIOS
	assert(skb != 0);

	pbuf->skbufio = skb_skbufio(skb);
	pbuf->skb = skb;
	pbuf->ffunc = pbuffer_skbufiofree;
	pbuf->farg = 0;
#else
	assert(ffunc != 0);
	assert(bdata != 0);
	assert(off >= 0);
	assert(off + size <= bsize);

	pbuf->bufdata = bdata;
	pbuf->bufsize = bsize;
	pbuf->off = off;
	pbuf->size = size;
	pbuf->ffunc = ffunc;
	pbuf->farg = farg;
#endif
	pbuf->refcount = 1;
	pbuf->flow = flow;
	pbuf->chan = chan;
}

/*
 * Allocate a buffer and associate it with the given pbuf object.
 */
ani_error_t
ani_pbuf_init(ani_pbuf_t *pbuf, ani_flow_t *flow, ani_chan_t *chan,
	      size_t bsize, size_t off, size_t size)
{
#ifdef OSKIT_SKBUFIOS
	oskit_skbufio_t *skbufio;
	struct sk_buff *skb;
	oskit_error_t err;

	assert(off + size <= bsize);

	err = oskit_skbufio_create(size, off, bsize-(off+size), &skbufio);
	if (err != 0)
		return ani_error_set(err, "oskit_skbufio_create failed");
	err = oskit_skbufio_metadata_map(skbufio, SKB_VERSION_NUMBER, &skb);
	assert(err == 0);

	ani_pbuf_init_external(pbuf, flow, chan, skb);
#else
	void *bdata;

	bdata = pbuffer_memalloc(flow, bsize);
	if (bdata == 0) {
		ani_mempool_free(ani_flow_mempool(flow), pbuf, sizeof *pbuf);
		return ani_error_set(ENOMEM, "no memory for pbuffer");
	}

	ani_pbuf_init_external(pbuf, flow, chan, bdata, bsize, off, size,
			       pbuffer_memfree, 0);
#endif

	return ani_error_clear();
}

void
ani_pbuf_deinit(ani_pbuf_t *pbuf)
{
	assert(pbuf != 0);

	if (ani_pbuf_isvalid(pbuf))
		ani_pbuf_release(pbuf);
}

/*
 * Create a new pbuffer referencing the same data buffer.
 */
ani_error_t
ani_pbuf_clone(ani_pbuf_t *pbuf, ani_pbuf_t *newpbuf)
{
	ani_error_t err;

	assert(pbuf != 0 && newpbuf != 0);
	assert_ani_flow_valid(pbuf->flow);
	assert(pbuf->chan == 0);

#ifdef OSKIT_SKBUFIOS
	assert(pbuf->skb != 0);
	oskit_skbufio_addref(pbuf->skbufio);
	ani_pbuf_init_external(newpbuf, pbuf->flow, 0, pbuf->skb);
	err = ani_error_clear();
#else
	err = ani_error_set(AN_ERR_NOTSUPPORTED,
			    "ani_pbuf_clone hasn't been written yet.");
#endif

	return err;
}

/*
 * Set the offset of the active region of the pbuffer.
 * Implicitly changes the size of the active region of the buffer.
 */
ani_error_t
ani_pbuf_setoffset(ani_pbuf_t *pbuf, size_t newoff, void **out_data)
{
	size_t offset, eoffset;

	assert(pbuf != 0);
	assert(ani_pbuf_isvalid(pbuf));

	offset = ani_pbuf_offset(pbuf);
	if (newoff != offset) {
		eoffset = offset + ani_pbuf_size(pbuf);
		if (newoff < 0 || newoff > eoffset)
			return ani_error_set(EINVAL,
					     "invalid pbuf offset specified");

#ifdef OSKIT_SKBUFIOS
		newoff -= offset;
		if (newoff < 0)
			skb_push(pbuf->skb, -newoff);
		else
			skb_pull(pbuf->skb, newoff);
#else
		pbuf->off = newoff;
		pbuf->size = eoffset - newoff;
#endif
	}

	if (out_data)
		*out_data = ani_pbuf_data(pbuf);
	return ani_error_clear();
}

/*
 * Set the size of the active region of the pbuffer.
 */
ani_error_t
ani_pbuf_setsize(ani_pbuf_t *pbuf, size_t newsize, void **out_data)
{
	size_t size;

	assert(pbuf != 0);
	assert(ani_pbuf_isvalid(pbuf));

	size = ani_pbuf_size(pbuf);
	if (newsize != size) {
		if (newsize < 0 ||
		    ani_pbuf_offset(pbuf) + newsize > ani_pbuf_bufsize(pbuf))
			return ani_error_set(EINVAL,
					     "invalid pbuf size specified");

#ifdef OSKIT_SKBUFIOS
		newsize -= size;
		if (newsize > 0)
			skb_put(pbuf->skb, newsize);
		else
			skb_get(pbuf->skb, -newsize);
#else
		pbuf->size = newsize;
#endif
	}

	if (out_data)
		*out_data = ani_pbuf_data(pbuf);

	return 0;
}

/*
 * Open a buffer of to its max size.
 * Returns a pointer to the active region.
 */
void *
ani_pbuf_setmax(ani_pbuf_t *pbuf)
{
	assert(pbuf != 0);
	assert(ani_pbuf_isvalid(pbuf));

#ifdef OSKIT_SKBUFIOS
	/* XXX naked sk_buff field refs */
	pbuf->skb->data = pbuf->skb->head;
	pbuf->skb->tail = pbuf->skb->end;
	pbuf->skb->len = pbuf->skb->truesize;
#else
	pbuf->off = 0;
	pbuf->size = pbuf->bufsize;
#endif

	return ani_pbuf_data(pbuf);
}

/*
 * Swap the memory buffers of two buffer objects.
 */
void
ani_pbuf_swap(ani_pbuf_t *pbuf1, ani_pbuf_t *pbuf2)
{
	ani_pbuf_t tmp;

	assert(pbuf1->refcount == 1);
	assert(pbuf2->refcount == 1);
	assert(ani_pbuf_bufsize(pbuf1) == ani_pbuf_bufsize(pbuf2));

	tmp = *pbuf1;
	*pbuf1 = *pbuf2;

	/*
	 * Retain flow and channel association
	 * XXX need to adjust accounting if buffer sizes don't match.
	 */
	pbuf1->flow = tmp.flow;
	pbuf1->chan = tmp.chan;
	tmp.flow = pbuf2->flow;
	tmp.chan = pbuf2->chan;

	*pbuf2 = tmp;
}

#ifdef OSKIT
#ifndef OSKIT_SKBUFIOS
static void
ani_pbuf_bufiofree(ani_pbuf_t *pbuf, void *arg)
{
	oskit_bufio_t *bufio = arg;

	assert(bufio != 0);
	assert(pbuf->bufdata != 0);

	oskit_bufio_unmap(bufio, pbuf->bufdata, 0, pbuf->bufsize);
	oskit_bufio_release(bufio);
	pbuf->bufdata = 0;
}
#endif

/*
 * Wrap an OSKit bufio with a NodeOS packet buffer.
 * The caller should have a reference to the bufio.  This routine will
 * take a bufio reference if successful.
 */
ani_error_t
ani_pbuf_frombufio(ani_flow_t *flow, ani_chan_t *chan, oskit_bufio_t *bufio,
		   size_t off, size_t size, ani_pbuf_t **out_pbuf)
{
	size_t bsize;
	ani_pbuf_t *pbuf;
	ani_error_t err;
	int ierr;
#ifdef OSKIT_SKBUFIOS
	oskit_skbufio_t *skbufio;
	struct sk_buff *skb;
#else
	void *bdata;
	oskit_off_t bs;
#endif

	assert_ani_flow_valid(flow);
	assert(bufio != 0);
	assert(off >= 0);

#ifdef OSKIT_SKBUFIOS
	/*
	 * Extract the skbufio interface and map the skbuff meta info.
	 * This call grabs a reference to the skbufio.
	 */
	ierr = oskit_bufio_query(bufio, &oskit_skbufio_iid, (void **)&skbufio);
	if (ierr != 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return ani_error_set(EINVAL, "oskit_bufio_query failed");
	}
	ierr = oskit_skbufio_metadata_map(skbufio, SKB_VERSION_NUMBER, &skb);
	if (ierr != 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		oskit_skbufio_release(skbufio);
		return ani_error_set(EINVAL,
				     "oskit_skbufio_metadata_map failed");
	}
	bsize = skb->truesize;
	assert(bsize == NODEOS_CHAN_BUFSIZE);

	/*
	 * Get a large enough buffer from the channel.
	 * If it can't pay, we drop the packet.
	 */
	err = ani_chan_pbuf_alloc(chan, bsize, &pbuf);
	if (err) {
		oskit_skbufio_metadata_unmap(skbufio, skb);
		oskit_skbufio_release(skbufio);
		return err;
	}
	assert(pbuf->flow == flow);

	/*
	 * XXX free the actual data buffer as we cannot return it
	 * to the OSKit driver to use right now.
	 */
	ani_pbuf_release(pbuf);

	/*
	 * Pop the header if required and initialize the pbuf with the
	 * new buffer info.
	 */
	if (off != 0)
		skb_pull(skb, off);
	ani_pbuf_init_external(pbuf, flow, chan, skb);
#else
	/*
	 * Determine the size of the incoming packet and map it
	 */
	ierr = oskit_bufio_getsize(bufio, &bs);
	if (ierr != 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return ani_error_set(ierr, "oskit_bufio_getsize failed");
	}
	bsize = (size_t)bs;
	if (off + size > bsize) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return ani_error_set(EINVAL, "pbuf too big for bufio");
	}
	ierr = oskit_bufio_map(bufio, &bdata, 0, bsize);
	if (ierr != 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return ani_error_set(ierr, "oskit_bufio_map failed");
	}

	/*
	 * Get a large enough buffer from the channel.
	 * If it can't pay, we drop the packet.
	 */
	err = ani_chan_pbuf_alloc(chan, bsize, &pbuf);
	if (ani_error_occured(err)) {
		oskit_bufio_unmap(bufio, bdata, 0, bsize);
		return err;
	}
	assert(pbuf->flow == flow);

	/*
	 * We copy small packets for two reasons, 1) because its possibly
	 * more efficient, 2) because we don't want to be swapping buffers
	 * that are too small into our packet buffer pool.
	 */
	if (bsize <= NODEOS_CHAN_COPYSIZE) {
		memcpy(ani_pbuf_data(pbuf), bdata+off, size);
		CHANSTATS(chan->stats.inbufcopies++);
	}
	/*
	 * Packet is large enough to swap
	 */
	else {
		/*
		 * XXX free the actual data buffer as we cannot return it
		 * to the OSKit driver to use right now.
		 */
		ani_pbuf_release(pbuf);
		/*
		 * Intialize the pbuf with the new buffer info.
		 */
		ani_pbuf_init_external(pbuf, flow, chan, bdata, bsize, off,
				       size, ani_pbuf_bufiofree, bufio);
		/*
		 * Keep the bufio alive
		 */
		oskit_bufio_addref(bufio);
	}
#endif

	*out_pbuf = pbuf;
	return ani_error_clear();
}

ani_error_t
ani_pbuf_tobufio(ani_pbuf_t *pbuf, void *hdr, size_t hsize, oskit_netio_t *nio,
		 oskit_bufio_t **out_bufio, int *didcopy)
{
	oskit_bufio_t *bufio;
	oskit_size_t psize;
	int ierr;

#ifdef OSKIT_SKBUFIOS
	oskit_skbufio_t *skbufio;
	struct sk_buff *skb;
	assert(pbuf->skb != 0);

	/*
	 * Make sure there is room for the header in the skbuff.
	 * If not, we'll have to create a new skbuff.
	 */
	skb = pbuf->skb;
	if (skb_headroom(skb) >= hsize) {
		if (hsize != 0)
			memcpy(skb_push(skb, hsize), hdr, hsize);
		bufio = (oskit_bufio_t *)pbuf->skbufio;
		oskit_bufio_addref(bufio);
		*out_bufio = bufio;
		*didcopy = 0;
		return ani_error_clear();
	}

	/*
	 * Create a new skbufio and copy the data.
	 * XXX should be a netio function to return an skbufio?
	 */
	psize = ani_pbuf_size(pbuf);
	assert(hsize + psize <= NODEOS_CHAN_BUFSIZE);
	ierr = oskit_skbufio_create(0, 0, NODEOS_CHAN_BUFSIZE, &skbufio);
	if (ierr != 0)
		return ani_error_set(ierr, "oskit_skbufio_create failed");
	ierr = oskit_skbufio_metadata_map(skbufio, SKB_VERSION_NUMBER, &skb);
	if (ierr != 0) {
		oskit_skbufio_release(skbufio);
		return ani_error_set(ierr, "oskit_skbufio_metadata map failed");
	}
	if (hsize)
		memcpy(skb_put(skb, hsize), hdr, hsize);
	memcpy(skb_put(skb, psize), ani_pbuf_data(pbuf), psize);
	oskit_skbufio_metadata_unmap(skbufio, skb);
	bufio = (oskit_bufio_t *)skbufio;
#else
	void *bdata;

	/*
	 * Packet buffer is already a wrapped bufio,
	 *
	 * We can just use the existing bufio if, after pre-pending any
	 * header, the pbuf offset would be zero.  Note that we do not
	 * actually modify the pbuf offsets, we just copy the header into
	 * the buffer before the official start of data.
	 */
	if (pbuf->ffunc == ani_pbuf_bufiofree && pbuf->off - hsize == 0) {
		if (hsize != 0)
			memcpy(pbuf->bufdata, hdr, hsize);
		bufio = pbuf->farg;
		oskit_bufio_addref(bufio);
		*out_bufio = bufio;
		*didcopy = 0;
		return 0;
	}

	/*
	 * Otherwise cons up a bufio and copy the header/data.
	 */
	psize = ani_pbuf_size(pbuf);
#ifdef OSKIT_NEW_NETIOS
	ierr = oskit_netio_alloc_bufio(nio, hsize + psize, &bufio);
	assert(ierr != OSKIT_E_NOTIMPL);
	if (ierr != 0)
		return ani_error_set(ierr, "oskit_netio_alloc_bufio failed");
#else
	bufio = oskit_bufio_create(hsize + psize);
	if (bufio == 0)
		return ani_error_set(ENOMEM, "no memory for bufio");
#endif

	/*
	 * Copy the header/data.
	 */
	ierr = oskit_bufio_map(bufio, &bdata, 0, hsize + psize);
	if (ierr == 0) {
		if (hsize)
			memcpy(bdata, hdr, hsize);
		memcpy(bdata+hsize, ani_pbuf_data(pbuf), psize);
		oskit_bufio_unmap(bufio, bdata, 0, hsize + psize);
	} else {
		oskit_size_t off = 0;

		if (hsize) {
			ierr = oskit_bufio_write(bufio, hdr, 0, hsize, &off);
			if (ierr != 0 || off != hsize) {
				oskit_bufio_release(bufio);
				return ani_error_set(ierr,
						"oskit_bufio_write hdr failed");
			}
		}
		ierr = oskit_bufio_write(bufio, ani_pbuf_data(pbuf), off,
					 psize, &off);
		if (ierr != 0 || off != psize) {
			oskit_bufio_release(bufio);
			return ani_error_set(ierr,
					     "oskit_bufio_write data failed");
		}
	}
#endif

	*out_bufio = bufio;
	*didcopy = 1;
	return ani_error_clear();
}

#else	/* OSKIT */

ani_error_t
ani_pbuf_tomem(ani_pbuf_t *pbuf, void *hdr, size_t hsize,
	       void **out_buf, size_t *out_size, int *didcopy)
{
	void *buf;

	/*
	 * See if we can prepend the header to the pbuf.
	 * As in the tobufio case above, we don't modify the pbuf offset.
	 */
	if (pbuf->off >= hsize) {
		buf = ani_pbuf_data(pbuf);
		if (hsize != 0) {
			buf -= hsize;
			memcpy(buf, hdr, hsize);
		}

		*out_buf = buf;
		*out_size = ani_pbuf_size(pbuf) + hsize;
		*didcopy = 0;
		return ani_error_clear();
	}

	/*
	 * Otherwise, copy the header and packet to a contiguous buffer.
	 */
	buf = ani_mempool_alloc(ani_flow_mempool(ani_flow_current()),
				ani_pbuf_size(pbuf) + hsize);
	if (buf == 0)
		return ani_error_set(ENOMEM, "no memory for pbuf");

	memcpy(buf, hdr, hsize);
	memcpy(buf+hsize, ani_pbuf_data(pbuf), ani_pbuf_size(pbuf));

	*out_buf = buf;
	*out_size = ani_pbuf_size(pbuf) + hsize;
	*didcopy = 1;
	return ani_error_clear();
}

void
ani_pbuf_relmem(void *buf, size_t bufsize)
{
	ani_mempool_free(ani_flow_mempool(ani_flow_current()), buf, bufsize);
}
#endif	/* OSKIT */

