/*
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Stride Packet Scheduler.
 * 
 * Based on PHD thesis by Carl A. Waldspurger.
 *
 * Carl A. Waldspurger. Lottery and Stride Scheduling: Flexible Proportional
 * Share Resource Management, Ph.D. dissertation, Massachusetts Institute
 * of Technology, September 1995. Also appears as Technical Report
 * MIT/LCS/TR-667. See http://www.waldspurger.org/carl/papers.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include "pq_impl.h"
#include "../os.h"

#define LOCK(pq)	ani_ilock_lock(&(pq)->lock)
#define UNLOCK(pq)	ani_ilock_unlock(&(pq)->lock)

/*
 * Each output device has its own scheduler. Maintain a list of them
 * for the watchdog.
 */
static pq_impl_t	*pq_schedulers;

/*
 * Need this to protect the above global.
 */
static ani_ilock_t	pqlock;

/*
 * Timer to periodically keep packets moving.
 * Also acts as the initialized flag.
 */
static void		*pqtimer;


static int
pq_create_default(pq_impl_t **out_pq)
{
	pq_impl_t	*pq;

	if ((pq = (pq_impl_t *) malloc(sizeof(pq_impl_t))) == NULL)
		return ENOMEM;
	memset(pq, 0, sizeof(*pq));

	pq->refcount = 1;
	queue_init(&pq->clients);
	ani_ilock_init(&pq->lock);

	*out_pq = pq;
	return 0;
}

static void
pq_destroy(pq_impl_t *pq)
{
	free(pq);
}

/*
 * Initialize entry state for a client. 
 */
static void
pq_client_init(pq_impl_t *pq, int tickets)
{
	pq->tickets = tickets;
	pq->stride  = STRIDE / tickets;
	pq->remain  = pq->stride;
}

/*
 * Update global pass based on number of bytes transferred.
 */
static inline void
pq_global_pass_update(pq_impl_t *pq)
{
	int	transferred;

	transferred           = pq->bytes_transferred;
	pq->bytes_transferred = 0;

	pq->pass += (pq->stride * transferred) / QUANTUM;
}

/*
 * Compute new tickets, stride when client enters/leaves.
 */
static inline void
pq_global_tickets_update(pq_impl_t *pq, int delta)
{
	pq->tickets += delta;
	if (pq->tickets)
		pq->stride = STRIDE / pq->tickets;
	else
		pq->stride = STRIDE;
}

/*
 * A leaf is added to the group (now has packets to send).
 */
static void
pq_client_join(pq_impl_t *pq)
{
	pq_global_pass_update(pq->schedlink);
	
	/*
	 * New pass
	 */
	pq->pass = pq->schedlink->pass + pq->remain;

	pq_global_tickets_update(pq->schedlink, pq->tickets);

	/*
	 * Insert into unordered queue.
	 */
	queue_enter(&pq->schedlink->clients, pq, pq_impl_t *, clients);
}

/*
 * A leaf is removed from the group (no longer has packets to send).
 */
static void
pq_client_leave(pq_impl_t *pq)
{
	pq_global_pass_update(pq->schedlink);
	
	/*
	 * New remain
	 */
	pq->remain = pq->pass - pq->schedlink->pass;

	pq_global_tickets_update(pq->schedlink, -(pq->tickets));

	/*
	 * Remove from queue.
	 */
	queue_remove(&pq->schedlink->clients, pq, pq_impl_t *, clients);
}

/*
 * Try to send a packet, either as the result of a client trying to send
 * packet, or because the timer kicked off and is trying push out the
 * backlog.
 */
static int
pq_send_next(pq_impl_t *pq)
{
	pq_impl_t	*pnext = (pq_impl_t *) queue_first(&(pq->clients));
	pq_impl_t	*ptmp;
	void		*p;
	int		psize;
	int		err, sent = 0;

	/*
	 * This needs to be a proper priority queue!
	 */
	queue_iterate(&(pq->clients), ptmp, pq_impl_t *, clients) {
		if (ptmp->pass < pnext->pass)
			pnext = ptmp;
	}
	assert(! pqueue_empty(pnext->pqueue));

	/*
	 * Got one!
	 */
	pqueue_first(pnext->pqueue, &p, &psize);

#ifdef  OSKIT
	err = oskit_netio_push(pq->netio, (oskit_bufio_t *) p, psize);
	if (!err) {
		sent = 1;
		oskit_bufio_release((oskit_bufio_t *) p);
	}
#endif
	if (sent) {
		DOSTAT(pq->sent++);
		pqueue_advance(pnext->pqueue);
		pq->bytes_transferred += psize;

		/*
		 * And update the pass ...
		 */
		pnext->pass += (pnext->stride * psize) / QUANTUM;
	}
	else
		DOSTAT(pq->stalls++);
	
	/*
	 * If the client queue is now empty (no more packets to send), then
	 * it needs to leave the field.
	 */
	if (pqueue_empty(pnext->pqueue)) {
		pq_client_leave(pnext);
		pnext->ready = 0;
	}

	return sent;
}

/*
 * Timer to periodically look for stalled packets.
 *
 * Go through and see if packets need to be sent. We need this because
 * we do not have an interface that tells us when we can send packets.
 */
static void
pq_timeout_handler(void *arg0, void *arg1)
{
	pq_impl_t	*pq;

	ani_ilock_lock(&pqlock);

	if ((pq = pq_schedulers) == NULL) {
		ani_ilock_unlock(&pqlock);
		return;
	}
	
	while (pq) {
		int		i;
		pq_impl_t	*pqnext;

		LOCK(pq);
		for (i = 0; i < pq->burstlen; i++) {
			if (queue_empty(&(pq->clients)))
				break;

			if (! pq_send_next(pq))
				break;
		}
		pqnext = pq->schedlink;
		UNLOCK(pq);
		pq = pqnext;
	}
	ani_ilock_unlock(&pqlock);
}

/*
 * Public Interface.
 */
int
pq_create_scheduler(oskit_netio_t *netio,
		    int maxburst, pq_scheduler_t **out_pq)
{
	pq_impl_t	*pq;
	int		err;
	static int	inited;

	if ((err = pq_create_default(&pq)))
		return err;

	pq->burstlen = (maxburst == 0) ? DEFAULT_BURST_LENGTH : maxburst;

#ifdef  OSKIT
	oskit_netio_addref(netio);
	pq->netio    = netio;
#endif

	if (! pqtimer) {
		unsigned long timerres;

		/*
		 * Initialize this first, since the handler uses it!
		 */
		ani_ilock_init(&pqlock);
		
		/*
		 * Start the timer ...
		 */
		pqtimer = os_timer_init(pq_timeout_handler, 0, &timerres);
		os_timer_start(pqtimer, timerres);

		inited = 1;
	}

	/*
	 * Link into list of device schedulers.
	 */
	ani_ilock_lock(&pqlock);
	pq->schedlink = pq_schedulers;
	pq_schedulers = pq;
	ani_ilock_unlock(&pqlock);

	*out_pq = (pq_scheduler_t *) pq;
	return 0;
}

int
pq_destroy_scheduler(pq_scheduler_t *node)
{
	pq_impl_t	*pq = (pq_impl_t *) node;
	pq_impl_t	**p, *temp;

	LOCK(pq);
	if (pq->refcount != 1) {
		UNLOCK(pq);
		return EINVAL;
	}

#ifdef  STATS
	printf("Destroying PQ Scheduler: sent %d, dropped %d, stalls %d\n",
	       pq->sent, pq->drops, pq->stalls);
#endif

	/*
	 * Unlock to avoid deadlock with pq_timeout_handler()
	 */
	UNLOCK(pq);
	
	/*
	 * Unlink from list of device schedulers.
	 */
	ani_ilock_lock(&pqlock);
	p = &pq_schedulers;
	while (((temp = *p) != NULL) && temp != pq)
		p = &temp->schedlink;
	assert(temp);
	*p = temp->schedlink;
	
	/*
	 * Last scheduler. Kill off the timer.
	 */
	if (!pq_schedulers) {
		os_timer_deinit(pqtimer);
		pqtimer = 0;
	}
	ani_ilock_unlock(&pqlock);

#ifdef  OSKIT
	oskit_netio_release(pq->netio);
#endif
	pq_destroy(pq);
	return 0;
}

int
pq_create_client(pq_scheduler_t *scheduler,
		 int maxqsize, int tickets, pq_client_t **out_pq)
{
	pq_impl_t	*pq;
	pq_impl_t	*pq_sched = (pq_impl_t *) scheduler;
	int		err;

	if (tickets > MAXTICKETS)
		return EINVAL;
	
	if ((err = pq_create_default(&pq)))
		return err;

	if (!maxqsize)
		maxqsize = DEFAULT_PACKET_QUEUE_LENGTH;

	if ((pq->pqueue = pqueue_alloc(maxqsize)) == NULL) {
		pq_destroy(pq);
		return ENOMEM;
	}

	pq_client_init(pq, tickets);
	LOCK(pq_sched);
	pq->schedlink = pq_sched;
	pq->schedlink->refcount++;
	UNLOCK(pq_sched);

	*out_pq = (pq_client_t *) pq;
	return 0;
}

void
pq_destroy_client(pq_client_t *node)
{
	pq_impl_t	*pq = (pq_impl_t *) node;

	/*
	 * Kill of the scheduler connection first.
	 */
	LOCK(pq->schedlink);
	pq_client_leave(pq);
	pq->schedlink->refcount--;
	UNLOCK(pq->schedlink);

#ifdef  STATS
	printf("Destroying PQ Client: sent %d, dropped %d, stalls %d\n",
	       pq->sent, pq->drops, pq->stalls);
#endif

	if (pq->ready) {
		void		*p;
		int		size;
		
		/*
		 * Have to kill off the packets.
		 */
		assert(! pqueue_empty(pq->pqueue));
		while (! pqueue_empty(pq->pqueue)) {
			pqueue_dequeue(pq->pqueue, &p, &size);
#ifdef	OSKIT
			oskit_bufio_release((oskit_bufio_t *) p);
#endif
		}
	}
	pqueue_free(pq->pqueue);
	pq_destroy(pq);
}

/*
 * Change a ticket allocation. Return old tickets if requested.
 */
void
pq_modify_tickets(pq_client_t *node, int tickets, int *old_tickets)
{
	pq_impl_t	*pq = (pq_impl_t *) node;
	int		stride, remain;
	
	LOCK(pq->schedlink);

	if (pq->ready)
		pq_client_leave(pq);

	stride = STRIDE / tickets;
	remain = (pq->remain * stride) / pq->stride;

	if (old_tickets)
		*old_tickets = pq->tickets;

	pq->tickets = tickets;
	pq->stride  = stride;
	pq->remain  = remain;
	
	if (pq->ready)
		pq_client_join(pq);

	UNLOCK(pq->schedlink);
}

/*
 * Send a packet on a client node.
 */
int
pq_send_packet(pq_client_t *node, void *packet, int size)
{
	pq_impl_t	*pq = (pq_impl_t *) node;

	LOCK(pq->schedlink);

	/*
	 * Queue it up. Choke if no room and let the caller deal with
	 * it.
	 */
	DOSTAT(pq->sent++);
	if (pqueue_full(pq->pqueue)) {
		DOSTAT(pq->drops++);
		UNLOCK(pq->schedlink);
		return EOVERFLOW;
	}

#ifdef  OSKIT
	oskit_bufio_addref((oskit_bufio_t *) packet);
#endif
	/*
	 * There is probably an optimization we can perform here. If
	 * there are no other clients queued up to send, then we can
	 * probably send the packet right away and forget about queuing
	 * and scheduling. Under congestion queues will start to full
	 * and then scheduling matters, but without congestion it does
	 * not really matter who gets bandwidth. Right?
	 */
	pqueue_enqueue(pq->pqueue, packet, size);

	/*
	 * If the queue is going from empty, then we need to tell the
	 * scheduler (join the fray).
	 */
	if (!pq->ready) {
		pq_client_join(pq);
		pq->ready = 1;
	}

	/*
	 * Okay, look for the next packet to send.
	 */
	if (! pq_send_next(pq->schedlink))
		DOSTAT(pq->stalls++);

	UNLOCK(pq->schedlink);
	return 0;
}
