/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sched.h>

#include <nodeos/an_thread.h>
#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_thread.h>

#include "os.h"
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_threadpool.h>
#include <nodeos/impl/ani_time.h>

#include "ani_debug.h"

pthread_key_t ani_nodeos_thread;	/* thread struct for this pthread */
static pthread_key_t nodeos_threaddata;	/* nodeos exported thread data */

/*
 * For hack thread_sleep implementation.
 * Cond is never signalled, just allows us to do a timed wait.
 */
static pthread_mutex_t sleepmutex;
static pthread_cond_t sleepcond;

static void ani_thread_done(void *arg);
static void ani_thread_cleanup(void *arg);
static void *ani_thread_loop(void *arg);
static void ani_thread_intr_init(void);

void
ani_thread_setup(void)
{
	int err;

	/*
	 * Establish a thread_specific data key to hold the ani_thread_t
	 * associated with a pthread.  This is used to determine the
	 * current thread.  Init to zero for this primortial thread.
	 */
	err = pthread_key_create(&ani_nodeos_thread, 0);
	assert(err == 0);
	err = pthread_setspecific(ani_nodeos_thread, 0);
	assert(err == 0);

	/*
	 * Also need a key to use for the exported ani_thread_setdata
	 * interface.
	 */
	err = pthread_key_create(&nodeos_threaddata, 0);
	assert(err == 0);
	err = pthread_setspecific(nodeos_threaddata, 0);
	assert(err == 0);

	/*
	 * For thread sleep
	 */
	err = pthread_mutex_init(&sleepmutex, 0);
	assert(err == 0);
	err = pthread_cond_init(&sleepcond, 0);
	assert(err == 0);

	ani_thread_intr_init();
}

ani_error_t
ani_thread_init(ani_thread_t *thread, void *stackptr, size_t stacksize)
{
	int err;

	assert(thread != 0);

	memset(thread, 0, sizeof *thread);

	/*
	 * Set up the attributes for the thread.
	 * Currently we only set the stack size.
	 * We will set the stack pointer eventually.
	 */
	err = pthread_attr_init(&thread->attr);
	if (err != 0)
		return ani_error_set(err, "pthread_attr_init failed");
	if (stacksize != 0) {
		err = pthread_attr_setstacksize(&thread->attr, stacksize);
		if (err != 0)
			return ani_error_set(err,
					"pthread_attr_setstacksize failed");
	}

	/*
	 * XXX set global scope for the benefit of Solaris.
	 * Note: no check of the return value, calls fails with an error
	 * under FreeBSD 3.4.
	 */
	(void) pthread_attr_setscope(&thread->attr, PTHREAD_SCOPE_SYSTEM);

	/*
	 * Init some other easy stuff
	 */
	ani_ilock_init(&thread->lock);
	ani_icond_init(&thread->waiting);
	ani_icond_init(&thread->waitedon);

	thread->flags = THREAD_BEDEAD;

	return ani_error_clear();
}

ani_error_t
ani_thread_deinit(ani_thread_t *thread)
{
	assert(thread != 0);
	if ((thread->flags & THREAD_BEDEAD) == 0)
		return ani_error_set(EINVAL, "thread is active");

	ani_ilock_deinit(&thread->lock);
	ani_icond_deinit(&thread->waiting);
	ani_icond_deinit(&thread->waitedon);

	return ani_error_clear();
}

/* ani_thread_current is inlined in nodeos/impl/ani_thread.h */

ani_error_t
ani_thread_spawn(ani_threadpool_t *tpool, ani_thread_t **out_thread)
{
	ani_thread_t *thread;
	int stacksize;
	ani_error_t err;

	assert(tpool->params.ttype != AN_TH_EXPLICIT);

	if (tpool->curthreads >= tpool->params.threads.im.maxthreads)
		return ani_error_set(EAGAIN, "thread limit exceeded");

	stacksize = tpool->params.threads.im.stacksize;
	/* XXX someday allocate the actual stack */

	thread = ani_mempool_alloc(ani_flow_mempool(tpool->flow),
				   sizeof *thread);
	if (thread == 0)
		return ani_error_set(ENOMEM, "no memory for thread object");

	err = ani_thread_init(thread, 0, stacksize);
	if (ani_error_occured(err)) {
		ani_mempool_free(ani_flow_mempool(tpool->flow), thread,
				 sizeof *thread);
		return err;
	}

	err = ani_thread_start(thread, tpool);
	if (ani_error_occured(err)) {
		ani_thread_deinit(thread);
		ani_mempool_free(ani_flow_mempool(tpool->flow), thread,
				 sizeof *thread);
		return err;
	}

	if (out_thread)
		*out_thread = thread;
	return ani_error_clear();
}

/*
 * Start entrypoint for explicit threads.
 */
ani_error_t
ani_thread_start(ani_thread_t *thread, ani_threadpool_t *tpool)
{
#if 0
	void *stackptr;
#endif
	size_t stacksize;
	int ierr;

	assert_ani_threadpool_valid(tpool);
	assert(ani_threadpool_islocked(tpool));
	assert(thread != 0);

	if ((thread->flags & THREAD_BEDEAD) == 0)
		return ani_error_set(EINVAL, "thread not associated with pool");

#if 0
	ierr = pthread_attr_getstackaddr(&thread->attr, &stackptr);
	assert(ierr == 0);
#endif
	ierr = pthread_attr_getstacksize(&thread->attr, &stacksize);
	assert(ierr == 0);

	/*
	 * Create the pthread, it will block immediately on the pool lock.
	 */
	thread->pool = tpool;
	ierr = pthread_create(&thread->id, &thread->attr,
			      ani_thread_loop, thread);
	if (ierr != 0)
		return ani_error_set(ierr, "pthread_create failed");

	ierr = pthread_detach(thread->id);
	assert(ierr == 0);

	/* Set the default scheduling priority */
	thread->nprio = AN_THREAD_PRIO_NORMAL;
	ierr = os_sched_setprio(thread, thread->nprio);
	assert(ierr == 0);

	/* We are committed, finish up */
	thread->flags = THREAD_INKERNEL | (thread->flags & THREAD_ONESHOT);
	if ((thread->flags & THREAD_ONESHOT) == 0) {
		thread->next = tpool->threads;
		tpool->threads = thread;
		tpool->curthreads++;
	} else {
		thread->next = tpool->osthreads;
		tpool->osthreads = thread;
		tpool->curosthreads++;
	}
	ani_threadpool_reference(tpool);

	return ani_error_clear();
}

/*
 * All NodeOS threads spend their lives here with brief forays into EE/AA code.
 */
static void *
ani_thread_loop(void *arg)
{
	ani_thread_t *thread = arg;
	ani_threadpool_t *tpool;
	struct ani_threadwork *w, wtmp;

	/*
	 * Record our NodeOS identity in the pthread so thread_current works.
	 */
	pthread_setspecific(ani_nodeos_thread, thread);
	assert_ani_thread_valid(thread);

	/*
	 * Get the threadpool lock.
	 * We will block here until our creator is finished initializing us.
	 */
	tpool = thread->pool;
	assert(tpool != 0);
	ani_threadpool_lock(tpool);
	DBG(DBG_THREADS, ani_dprintf("TP%p: thread %p started\n", tpool, thread));

	/*
	 * Now that we are official, push a cleanup handler, and go to work.
	 */
	pthread_cleanup_push(ani_thread_cleanup, thread);
	while (1) {
		/*
		 * If there is no work to do we either block waiting for
		 * more or die if there are enough threads in reserve.
		 */
		while (thread->work == 0 && tpool->workhead == 0) {
			if (tpool->freethreads >= tpool->reserve)
				goto die;
			thread->freenext = tpool->freelist;
			tpool->freelist = thread;
			tpool->freethreads++;
#ifdef THREADPOOL_STATS
			if (tpool->freethreads > tpool->stats.maxcached)
				tpool->stats.maxcached = tpool->freethreads;
#endif
			/*
			 * There is now a thread available to do work.
			 * Wake anyone who might have been waiting for that.
			 */
			if ((tpool->flags & TPOOL_WAITING) != 0) {
				tpool->flags &= ~TPOOL_WAITING;
				ani_icond_broadcast(&tpool->waiting);
			}
			DBG(DBG_THREADS|DBG_THREADPOOLS, {
				ani_dprintf("TP%p: thread %p sleeping, freethreads=%d\n",
					    tpool, thread, tpool->freethreads);
			});
			ani_icond_wait(&thread->waiting, &tpool->lock);
			DBG(DBG_THREADS|DBG_THREADPOOLS, {
				ani_dprintf("TP%p: thread %p awake, freethreads=%d\n",
					    tpool, thread, tpool->freethreads);
			});

			assert_ani_thread_valid(thread);
		}
		assert((thread->flags & THREAD_ISFREE) == 0);

#ifdef THREADPOOL_STATS
		{
			int active = tpool->curthreads - tpool->freethreads;

			if (active > tpool->stats.maxactive)
				tpool->stats.maxactive = active;
		}
#endif

		/*
		 * First look for work specifically targetted to this thread.
		 * Otherwise, grab the first item off the list.
		 */
		w = thread->work;
		if (w == 0) {
			w = tpool->workhead;
			if (w->next == 0) {
				assert(tpool->worktail == w);
				tpool->workhead = tpool->worktail = 0;
			} else
				tpool->workhead = w->next;
			tpool->worklen--;
		} else
			thread->work = 0;

		/*
		 * Make a local copy so we can free the heap element
		 */
		wtmp = *w;
		ani_mempool_free(ani_flow_mempool(tpool->flow), w, sizeof *w);
		w = &wtmp;

		thread->chan = w->chan;

		DBG(DBG_THREADS, {
			ani_dprintf("TP%p: thread %p doing %p(%p)\n",
				    tpool, thread, w->func, w->farg);
		});

		/*
		 * Run the desired function.
		 *
		 * The sigsetjmp catches threads that call ani_thread_exit
		 * explicitly (as opposed to returning from the registered
		 * upcall func).  Those threads will siglongjmp back to
		 * here unwinding their stacks.
		 *
		 * The cleanup handler ensure that anybody awaiting this
		 * thread is properly awakened in the event of a cancellation.
		 * Note the "1" argument to cleanup_pop that causes
		 * ani_thread_done to be called even when no cancellation
		 * occured.
		 */
		pthread_cleanup_push(ani_thread_done, thread);
		ani_threadpool_unlock(tpool);
		if (sigsetjmp(thread->jmpbuf, 0) == 0) {
			if (w->inkern == 0) {
				thread->flags |= THREAD_INTSON;
				NODEOS_LEAVE;
				w->func(w->farg[0], w->farg[1]);
				NODEOS_ENTER;
				assert_ani_thread_valid(thread);
			} else
				w->func(w->farg[0], w->farg[1]);
		} else
			thread = ani_thread_current();
		tpool = thread->pool;
		pthread_cleanup_pop(1);	/* calls ani_thread_done */
		ani_threadpool_lock(tpool);
		DBG(DBG_THREADS, {
			ani_dprintf("TP%p: thread %p did %p(%p)\n",
				    tpool, thread, w->func, w->farg);
		});

		/*
		 * If this was a one-shot thread, just exit now.
		 */
		if ((thread->flags & THREAD_ONESHOT) != 0)
			goto die2;
	}
 die:
	DBG(DBG_THREADS|DBG_THREADPOOLS, {
		ani_dprintf("TP%p: thread %p dying, freethreads=%d, reserve=%d\n",
			    tpool, thread, tpool->freethreads, tpool->reserve);
	});
	
	/*
	 * Wake up anyone waiting for the freelist to drop to the desired
	 * reserve level.
	 */
	if ((tpool->flags & TPOOL_WAITING) != 0 &&
	    tpool->freethreads == tpool->reserve) {
		tpool->flags &= ~TPOOL_WAITING;
		ani_icond_broadcast(&tpool->waiting);
	}

 die2:
	ani_threadpool_unlock(thread->pool);

	DBG(DBG_THREADS|DBG_THREADPOOLS, {
		ani_dprintf("TP%p: thread %p pthread_exit'ing.\n",
			    tpool, thread);
	});

	/*
	 * Call pthread_exit explicitly here instead of returning
	 * since the cleanup handler must be "in scope" when we exit.
	 */
	pthread_exit(0);
	pthread_cleanup_pop(1);

	DBG(DBG_THREADS|DBG_THREADPOOLS, {
		ani_dprintf("TP%p: thread %p returning from whence it came (post exit?!).\n",
			    tpool, thread);
	});

	return 0;
}

/*
 * Kill (cancel) a thread and optionally wait for it to die.
 * Thread's pool should be locked on entry.
 */
void
ani_thread_kill(ani_thread_t *thread, int dowait)
{
	ani_threadpool_t *tpool;

	assert(thread != ani_thread_current());

	tpool = thread->pool;
	assert_ani_threadpool_valid(tpool);

	ani_threadpool_lock(tpool);
	ani_thread_kill_locked(thread, dowait);
	ani_threadpool_unlock(tpool);
}

void
ani_thread_kill_locked(ani_thread_t *thread, int dowait)
{
	ani_error_t err;
	int ierr;

	ierr = pthread_cancel(thread->id);
	assert(ierr == 0);

	if (dowait) {
		err = ani_threadpool_thread_wait(thread->pool, thread, 0);
		assert(!ani_error_occured(err));
	}
}

void
ani_thread_exit(void)
{
	/* Just exit, the actual work is done in the cleanup handler */
	pthread_exit(0);
}

/*
 * Wake any thread waiting for this one to finish and wait for it to
 * acknowledge.  Called from the main thread loop (or pthread_exit) to
 * ensure a thread's previous commitments are completed before returning
 * to the common pool (or exiting).
 */
static void
ani_thread_done(void *arg)
{
	ani_thread_t *thread = arg;
	ani_threadpool_t *tpool = thread->pool;

	assert_ani_threadpool_valid(tpool);

	if (thread->chan) {
		ani_chan_thread_exit(thread->chan);
		thread->chan = 0;
	}

	ani_threadpool_lock(tpool);
	if (thread->flags & THREAD_WAITEDON) {
		DBG(DBG_THREADS, {
			ani_dprintf("TP%p: thread %p finishing\n", tpool, thread);
		});
		thread->flags &= ~THREAD_WAITEDON;
		ani_icond_signal(&thread->waitedon);
		thread->flags |= THREAD_DONE;
		while ((thread->flags & THREAD_DONE) != 0)
			ani_icond_wait(&thread->waiting, &tpool->lock);
		DBG(DBG_THREADS, {
			ani_dprintf("TP%p: thread %p finished\n", tpool, thread);
		});
	}
	assert((thread->flags & (THREAD_WAITEDON|THREAD_DONE)) == 0);
	ani_threadpool_unlock(tpool);
}

/*
 * Cleanup a thread's resources.
 * This is the pthread cleanup handler called for every thread when it
 * exits or is canceled.
 */
static void
ani_thread_cleanup(void *arg)
{
	ani_thread_t *thread;
	ani_thread_t **pth, *th;
	ani_threadpool_t *tpool;

	thread = arg;
	assert(thread == ani_thread_current());
	assert((thread->flags & THREAD_WAITEDON) == 0);
	assert((thread->flags & THREAD_BEDEAD) == 0);

	tpool = thread->pool;
	assert_ani_threadpool_valid(tpool);

	ani_threadpool_lock(tpool);
	DBG(DBG_THREADS, {
		ani_dprintf("TP%p: thread %p cleanup oneshot=%s curthreads=%d curosthreads=%d\n",
			    tpool, thread,
			    ((thread->flags & THREAD_ONESHOT) == 0)?"no":"yes",
			    tpool->curthreads, tpool->curosthreads);
	});

	if ((thread->flags & THREAD_ONESHOT) == 0)
		pth = &tpool->threads;
	else
		pth = &tpool->osthreads;
	while ((th = *pth) != 0) {
		if (th == thread)
			break;
		pth = &th->next;
	}
	assert(th != 0);

	/*
	 * XXX we could just remove the thread from the pool freelist if
	 * it is there, but right now we should never be cancelling a thread
	 * that is on the freelist (we wake it up and let it die instead).
	 */
	assert((thread->flags & THREAD_ISFREE) == 0);

	*pth = thread->next;
	thread->flags |= THREAD_BEDEAD;

	if ((thread->flags & THREAD_ONESHOT) == 0)
		tpool->curthreads--;
	else
		tpool->curosthreads--;

	if (tpool->curthreads + tpool->curosthreads == 0 &&
	    (tpool->flags & TPOOL_WAITING) != 0) {
		tpool->flags &= ~TPOOL_WAITING;
		ani_icond_broadcast(&tpool->waiting);
	}

	ani_threadpool_unlock(tpool);
	ani_threadpool_release(tpool);
}

void
ani_thread_yield(void)
{
	sched_yield();
}

/*
 * XXX hack, uninterruptable (though cancelable) sleep implementation.
 * XXX assumes that an_timespec is same as timespec
 *
 * I'm not convinced ani_thread_sleep is all that useful or I would spend some
 * time doing it right.
 */
void
ani_thread_sleep(an_timespec_t *delay)
{
	int err;
	an_timespec_t wakeup;

	ani_time_future(&wakeup, delay);

	err = pthread_mutex_lock(&sleepmutex);
	assert(err == 0);

	do {
		pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock,
				     &sleepmutex);
		err = pthread_cond_timedwait(&sleepcond, &sleepmutex, &wakeup);
		pthread_cleanup_pop(0);
	} while (err == 0);
	assert(err == EINVAL || err == ETIMEDOUT);

	err = pthread_mutex_unlock(&sleepmutex);
	assert(err == 0);
}

void *
ani_thread_getdata(void)
{
	return pthread_getspecific(nodeos_threaddata);
}

void
ani_thread_setdata(void *data)
{
	int err;

	err = pthread_setspecific(nodeos_threaddata, data);
	assert(err == 0);
}

int
ani_thread_getprio(ani_thread_t *thread)
{
	return thread->nprio;
}

ani_error_t
ani_thread_setprio(ani_thread_t *thread, int nprio)
{
	int err;

	err = os_sched_setprio(thread, nprio);
	if (err != 0)
		return ani_error_set(err, "os_sched_setprio failed");

	thread->nprio = nprio;
	return ani_error_clear();
}

/*
 * Get various bits of stack information on a specific
 * thread.  The thread may be the current thread.
 * If the target thread is running (on an SMP machine)
 * the results are undefined.
 *
 * XXX this call may evolve into a more generic ani_thread_info call.
 */
ani_error_t
ani_thread_stackinfo(ani_thread_t *thread, void **base, size_t *size,
		     void **ptr)
{
#ifdef OSKIT
	struct pthread_state ps;
	int err;

	err = oskit_pthread_getstate(thread->id, &ps);
	if (err != 0)
		return ani_error_set(err, "oskit_pthread_getstate failed");

	*base = (void *)ps.stackbase;
	*size = ps.stacksize;
	*ptr = (void *)ps.stackptr;
	return ani_error_clear();
#else
	return ani_error_set(ENODEV, "ani_thread_stackinfo not implemented");
#endif
}

/*
 * Thread "interrupt" support.
 *
 * With pthreads, signals are the only way we can do this.
 * Pthread cancelation implies the death of the canceled thread,
 * NodeOS interrupts do not.
 */
#include <signal.h>

#define SIG_THREADINTR	SIGUSR1 /* XXX */
sigset_t ani_thread_intrmask;

static void ani_thread_intr_handler(int sig, siginfo_t *si, void *sc);

static void
ani_thread_intr_init(void)
{
	struct sigaction act;

	sigemptyset(&ani_thread_intrmask);
	sigaddset(&ani_thread_intrmask, SIG_THREADINTR);

	/*
	 * Catch relevant signals.
	 */
#ifdef SA_SIGINFO
	act.sa_flags = SA_SIGINFO;
	act.sa_sigaction = ani_thread_intr_handler;
#else
	(void*)(act.sa_handler) = (void*)ani_thread_intr_handler;
#endif
	sigemptyset(&act.sa_mask);

	sigaction(SIG_THREADINTR, &act, 0);
	sigaction(SIGILL, &act, 0);
#ifdef SIGEMT
	sigaction(SIGEMT, &act, 0);
#endif
	sigaction(SIGFPE, &act, 0);
	sigaction(SIGBUS, &act, 0);
	sigaction(SIGSEGV, &act, 0);
}

void
ani_thread_interrupt(ani_thread_t *thread)
{
	int err;
	ani_cond_t *cond;

	ani_thread_lock(thread);

	/*
	 * Send a signal to said thread
	 */
	err = pthread_kill(thread->id, SIG_THREADINTR);
	assert(err == 0);

	/*
	 * If thread is in a cond wait, wake it up.
	 * This gets really strange because we must acquire the cond
	 * lock before the thread lock (because of how ani_thread_condwait
	 * works).
	 */
 again:
	cond = thread->cond;
	if (cond) {
		/*
		 * Cond lock is held.  Release the thread lock and wait
		 * for the cond.  Then reaquire the thread lock and see
		 * if it is still waiting on the same cond.  If not,
		 * start over.
		 */
		if (ani_cond_trylock(cond)) {
			ani_thread_unlock(thread);
			ani_cond_lock(cond);
			ani_thread_lock(thread);
			if (thread->cond != cond) {
				ani_cond_unlock(cond);
				goto again;
			}
		}
		ani_icond_broadcast(&cond->icond);
		ani_cond_unlock(cond);
	}

	ani_thread_unlock(thread);
}

void
ani_thread_block_interrupt(void)
{
	ani_thread_current()->flags &= ~THREAD_INTSON;
}

void
ani_thread_unblock_interrupt(void)
{
	ani_thread_current()->flags |= THREAD_INTSON;
}

void
ani_thread_sethandler(an_threadintr_func_t handler)
{
	ani_thread_current()->handler = handler;
}

/*
 * Signal handler for all exceptions and NodeOS thread interrupts
 *
 * XXX we are probably doing a lot of things here that a pthread signal
 * handler should not do.
 */
static void
ani_thread_intr_handler(int sig, siginfo_t *si, void *sc)
{
	ani_thread_t *thread = ani_thread_current();
	int intson, type;

	if (thread == 0 || (thread->flags & THREAD_INKERNEL)) {
		os_dump_state(sc);
		panic("NodeOS: got signal %d\n", sig);
	}

	NODEOS_ENTER;

	/*
	 * No handler for thread, just exit
	 *
	 * XXX should kill the flow.  Rationale: if an EE plans on
	 * interrupting a thread, it should be prepared to catch it.
	 * And if a thread faults, that is clearly a flow/EE error.
	 * By just killing the thread as we do now, we have effectively
	 * introduced an unsafe ani_thread_kill operation.  Me bad.
	 */
	 if (thread->handler == 0) {
		 printf("ani_thread_intr_handler: thread has no handler:\n");
		 os_dump_state(sc);
		 (void)pthread_cancel(thread->id);
		 return;
	 }

	/*
	 * Otherwise, call the handler with interrupts disabled
	 */
	intson = (thread->flags & THREAD_INTSON);
	thread->flags &= ~THREAD_INTSON;

	/*
	 * Map the signal type to the appropriate AN_THREAD_ constant.
	 */
	switch (sig)
	{
	case SIG_THREADINTR:
		assert(intson);
		type = AN_THREAD_INTR;
		break;
	case SIGFPE:
		type = AN_THREAD_TRAP_FPU;
		break;
#if defined(SIGBUS)
	case SIGBUS:  /* map to SEGV */
#endif
	case SIGSEGV:
	default:
		type = AN_THREAD_TRAP_SEGV;
		break;
	}

	NODEOS_LEAVE;
	thread->handler(type, sc);
	NODEOS_ENTER;

	thread->flags |= intson;
	NODEOS_LEAVE;
}

/*
 * Block a thread on the specified condition variable.
 * The thread may be awakened either by the condition being signaled through
 * an API call or because of ani_thread_interrupt.  We do not differentiate the
 * two cases here.
 *
 * If abstime is not NULL, then the cond wait may timeout if the given
 * absolute time is reached.
 *
 * Called with thread and condition locked.
 */
void
ani_thread_condtimedwait(ani_cond_t *cond, an_timespec_t *abstime)
{
	ani_thread_t *thread = ani_thread_current();
	assert(thread->cond == NULL);

	thread->cond = cond;
	if (abstime == NULL) 
	{
		ani_icond_wait(&cond->icond, &cond->ilock);
	}
	else
	{
		ani_icond_timedwait(&cond->icond, &cond->ilock,
				    abstime->tv_sec, abstime->tv_nsec);
	}
	
	thread->cond = NULL;
}
